/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
*
* madgecb.c
*
* Copyright (c) Madge Networks Ltd
*
*
*****************************************************************************/
/* Please refer to the LICENSE file supplied with this source code */


#include <linux/config.h>
#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif

#if LINUX_VERSION_CODE < 0x20402
#include <pcmcia/k_compat.h>
#else
#include "k_compat.h"
#endif
 
#include "driver.h"

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/netdevice.h>

/* new ones */
#include <linux/types.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/irq.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include <pcmcia/driver_ops.h>

#ifdef DEBUG
	#include "mpoke.h"
#endif

#if LINUX_VERSION_CODE < 0x20155
#include <linux/bios32.h>
#endif

#include <asm/irq.h>                    /* For NR_IRQS only. */
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/string.h>

#include "madgecb.h"

static const char * version = 
        MTOKCB_PRODUCT_NAME "\n" MTOKCB_PRODUCT_COPYRIGHT "\n";

#ifdef MVER_STRINGS
/*-------------------------------------------------------------------------
|
| PRODUCT TRACKING STRINGS - These MUST NOT be changed.
|
|-------------------------------------------------------------------------*/
char mtokcb_mver_title[] = "VeRsIoN=" MTOKCB_PRODUCT_NAME;
char mtokcb_mver_copyright[] = "VeRsIoN=" MTOKCB_PRODUCT_COPYRIGHT;
char mtokcb_mver_platform[] = "VeRsIoN=PLATFORM: " MTOKCB_PRODUCT_PLATFORM;
char mtokcb_mver_version[61] = "VeRsIoN=VERSION: v" MTOKCB_VERSION;
char mtok_cb_crcp[] = "MaDgE=XXXX";
#endif                

static dev_node_t *mtokcb_attach(dev_locator_t *loc);
static void mtokcb_detach(dev_node_t * node);

extern LINUXDEV *mtok_get_dev_by_name(char * name); 

/* Initialise, add us into list, display details */
extern LINUXDEV * mtok_probe1(LINUXDEV *dev,
                            ADAPTER_ID_T  * padapter_id,int irq,
                            int maxframesize,int txslots,int rxslots,
                            int ringspeed,int dtr,void *UsrData);
extern void mtok_remove_dev(LINUXDEV *dev); 
extern void mtok_cleanup(void *);

void mtokcb_suspend(dev_node_t * dev)
{
}

void mtokcb_resume(dev_node_t *dev)
{
}


struct driver_operations madge_ops = {
    "mtok_cb",
   	mtokcb_attach, 
   	mtokcb_suspend,
   	mtokcb_resume,
   	mtokcb_detach
};    


typedef struct {
    u16                 vendorid;
    u16                 deviceid;
    u8                  revision;
    u16                 subsystemid;
    ADAPTER_TYPE_T      adapter_type;
    ADAPTER_SUBTYPE_T   adapter_subtype;
} mtok_card_id;

static mtok_card_id cbcards[] = {
#ifdef CARDBUS2_UDI       /*** Cardbus Mk 2 ***/
    {   
        PCI_MADGE_VENDOR_ID, CARDBUS2_DEVICE_ID, 
        0xFF, 0xFFFF, 
        ADAPTER_TYPE_KNOSSOS_CARDBUS, ADAPTER_SUBTYPE_CARDBUS2
    },
#endif
#ifdef OLICOM_CARDBUS_UDI
    {   
        PCI_OLICOM_VENDOR_ID, OLICOM_PCI_3250_DEVICE_ID, 
        0xFF, 0xFFFF,
        ADAPTER_TYPE_OLICOM_CARDBUS, ADAPTER_SUBTYPE_OLICOM_CARDBUS_3250
    },
#endif
};

#define NUM_CB_ADAPTERS (sizeof(cbcards)/sizeof(mtok_card_id))

static dev_node_t *mtokcb_attach(dev_locator_t *loc)
{
    LINUXDEV * dev;
    u16 dev_id, vendor_id;
    u32 io;
    u8 bus, devfn, irq;
    int i,found;
    ADAPTER_ID_T	adapter_id;
#ifndef OLD_PCI 
    struct pci_dev *pcidev;
#endif

#ifdef DEBUG 
    QPOKE(PO_YELLOW,"[Attach");
#endif

    if (loc->bus != LOC_PCI)
    {
#ifdef DEBUG      
	QPOKE(PO_RED,"NotPCI!");
        QPOKE(PO_YELLOW,"]");
#endif
        return NULL;
    }
    
    bus = loc->b.pci.bus;
    devfn = loc->b.pci.devfn;

#ifdef DEBUG	
    QPOKEW(PO_MAGENTA,"Bus=",bus);
    QPOKEW(PO_CYAN,"Devfn=",devfn);
#endif

#ifdef OLD_PCI
    pcibios_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &io);
    pcibios_read_config_byte(bus, devfn, PCI_INTERRUPT_LINE, &irq);
    pcibios_read_config_word(bus, devfn, PCI_VENDOR_ID, &vendor_id);
    pcibios_read_config_word(bus, devfn, PCI_DEVICE_ID, &dev_id);
#else
    pcidev = (struct pci_dev *)pci_find_slot(bus,devfn);
    pci_read_config_dword(pcidev, PCI_BASE_ADDRESS_0, &io);
    pci_read_config_byte(pcidev, PCI_INTERRUPT_LINE, &irq);
    pci_read_config_word(pcidev, PCI_VENDOR_ID, &vendor_id);
    pci_read_config_word(pcidev, PCI_DEVICE_ID, &dev_id);
#endif /* OLD_PCI */

#ifdef DEBUG	
    QPOKEW(PO_YELLOW,"Vendor=",vendor_id);
    QPOKEW(PO_GREEN,"Device=",dev_id);
#endif

    io &= ~3;

    if (io == 0 || irq == 0) {
        printk(KERN_ERR "The Card Bus Token Ring interface was not "
                        "assigned an %s.\n" KERN_ERR "  It will not be activated.\n",
                        io == 0 ? "I/O address" : "IRQ");
        return NULL;
    }

    found = 0;
    for(i=0;i<NUM_CB_ADAPTERS;i++)
    {
        if (vendor_id != cbcards[i].vendorid)
            continue;

        if (dev_id != cbcards[i].deviceid)
            continue;
                                                
       	adapter_id.adapter_type = cbcards[i].adapter_type;
       	adapter_id.adapter_subtype = cbcards[i].adapter_subtype;
       	adapter_id.field_used_to_identify_card = USE_PCI_NUMBER;
       	adapter_id.pci_number = (bus << 8) | devfn;
       	adapter_id.io_base = io;
       	adapter_id.adapter_state = ADAPTER_STATE_UNKNOWN;

        found = 1;
        break;
    }	

    if (!found)
    {
#ifdef DEBUG	
        QPOKE(PO_RED,"CardNotRecognised!");
        QPOKE(PO_YELLOW,"]");
#endif
        return NULL;
    }
    
	printk(KERN_INFO "mtok_cb: Attaching to Cardbus Adapter at IO = %X IRQ = %d\n",io,irq);

	dev = mtok_probe1(NULL,&adapter_id,irq,0,0,0,0,0,NULL);

	if (dev)
	{
		dev_node_t * node = kmalloc(sizeof(dev_node_t),GFP_KERNEL);
		strcpy(node->dev_name,dev->name);
		node->major = node->minor = 0;
		node->next = NULL;
        	MOD_INC_USE_COUNT;
		return node;
	}
	return NULL;                                                                                                                               
}



static void mtokcb_detach(dev_node_t * node)
{
    LINUXDEV *dev = mtok_get_dev_by_name(node->dev_name);

#ifdef DEBUG	
    QPOKE(PO_YELLOW,"[Detach");
#endif

	if (!dev)
	{
#ifdef DEBUG	
	        QPOKE(PO_RED,"Dev=NULL");
	        QPOKE(PO_YELLOW,"]");
#endif
		return;
	}

	
	mtok_cleanup(dev->priv);
	mtok_remove_dev(dev);
	kfree(dev);
	MOD_DEC_USE_COUNT;

#ifdef DEBUG	
	QPOKE(PO_YELLOW,"]");
#endif

}

int init_module(void)
{
    printk(version);
    register_driver(&madge_ops);
    return 0;
}

void cleanup_module(void)
{
    unregister_driver(&madge_ops);
}
