/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
*
* mpoke.c
*
* Linux pokeout driver.
*
* Copyright (c) Madge Networks Ltd
* Company Confidential
*
*****************************************************************************/

#include <linux/autoconf.h>

#include <linux/types.h>
#include <linux/module.h>
#include <stdarg.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <asm/io.h>
#include <asm/atomic.h>
/* #include <asm/spinlock.h> */
#include <asm/uaccess.h>

#include "mpoke.h"

/*----------------------------------------------------------------------------
|
| Version information.
|
|---------------------------------------------------------------------------*/

#define VERSION_NUMBER "1.00.00"


/*----------------------------------------------------------------------------

1.00.00         PBA         19/10/1999

                First version.
                
----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
|
| Configuration constants.
|
|---------------------------------------------------------------------------*/

#define MAJOR_NUMBER       120     /* Our device file's major number. */
#define DRIVER_NAME        "mpoke" /* Our public module name. */

#define POKEBUF_SIZE       16384   /* Size of the pokeout ring buffer. Must */
                                   /* be a power of 2. */

#define MAX_CALLERS        32      /* Maximum number of clients. */

#define VGA_MEMBASE        0xb8000 /* Physical base of VGA memory. */
#define VGA_MEMSIZE        4000    /* Size of VGA memory extent to use for */
                                   /* pokeouts. */
#define VGA_CMDPORT        0x3d4   /* Address of VGA control/command port. */
#define VGA_CMD_SETMEMBASE 12      /* VGA command to set the base of visible */
                                   /* memory. */

#define VGA_RESET_TIME     (10 * HZ) /* We will reset the visible extend of */
                                     /* VGA memory each VGA_RESET_TIME */
                                     /* jiffies. */
                                     
                                     
/*----------------------------------------------------------------------------
|
| Debugging.
|
|---------------------------------------------------------------------------*/

#ifdef __DEBUG__

#define MDBG1(_x)   (_x)
#define MDBG2(_x)   (_x)
#define MDBG3(_x)   (_x)

#define MPOKEF      printk("mpoke %04d: ", __LINE__), printk

#define STATIC

#else /* __DEBUG__ */

#define MDBG1(_x)
#define MDBG2(_x)
#define MDBG3(_x)

#define MPOKEF

#define STATIC      static

#endif /* __DEBUG__ */


/*----------------------------------------------------------------------------
|
| Extra return code.
|
|---------------------------------------------------------------------------*/

#define EOK  0


/*----------------------------------------------------------------------------
|
| BOOLEAN type.
|
|---------------------------------------------------------------------------*/

#define BOOLEAN  int
#define TRUE     1
#define FALSE    0


/*----------------------------------------------------------------------------
|
| List of registered clients.
|
|---------------------------------------------------------------------------*/

STATIC
char mpoke_caller_names[MAX_CALLERS][32] = {{0}};


/*----------------------------------------------------------------------------
|
| ANSI colour strings.
|
|---------------------------------------------------------------------------*/

STATIC
char * mpoke_ansi_colour_codes[32] =
{
    "\033[0;37;40m",    /* Dim white on black. */
    "\033[0;30;40m",    /* Dim black on black. */
    "\033[0;34;40m",    /* Dim blue on black. */
    "\033[0;32;40m",    /* Dim green on black. */
    "\033[0;36;40m",    /* Dim cyan on black. */
    "\033[0;31;40m",    /* Dim red on black. */
    "\033[0;35;40m",    /* Dim magenta on black. */
    "\033[0;33;40m",    /* Dim yellow on black. */
    
    "\033[0;1;37;40m",  /* Bright white on black. */
    "\033[0;1;30;40m",  /* Bright black on black. */
    "\033[0;1;34;40m",  /* Bright blue on black. */
    "\033[0;1;32;40m",  /* Bright green on black. */
    "\033[0;1;36;40m",  /* Bright cyan on black. */
    "\033[0;1;31;40m",  /* Bright red on black. */
    "\033[0;1;35;40m",  /* Bright magenta on black. */
    "\033[0;1;33;40m",  /* Bright yellow on black. */
    
    "\033[0;37;47m",    /* Dim white on white. */
    "\033[0;30;47m",    /* Dim black on white. */
    "\033[0;34;47m",    /* Dim blue on white. */
    "\033[0;32;47m",    /* Dim green on white. */
    "\033[0;36;47m",    /* Dim cyan on white. */
    "\033[0;31;47m",    /* Dim red on white. */
    "\033[0;35;47m",    /* Dim magenta on white. */
    "\033[0;33;47m",    /* Dim yellow on white. */
    
    "\033[0;1;37;47m",  /* Bright white on white. */
    "\033[0;1;30;47m",  /* Bright black on white. */
    "\033[0;1;34;47m",  /* Bright blue on white. */
    "\033[0;1;32;47m",  /* Bright green on white. */
    "\033[0;1;36;47m",  /* Bright cyan on white. */
    "\033[0;1;31;47m",  /* Bright red on white. */
    "\033[0;1;35;47m",  /* Bright magenta on white. */
    "\033[0;1;33;47m",  /* Bright yellow on white. */
};


/*----------------------------------------------------------------------------
|
| VGA colour attributes.
|
|---------------------------------------------------------------------------*/

STATIC
u16 mpoke_vga_colour_attr[32] =
{
    0x0700,    /* Dim white on black. */
    0x0000,    /* Dim black on black. */
    0x0100,    /* Dim blue on black. */
    0x0200,    /* Dim green on black. */
    0x0300,    /* Dim cyan on black. */
    0x0400,    /* Dim red on black. */
    0x0500,    /* Dim magenta on black. */
    0x0600,    /* Dim yellow on black. */
    
    0x0f00,    /* Bright white on black. */
    0x0800,    /* Bright black on black. */
    0x0900,    /* Bright blue on black. */
    0x0a00,    /* Bright green on black. */
    0x0b00,    /* Bright cyan on black. */
    0x0c00,    /* Bright red on black. */
    0x0d00,    /* Bright magenta on black. */
    0x0e00,    /* Bright yellow on black. */
    
    0x7700,    /* Dim white on white. */
    0x7000,    /* Dim black on white. */
    0x7100,    /* Dim blue on white. */
    0x7200,    /* Dim green on white. */
    0x7300,    /* Dim cyan on white. */
    0x7400,    /* Dim red on white. */
    0x7500,    /* Dim magenta on white. */
    0x7600,    /* Dim yellow on white. */
    
    0x7f00,    /* Bright white on white. */
    0x7800,    /* Bright black on white. */
    0x7900,    /* Bright blue on white. */
    0x7a00,    /* Bright green on white. */
    0x7b00,    /* Bright cyan on white. */
    0x7c00,    /* Bright red on white. */
    0x7d00,    /* Bright magenta on white. */
    0x7e00,    /* Bright yellow on white. */
};
    
    
/*----------------------------------------------------------------------------
|
| Hex characters.
|
|---------------------------------------------------------------------------*/

STATIC
char mpoke_hex_chars[16] = "0123456789abcdef";

  
/*----------------------------------------------------------------------------
|
| The cyclic pokeout buffer.
|
|---------------------------------------------------------------------------*/

STATIC
char mpoke_pokebuf[POKEBUF_SIZE];

/*
 * The head pointer is only changed by the buffer writing code. The tail
 * pointer is only changed by the buffer reading code. Both pointers need
 * to be atomic as both readers and writers may read the values.
 */
 
STATIC
atomic_t mpoke_pokebuf_head; 

STATIC             
atomic_t mpoke_pokebuf_tail; 


/*----------------------------------------------------------------------------
|
| Pointer to the function that actually pokes out a string.
|
|---------------------------------------------------------------------------*/

void (*mpoke_outstr)(int ident, int colour, char * str);


/*----------------------------------------------------------------------------
|
| Miscellaneous global state.
|
|---------------------------------------------------------------------------*/

STATIC
int mpoke_open_count = 0;            /* Count of open() operations. */
  
STATIC
int mpoke_prev_ident = -1;           /* Identifier of the last caller. */

STATIC
int mpoke_prev_colour = -1;          /* Previous colour number used. */

STATIC
int mpoke_vga_offs = 0;              /* Offset into the VGA memory used for */
                                     /* pokeouts. */
                                     
STATIC
unsigned long mpoke_reset_time;      /* Next jiffy at which we should reset */
                                     /* the VGA visible memory base. */
                                     
STATIC
BOOLEAN mpoke_using_colour = TRUE;   /* Are we doing colour pokeouts? */

STATIC
spinlock_t mpoke_lock;               /* Spin-lock used to ensure that we're */
                                     /* single threaded. */
STATIC struct
wait_queue * mpoke_read_wait = NULL; /* Wait queue on which to block */
                                     /* mpoke_read() when there's no data to */
                                     /* read. */
                                    
STATIC
atomic_t mpoke_read_blocked;         /* Are we blocked inside mpoke_read()? */
                                    
                                    
/*----------------------------------------------------------------------------
|
| Entry point table for a character device.
|
|---------------------------------------------------------------------------*/

STATIC ssize_t
mpoke_read(
    struct file  * filep,
    char         * bufp,
    size_t         count,
    loff_t       * posp);    

STATIC ssize_t
mpoke_write(
    struct file  * filep,
    const char   * bufp,
    size_t         count,
    loff_t       * posp);    

STATIC int
mpoke_open(
    struct inode * inodep,
    struct file  * filep);
    
STATIC int
mpoke_release(
    struct inode * inodep,
    struct file  * filep);
    
STATIC
struct file_operations mpoke_fops =
{
    NULL,            /* lseek */
    mpoke_read,      /* read */
    mpoke_write,     /* write */
    NULL,            /* readdir */
    NULL,            /* poll */
    NULL,            /* ioctl */
    NULL,            /* mmap */
    mpoke_open,      /* open */
    NULL,            /* flush */
    mpoke_release,   /* release */
    NULL,            /* fsync */
    NULL,            /* fasync */
    NULL,            /* check_media_change */
    NULL,            /* revalidate */
    NULL             /* lock */
};

    
/*----------------------------------------------------------------------------
|
| Macro       - POKEBUF_APPEND
|
| Parameters  - str -> A string to append to the pokeout buffer.
|
| Purpose     - Append a string to the cyclic pokeout buffer.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

#define POKEBUF_APPEND(str)                                     \
{                                                               \
    char * ptr;                                                 \
    int    head;                                                \
    int    new_head;                                            \
                                                                \
    head = atomic_read(&mpoke_pokebuf_head);                    \
                                                                \
    for (ptr = (str); *ptr != '\0'; ptr++)                      \
    {                                                           \
        new_head = (head + 1) & (POKEBUF_SIZE - 1);             \
                                                                \
        if (new_head != atomic_read(&mpoke_pokebuf_tail))       \
        {                                                       \
            mpoke_pokebuf[head] = *ptr;                         \
            head                = new_head;                     \
        }                                                       \
    }                                                           \
                                                                \
    atomic_set(&mpoke_pokebuf_head, head);                      \
}
 

/*----------------------------------------------------------------------------
|
| Function    - mpoke_strtobuf
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               str    -> A string to add to the buffer.
|
| Purpose     - Append a string to the pokeout buffer. ANSI colour information 
|               and caller identification are added as needed.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

STATIC void
mpoke_strtobuf(
    int    ident,
    int    colour,
    char * str
    )
{
    unsigned long flags;

    /*
     * Make sure that we are single threaded.
     */
     
    spin_lock_irqsave(&mpoke_lock, flags);
    
    /*
     * If the current caller is different from the previous one then insert
     * the caller's id at the front of the pokeout. If we are using colour
     * then we need to make sure that the identifier comes out in white.
     */
     
    if (ident != mpoke_prev_ident)
    {
        if (mpoke_using_colour && mpoke_prev_colour != MPOKE_WHITE)
        {
            POKEBUF_APPEND(mpoke_ansi_colour_codes[MPOKE_WHITE]);
            mpoke_prev_colour = MPOKE_WHITE;
        }
        
        POKEBUF_APPEND(mpoke_caller_names[ident]);
        mpoke_prev_ident = ident;
    }
    
    /*
     * If we are using colours then insert the ANSI colour string. Don't
     * bother inserting the ANSI string if the previous colour used is the 
     * same as the selected colour.
     */
     
    if (mpoke_using_colour && mpoke_prev_colour != colour)
    {
        POKEBUF_APPEND(mpoke_ansi_colour_codes[colour]);
        mpoke_prev_colour = colour;
    }
    
    /*
     * Now output the caller's string.
     */
     
    POKEBUF_APPEND(str); 

    /*
     * If mpoke_read() is blocked waiting for data then wake it up.
     */
     
    if (atomic_read(&mpoke_read_blocked))
    {
        wake_up_interruptible(&mpoke_read_wait);
    }

    /*
     * Drop our lock.
     */
     
    spin_unlock_irqrestore(&mpoke_lock, flags);
}


/*----------------------------------------------------------------------------
|
| Macro       - VGA_SETVISIBLE
|
| Parameters  - offs -> Offset from VGA_MEMBASE that should be the start
|                       of the visible VGA memory.
|
| Purpose     - Tell the VGA adapter to start displaying text from its memory
|               at base + offs. i.e. the character cell at (0xb8000 + offs) 
|               should be in the top left hand corner of the screen.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

/*
 * This code is a little subtle. Each writew() function is really writing
 * two separate bytes. The first is a VGA_CMD_SETMEMBASE (+1) command
 * code to the VGA_CMDPORT register. The second byte (in the top 8 bits)
 * is part of the offset.
 */
 
#define VGA_SETVISIBLE(offs)                                  \
{                                                             \
    u16 v1;                                                   \
    u16 v2;                                                   \
                                                              \
    v1 = VGA_CMD_SETMEMBASE + ((offs) & 0xff00);              \
    v2 = VGA_CMD_SETMEMBASE + 1 + (((offs) << 8) & 0xff00);   \
                                                              \
    outw(v1, VGA_CMDPORT);                                    \
    outw(v2, VGA_CMDPORT);                                    \
}


/*----------------------------------------------------------------------------
|
| Macro       - VGA_CLEAR
|
| Parameters  - None.
|
| Purpose     - Blank VGA memory.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

#define VGA_CLEAR()                        \
{                                          \
    int i;                                 \
                                           \
    for (i = 0; i < VGA_MEMSIZE; i += 2)   \
    {                                      \
        writew(0x0f20, VGA_MEMBASE + i);   \
    }                                      \
}

         
/*----------------------------------------------------------------------------
|
| Macro       - VGA_WRITE
|
| Parameters  - str  -> A string to write to VGA memory.
|               attr -> Word with VGA attribute in the hi byte and 0 in the
|                       lo byte.
|
| Purpose     - Write a string with the given attribute to VGA memory.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

#define VGA_WRITE(str, attr)                                    \
{                                                               \
    char * ptr;                                                 \
    u16    attrw;                                               \
                                                                \
    attrw = (u16) (attr);                                       \
                                                                \
    for (ptr = (str); *ptr != '\0'; ptr++)                      \
    {                                                           \
        writew(attrw | *ptr, VGA_MEMBASE + mpoke_vga_offs);     \
                                                                \
        mpoke_vga_offs += 2;                                    \
                                                                \
        if (mpoke_vga_offs >= VGA_MEMSIZE)                      \
        {                                                       \
            mpoke_vga_offs = 0;                                 \
        }                                                       \
    }                                                           \
                                                                \
    writew(0x8f01, VGA_MEMBASE + mpoke_vga_offs);               \
}
 

/*----------------------------------------------------------------------------
|
| Function    - mpoke_strtovga
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               str    -> A string to write to the VGA screen.
|
| Purpose     - Write a string to the VGA screen. Caller identification is
|               added as needed.
|
| Returns     - Nothing.
|
| Notes       - This function changes the visible portion of VGA memory so
|               it will mess up what Linux thinks is visible on the console.
|
|---------------------------------------------------------------------------*/

STATIC void
mpoke_strtovga(
    int    ident,
    int    colour,
    char * str
    )
{
    unsigned long flags;

    /*
     * Make sure that we are single threaded.
     */
     
    spin_lock_irqsave(&mpoke_lock, flags);

    /*
     * Is it time to reset the base of visible VGA memory? We do this
     * every so often so that if the user has changed virtual consoles
     * then pokeouts come back again.
     */
     
    if (jiffies >= mpoke_reset_time)
    {
        /*
         * Reset visible the visible memory base.
         */
         
        VGA_SETVISIBLE(0);
        
        /*
         * If this is the first time we've done this then blank VGA memory.
         */
         
        if (mpoke_reset_time == 0)
        {
            VGA_CLEAR();
        }
        
        /*
         * Arrange to do it again after a while.
         */
          
        mpoke_reset_time = jiffies + VGA_RESET_TIME;
    }
         
    /*
     * If the current caller is different from the previous one then insert
     * the caller's id at the front of the pokeout.
     */
     
    if (ident != mpoke_prev_ident)
    {
        VGA_WRITE(
            mpoke_caller_names[ident], 
            mpoke_vga_colour_attr[MPOKE_WHITE]);
        mpoke_prev_ident = ident;
    }
    
    /*
     * Now output the caller's string.
     */
     
    if (!mpoke_using_colour)
    {
        colour = MPOKE_WHITE;
    }
    
    VGA_WRITE(str, mpoke_vga_colour_attr[colour]); 

    /*
     * Drop our lock.
     */
     
    spin_unlock_irqrestore(&mpoke_lock, flags);
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_strtoconsole
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               str    -> A string to write to the console.
|
| Purpose     - Write a string to the console screen. Caller identification is
|               added as needed.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

/*
 * The console_print() function comes from the linux/kernel/printk.c module.
 * It writes a string to the current console.
 */
 
STATIC void
mpoke_strtoconsole(
    int    ident,
    int    colour,
    char * str
    )
{
    unsigned long flags;

    /*
     * Make sure that we are single threaded.
     */
     
    spin_lock_irqsave(&mpoke_lock, flags);

    /*
     * If the current caller is different from the previous one then insert
     * the caller's id at the front of the pokeout.
     */
     
    if (ident != mpoke_prev_ident)
    {
        console_print(mpoke_caller_names[ident]);
        mpoke_prev_ident = ident;
    }
    
    /*
     * Now output the caller's string.
     */
     
    console_print(str); 

    /*
     * Drop our lock.
     */
     
    spin_unlock_irqrestore(&mpoke_lock, flags);
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_register
|
| Parameters  - name -> Name of the caller registering.
|
| Purpose     - Called to register with the pokeout driver. This function
|               must be called before any other mpoke functions.
|
| Returns     - A caller identifer to be passed into other mpoke functions.
|
|---------------------------------------------------------------------------*/

int
mpoke_register(
    char * name
    )
{
    int           ident;
    int           free_ident;
    int           i;
    unsigned long flags;

    MDBG3(MPOKEF("mpoke_resgister:\n"));

    /*
     * Make sure that we are single threaded.
     */
     
    spin_lock_irqsave(&mpoke_lock, flags);

    /*
     * Find out if this caller is already registered. If so use the
     * same identifier. Also look for a free slot in the list of registered
     * callers in case this caller is not already registered.
     */
     
    ident      = -1;
    free_ident = -1;
    
    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (strcmp(mpoke_caller_names[i], name) == 0)
        {
            ident = i;
            break;
        }
        
        if (mpoke_caller_names[i][0] == '\0')
        {
            free_ident = i;
        }
    }
    
    /*
     * If the caller isn't registered then use the free slot we tried to
     * find.
     */
     
    if (ident == -1)
    {
        ident = free_ident;
    }
          
    /*
     * If the caller isn't already registered we didn't find a free slot then 
     * it's all gone horribly wrong so tell the user.
     */

    if (ident == -1)
    {
        printk("mpoke: **** [%s] TOO MANY CALLERS ****\n", name);
        ident = 0;
    }
         
    /*
     * Update the list of caller names.
     */
     
    sprintf(mpoke_caller_names[ident], "[%s]", name);
              
    /*
     * Drop our lock.
     */
     
    spin_unlock_irqrestore(&mpoke_lock, flags);

    /*
     * Tell the caller it's identifier.
     */
     
    return ident; 
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_deregister
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|
| Purpose     - Called to deregister with the pokeout driver.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void
mpoke_deregister(
    int ident
    )
{
    unsigned long flags;

    MDBG3(MPOKEF("mpoke_deresgister:\n"));

    /*
     * Make sure that we are single threaded.
     */
     
    spin_lock_irqsave(&mpoke_lock, flags);
    
    /*
     * Clear the callers slot in the list of caller names.
     */
     
    mpoke_caller_names[ident][0] = '\0';
              
    /*
     * Drop our lock.
     */
     
    spin_unlock_irqrestore(&mpoke_lock, flags);
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_printf
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               fmt    -> A printf style format string.
|               ...    -> A variable number of arguments.
|
| Purpose     - Do a pokeout in the style of printf.
|
| Returns     - We don't really return anything. mpoke_printf() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_printf(
    int    ident,
    int    colour,
    char * fmt,
    ...
    )
{
    va_list args;
    char    buf[512];

    MDBG3(MPOKEF("mpoke_printf\n"));

    /*
     * Put the formatted data into the pokeout buffer.
     */
     
    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    (*mpoke_outstr)(ident, colour, buf); 

    return 0;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokec
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               c      -> A character to pokeout.
|
| Purpose     - Pokeout a character.
|
| Returns     - We don't really return anything. mpoke_pokec() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokec(
    int ident,
    int colour,
    int c
    )
{
    char buf[2];
    
    MDBG3(MPOKEF("mpoke_pokec\n"));

    /*
     * Put the word into the pokeout buffer.
     */
    
    buf[0] = (char) c;
    buf[1] = '\0';
    
    (*mpoke_outstr)(ident, colour, buf); 

    return 0;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokestr
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               str    -> A string to pokeout.
|
| Purpose     - Pokeout a string.
|
| Returns     - We don't really return anything. mpoke_pokestr() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokestr(
    int    ident,
    int    colour,
    char * str
    )
{
    MDBG3(MPOKEF("mpoke_pokestr\n"));

    /*
     * Put the string into the pokeout buffer.
     */
     
    (*mpoke_outstr)(ident, colour, str); 

    return 0;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokeb
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               b      -> An 8bit byte to pokeout.
|
| Purpose     - Pokeout an 8 bit byte in hex.
|
| Returns     - We don't really return anything. mpoke_pokeb() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokeb(
    int ident,
    int colour,
    int b
    )
{
    char buf[3];
    
    MDBG3(MPOKEF("mpoke_pokeb\n"));

    /*
     * Put the word into the pokeout buffer.
     */
    
    buf[0] = mpoke_hex_chars[(b >>  4) & 0xf];
    buf[1] = mpoke_hex_chars[(b      ) & 0xf];
    buf[2] = '\0';
    
    (*mpoke_outstr)(ident, colour, buf); 

    return 0;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokew
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               w      -> A 16bit word to pokeout.
|
| Purpose     - Pokeout a 16 bit word in hex.
|
| Returns     - We don't really return anything. mpoke_pokew() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokew(
    int      ident,
    int      colour,
    unsigned w
    )
{
    char buf[5];
    
    MDBG3(MPOKEF("mpoke_pokew\n"));

    /*
     * Put the word into the pokeout buffer.
     */
    
    buf[0] = mpoke_hex_chars[(w >> 12) & 0xf];
    buf[1] = mpoke_hex_chars[(w >>  8) & 0xf];
    buf[2] = mpoke_hex_chars[(w >>  4) & 0xf];
    buf[3] = mpoke_hex_chars[(w      ) & 0xf];
    buf[4] = '\0';
    
    (*mpoke_outstr)(ident, colour, buf); 

    return 0;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_pokedw
|
| Parameters  - ident  -> The caller's identification number as returned by
|                         mpoke_register().
|               colour -> A colour for the pokeout.
|               dw     -> A 32bit dword to pokeout.
|
| Purpose     - Pokeout a 32 bit dword in hex.
|
| Returns     - We don't really return anything. mpoke_pokedw() has a return 
|               value so that it can be used in conditional expressions.
|
|---------------------------------------------------------------------------*/

int
mpoke_pokedw(
    int           ident,
    int           colour,
    unsigned long dw
    )
{
    char buf[9];
    
    MDBG3(MPOKEF("mpoke_pokedw\n"));

    /*
     * Put the word into the pokeout buffer.
     */
    
    buf[0] = mpoke_hex_chars[(dw >> 28) & 0xf];
    buf[1] = mpoke_hex_chars[(dw >> 24) & 0xf];
    buf[2] = mpoke_hex_chars[(dw >> 20) & 0xf];
    buf[3] = mpoke_hex_chars[(dw >> 16) & 0xf];
    buf[4] = mpoke_hex_chars[(dw >> 12) & 0xf];
    buf[5] = mpoke_hex_chars[(dw >>  8) & 0xf];
    buf[6] = mpoke_hex_chars[(dw >>  4) & 0xf];
    buf[7] = mpoke_hex_chars[(dw      ) & 0xf];
    buf[8] = '\0';
    
    (*mpoke_outstr)(ident, colour, buf); 

    return 0;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_open
|
| Parameters  - inodep -> Pointer to the inode structure for the filesystem
|                         entry for this device.
|               filep  -> Pointer to a kernel open file structure.
|
| Purpose     - Called by the kernel when our device is opened. Simply 
|               updates the module's reference count.
|
| Returns     - An error code.
|
|---------------------------------------------------------------------------*/

STATIC int
mpoke_open(
    struct inode * inodep,
    struct file  * filep
    )
{
    MDBG3(MPOKEF("open:\n"));

    /*
     * We only support one open.
     */
     
    if (mpoke_open_count > 0)
    {
        printk(DRIVER_NAME ": only one open() supported\n");
        return -EBUSY;
    }
     
    /*
     * Increment our module's reference count.
     */
         
    MOD_INC_USE_COUNT;
    
    return EOK;
}
    

/*----------------------------------------------------------------------------
|
| Function    - mpoke_release
|
| Parameters  - inodep -> Pointer to the inode structure for the filesystem
|                         entry for this device.
|               filep  -> Pointer to a kernel open file structure.
|
| Purpose     - Called by the kernel when our device is closed. Simply
|               decrements the kernels reference count.
|
| Returns     - EOK.
|
|---------------------------------------------------------------------------*/

STATIC int
mpoke_release(
    struct inode * inodep,
    struct file  * filep
    )
{
    MDBG3(MPOKEF("release:\n"));

    /*
     * Decrement our module's reference count.
     */
         
    MOD_DEC_USE_COUNT;
    
    /*
     * Zero our open reference count.
     */
     
    mpoke_open_count = 0;
     
    return EOK;
}


/*----------------------------------------------------------------------------
|
| Function    - mpoke_read
|
| Parameters  - filep  -> Pointer to a kernel open file structure.
|               bufp   -> Pointer to a buffer to hold the data read.
|               count  -> The maximum number of bytes to read.
|               posp   -> A pointer to the file offset.
|
| Purpose     - Called by the kernel when a read() command is issued on
|               our device. Up to count bytes of data are written into
|               bufp.
|
| Returns     - The number of bytes read if there is data. 0 at the end of
|               file or an error code.
|
|---------------------------------------------------------------------------*/

STATIC ssize_t
mpoke_read(
    struct file  * filep,
    char         * bufp,
    size_t         count,
    loff_t       * posp
    )    
{
    int copied;
    int tail;
    
    MDBG3(MPOKEF("read:\n"));

    /*
     * Make a note of the pokeout buffer tail pointer. We are the only
     * entity to change this value so we can safely cache it in a local
     * variable.
     */
     
    tail = atomic_read(&mpoke_pokebuf_tail);
    
    /*
     * If the pokeout buffer is empty then block until some data is added.
     */
     
    while (tail == atomic_read(&mpoke_pokebuf_head))
    {
        /*
         * Tell the code that writes the pokeout buffer that we're
         * sleeping.
         */
         
        atomic_set(&mpoke_read_blocked, TRUE);
        
        /*
         * Sleep until someone writes to the pokeout buffer. There is a race
         * condition where a write comes in between our write to 
         * mpoke_read_blocked and us sleeping. To avoid this we set a timeout
         * on the sleep.
         */

        interruptible_sleep_on_timeout(&mpoke_read_wait, 5 * HZ);

        /*
         * Note that we're no longer sleeping.
         */
         
        atomic_set(&mpoke_read_blocked, FALSE);
        
        /*
         * If we were woken up by a signal then fail.
         */
         
        if (signal_pending(current))
        {
            return -ERESTARTSYS;
        }
    }
    
    /*
     * Copy up to count bytes from the pokeout buffer into bufp.
     */
     
    copied = 0;
    
    while (count > 0 && tail != atomic_read(&mpoke_pokebuf_head))
    {
        /*
         * Put user writes a data item into a user mode buffer. This is
         * not the fastest way of coding this. We should really use
         * copy_to_user() to copy whole extents of the buffer in one go.
         */
         
        put_user(mpoke_pokebuf[tail], bufp);
        
        tail = (tail + 1) & (POKEBUF_SIZE - 1);
        
        count--;
        bufp++;
        copied++;
    }

    /*
     * Set the pokeout buffer tail pointer to our cached value.
     */
     
    atomic_set(&mpoke_pokebuf_tail, tail);
        
    /*
     * Tell the user how much data we copied.
     */
    
    return copied;
}
            

/*----------------------------------------------------------------------------
|
| Function    - mpoke_write
|
| Parameters  - filep  -> Pointer to a kernel open file structure.
|               bufp   -> Pointer to a buffer containing the data to write.
|               count  -> The number of bytes to write.
|               posp   -> A pointer to the file offset.
|
| Purpose     - Called by the kernel when a write() command is issued on
|               our device. Doesn't do anything.
|
| Returns     - The number of bytes written or an error code.
|
|---------------------------------------------------------------------------*/

STATIC ssize_t
mpoke_write(
    struct file  * filep,
    const char   * bufp,
    size_t         count,
    loff_t       * posp
    )    
{
    MDBG3(MPOKEF("write:\n"));
    
    return count;
}


/*----------------------------------------------------------------------------
|
| Function    - init_module
|
| Parameters  - None.
|
| Purpose     - Called by the Linux kernel when the module is loaded.
|
| Returns     - An error code.
|
|---------------------------------------------------------------------------*/

/*
 * These are our installation parameters.
 */
 
static 
int colour = 1;

static 
int console = 0;

static
int rawvga = 0;

/*
 * Module information and parameter declarations.
 */
 
MODULE_AUTHOR("Madge Networks Ltd");
MODULE_DESCRIPTION("Madge Pokeout Driver v" VERSION_NUMBER);
MODULE_PARM(colour,  "i");
MODULE_PARM(console, "i");
MODULE_PARM(rawvga,  "i");

int
init_module(
    void
    )
{
    int ret_code;
    
    MDBG3(MPOKEF("init_module:\n"));

    /*
     * Tell the user that we've loaded.
     */
     
    printk(
        "\n"
        DRIVER_NAME 
        ": Madge Pokeout Driver v" 
        VERSION_NUMBER 
        " (" __TIME__ " " __DATE__ ")\n");
    
    /*
     * Register our entry points with the kernel.
     */
     
    ret_code = register_chrdev(MAJOR_NUMBER, DRIVER_NAME, &mpoke_fops);
    
    if (ret_code != EOK)
    {
        MDBG1(MPOKEF("register_chrdev() failed %d\n", ret_code)); 
        return ret_code;
    }

    /*
     * Note if we're using colour.
     */
     
    mpoke_using_colour = (colour != 0);
    
    /*
     * We may have been told to send pokeouts to the current console.
     */
     
    if (console != 0)
    {
        mpoke_outstr = mpoke_strtoconsole;
        printk("Pokeouts will be written to the current console\n");
    }
    
    /*
     * Or we may have been told to send pokeouts to the raw VGA device.
     */
     
    else if (rawvga != 0)
    {
        mpoke_outstr = mpoke_strtovga;
        printk("Pokeouts will be written to the VGA adapter\n");
        
        /*
         * We need to reset the base of visible VGA memory so we know
         * where to write pokeouts (the Linux console driver does clever
         * bits of re-mapping for virtual consoles and scrolling). Arrange
         * for mpoke_strtovga() to do this the first time it is called.
         */
         
        mpoke_reset_time = 0;
    }
    
    /*
     * The default is to send pokeouts to a cyclic buffer so that they can
     * be read from a device file.
     */
     
    else 
    { 
        mpoke_outstr = mpoke_strtobuf;
        printk(
            "Pokeouts can be read from the device node with major number %d\n", 
            MAJOR_NUMBER);
    }
    
    /*
     * Initialise the spin-lock we use to ensure we're single threaded. The
     * asignment is to stop a compiler warning when the locking macros
     * don't acutally us mpoke_lock.
     */

    mpoke_lock = mpoke_lock;     
    spin_lock_init(&mpoke_lock);

    /*
     * Initalise our atomic counters.
     */
     
    atomic_set(&mpoke_pokebuf_head, 0);
    atomic_set(&mpoke_pokebuf_tail, 0);
    atomic_set(&mpoke_read_blocked, FALSE);

    return EOK;
}


/*----------------------------------------------------------------------------
|
| Function    - cleanup_module
|
| Parameters  - None.
|
| Purpose     - Called by the Linux kernel when the module is unloaded.
|
| Returns     - Nothing.
|
|---------------------------------------------------------------------------*/

void
cleanup_module(
    void
    )
{
    MDBG3(MPOKEF("cleanup_module:\n"));
    
    /*
     * Deregister with the kernel.
     */
     
    unregister_chrdev(MAJOR_NUMBER, DRIVER_NAME);
}
    

/**** End of mpoke.c ******************************************************/
