/********************************************************************
*	THE USE BY YOU OF THIS SOFTWARE IS GOVERNED IN ALL
*	RESPECTS BY THE MADGE NETWORKS LTD OPEN SOURCE LICENSE.		
*	WE ADVISE YOU TO READ THE TERMS OF THE OPEN SOURCE LICENSE
*	WHICH CAN BE FOUND WITHIN THE LICENSE.TXT FILE WHICH IS PART OF
*	THE SOFTWARE OR BY VISITING THE URL
*	http://www.madge.com/Connect/Downloads/Software/.		
*	MADGE NETWORKS LTD RESERVES THE RIGHT TO SEEK REDRESS FOR
*	ANY VIOLATION OF THE LICENSE'S TERMS					
*********************************************************************/

/*****************************************************************************
*
* ver.h 
*
* Madge Token Ring Device Driver for Linux 2.0.x/2.2.x/2.4.x
*
* Version Independent header file.
*
* Copyright (c) Madge Networks Ltd
*
*    
*****************************************************************************/

#ifndef __VER_H
#define __VER_H


/*
 * Try to keep version dependancies in this file rather than the main source
 * file
 */

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
#endif


#ifndef LINUX_KERNEL_VERSION
#define LINUX_KERNEL_VERSION LINUX_VERSION_CODE
#endif

#ifndef LINUX_VERSION_CODE
#define LINUX_VERSION_CODE LINUX_KERNEL_VERSION
#endif


#define LINUX_MIN_VERSION	131106


#if LINUX_VERSION_CODE < LINUX_MIN_VERSION
#error "Linux Kernel is too old. Upgrade to 2.0.34 or later"
#endif

/* Ver >= 2.0.0 &&  Ver < 2.1.0 */
#if (LINUX_VERSION_CODE >= LINUX_MIN_VERSION && LINUX_VERSION_CODE < 0x020100)
#define LINUX20
#else
#define LINUX21
#endif

/*
 * At some point during kernel development (after 1.2) an IRQ 
 * sharing scheme was introduced and a new flag SA_SHIRQ was 
 * defined.
 * Some macros are created here to overcome this incompatibility
 *
 * Note the introduction of the _instance parameter in the calls
 */

#ifdef SA_SHIRQ

#ifndef FREE_IRQ
#define FREE_IRQ(_irq,_instance)	\
	free_irq(_irq,_instance)
#endif
	
#ifndef REQUEST_IRQ
#define REQUEST_IRQ(_irq,_handler,_flags,_name,_instance)	\
	request_irq(_irq,_handler,_flags,_name,_instance)
#endif

#define IRQ_ARGS(_irq,_instance,_pt_regs)	\
	(_irq,_instance,_pt_regs)
	
#define ADAPTERFROMIRQ(_irq,_dev_id) (MTOK_ADAPTER *)(((LINUXDEV *) _dev_id)->priv)

#else

#ifndef FREE_IRQ
#define FREE_IRQ(_irq,_instance)	\
	free_irq(_irq)
#endif
	
#ifndef REQUEST_IRQ
#define REQUEST_IRQ(_irq,_handler,_flags,_name,_instance)	\
	request_irq(_irq,_handler,0,_name)
#endif

#define IRQ_ARGS(_irq,_instance,_pt_regs)	\
	(_irq,_pt_regs)

#define ADAPTERFROMIRQ(_irq,_dev_id) (MTOK_ADAPTER *)(((LINUXDEV *) irq2dev_map[_irq])->priv)

#endif /* SA_SHIRQ */

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,1,16))
#define REGISTER_NETDEV(_dev) register_netdev(_dev)
#define UNREGISTER_NETDEV(_dev) unregister_netdev(_dev)
#else
#define REGISTER_NETDEV(_dev) register_trdev(_dev)
#define UNREGISTER_NETDEV(_dev) unregister_trdev(_dev)
#endif

#if LINUX_VERSION_CODE < 0x20138
#ifndef LINUX20
#define test_and_set_bit(val, addr) set_bit(val, addr)
#endif
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,14)
typedef struct device LINUXDEV;
#else
typedef struct net_device LINUXDEV;
#endif

#ifdef CONFIG_PROC_FS

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,29)

#define PROC_NET_REGISTER(a)	proc_register_dynamic(&proc_net,a);

#else

#define PROC_NET_REGISTER(a)	proc_net_register(a);

#endif

#define PROC_NET_UNREGISTER(a)	proc_net_unregister( (a)->low_ino);

#endif /* CONFIG_PROC_FS */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,18)
#define MTOK_KSYM(sym)  X(##sym),
#else
#define MTOK_KSYM(sym)	EXPORT_SYMBOL(##sym);
#endif


#ifdef LINUX20

#ifndef TR_MAXRIFLEN
#define TR_MAXRIFLEN	(18)
#endif

#endif

#ifdef LINUX21
#define MTOK_DEV_KFREE_SKB(s)	dev_kfree_skb(s)
#endif

#ifdef LINUX20
#define net_device_stats enet_statistics
#endif

#if LINUX_VERSION_CODE < 0x20155
#define OLD_PCI
#endif

/* Kernels prior to 2.1.53 (53d=35h) always passed down 18 bytes of SR info */
#if LINUX_VERSION_CODE < 0x20135
#define FIX_SR
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,13)
#define MTOK_PCI_IO_BASE(pdev) (pdev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK)
#else
#define MTOK_PCI_IO_BASE(pdev) (pdev->resource[0].start & PCI_BASE_ADDRESS_IO_MASK)
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,46)
#define MTOK_SET_TBUSY(dev)		(dev->tbusy = 1)	
#define MTOK_RESET_TBUSY(dev)	(dev->tbusy = 0)	
#define MTOK_SET_START(dev)		(dev->start = 1)	
#define MTOK_RESET_START(dev)	(dev->start = 0)	
#define MTOK_SET_INTR(dev)		(dev->interrupt = 1)	
#define MTOK_RESET_INTR(dev)	(dev->interrupt = 0)	
#define MTOK_FIRST_RESET_TBUSY(dev)	(dev->tbusy = 0)
#define MTOK_TEST_START(dev)	dev->start
#define MTOK_TEST_TBUSY(dev)	dev->tbusy
#define MTOK_MARK_BH()			mark_bh(NET_BH)

#ifdef LINUX20
#define MTOK_DEV_KFREE_SKB(s)	dev_kfree_skb(s, FREE_WRITE)
#define MTOK_DEV_KFREE_SKB_ANY(s)	dev_kfree_skb(s, FREE_WRITE)
#define MTOK_DEV_KFREE_SKB_IRQ(s)	dev_kfree_skb(s, FREE_WRITE)
#endif /* linux 20 */

#ifdef LINUX21
#define MTOK_DEV_KFREE_SKB(s)	dev_kfree_skb(s)
#define MTOK_DEV_KFREE_SKB_ANY(s)	dev_kfree_skb(s)
#define MTOK_DEV_KFREE_SKB_IRQ(s)	dev_kfree_skb(s)
#endif /* linux 21 */

#else  /* else for if linux code < 2.x..... */

#define MTOK_SET_TBUSY(dev)		netif_stop_queue(dev)
#define MTOK_RESET_TBUSY(dev)	netif_wake_queue(dev)
#define MTOK_SET_START(dev)	
#define MTOK_RESET_START(dev)
#define MTOK_SET_INTR(dev)
#define MTOK_RESET_INTR(dev)
#define MTOK_FIRST_RESET_TBUSY(dev)	netif_start_queue(dev)
#define MTOK_TEST_START(dev)	netif_running(dev)
#define MTOK_TEST_TBUSY(dev)	netif_queue_stopped(dev)
#define MTOK_MARK_BH()

#define MTOK_DEV_KFREE_SKB(s)	dev_kfree_skb(s)
#define MTOK_DEV_KFREE_SKB_ANY(s)	dev_kfree_skb_any(s)
#define MTOK_DEV_KFREE_SKB_IRQ(s)	dev_kfree_skb_irq(s)
#endif


#endif /* __VER_H */
