/* lchown system call.
   Copyright (C) 1998 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/*   */
	
#include <sysdep.h>
#include <errno.h>
#include <unistd.h>
#include <sys/syscall.h>

extern int __syscall_chown (const char *__file,
			    uid_t __owner, gid_t __group);
extern int __syscall_lchown (const char *__file,
			     uid_t __owner, gid_t __group);

int
__lchown (const char *file, uid_t owner, gid_t group)
{
   int err;
   static int libc_old_chown = 0 /* -1=old linux, 1=new linux, 0=unknown */;

   if (libc_old_chown >= 0)
     {
       /* New Linux kernel (>= 2.1) or still unknown: try syscall 182 */
       err = __syscall_lchown (file, owner, group);
       if (err != -1 || errno != ENOSYS)
	 {
	   libc_old_chown = 1;
	   return err;
	 }
       libc_old_chown = -1;
     }

   /* Old Linux kernel (<2.1): lchown semantics implemented by chown syscall */
   return __syscall_chown (file, owner, group);
}

#if defined PIC && defined DO_VERSIONING
default_symbol_version (__lchown, lchown, GLIBC_2.1);
#else
weak_alias (__lchown, lchown)
#endif
