/*
 * filename: close_ipsec.c
 *
 * last edit: 27 June 2002
 */

#include <linux/config.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_ether.h>
#include <linux/inetdevice.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/in.h>
#include <linux/kernel.h>
#include <linux/udp.h>
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <net/protocol.h>

#include "dev_behaviour.h"
#include "ipsec_tunnel.h"

int icmp_bypass = 0;
int cug = 1;
int cug_trap = 0;

static int is_dev_attached(struct net_device *dev)
{
	int ret;

	ret = -EFAULT;
	if (dev) {
		struct net_device *p;

		ret = 0;
		for (p = dev_base; p; p = p->next) {
			if (strncmp(p->name, "ipsec", 5))
				continue;

			if (p->priv && ((struct ipsecpriv *) p->priv)->dev == dev) {
				ret = 1;
				break;
			}
		}
	}

	return ret;
}


int closeipsec_update_behav(struct dev_behaviour_t *p, unsigned int value)
{
	int ret;

	ret = -EPERM;
#ifdef CONFIG_IPSEC_OPENDEV

	if (!p) return -EFAULT;

	if (atomic_read(&p->set) != 1 && !is_dev_attached(p->dev)) {
		if (value) {
			p->dev->hard_start_xmit = p->hard_start_xmit;
			ret = update_behav(p, DEV_OPEN, 1);
		} else {
			ret = update_behav(p, DEV_OPEN, 0);
		}
		
		atomic_set(&p->set, 1);
	}
#endif /* CONFIG_IPSEC_OPENDEV */

	return ret;
}


#ifdef CONFIG_IPSEC_NF

int closeipsec_var = 0;

int closeipsec_sysctl_strategy(ctl_table *ctl, int write, struct file * filp,
                               void *buffer, size_t *lenp)
{
	int ret;

	if (!write) {
		closeipsec_var = 0;
		ret = 0;

		if (ctl && ctl->de && ctl->de->parent) {
			struct net_device *dev;

			dev = dev_get_by_name(ctl->de->parent->name);
			if (dev) {
				struct dev_behaviour_t *p;

				ret = -ESRCH;
				for (p = get_first_behav(); p; p = p->next)
					if (p->ifindex == dev->ifindex) {
						closeipsec_var = (p->behaviour & DEV_OPEN) ? 1 : 0;
						break;
					}

				dev_put(dev);
			}
		}

		ret = proc_dointvec(ctl, write, filp, buffer, lenp);
		goto out;
	}

#ifndef CONFIG_IPSEC_OPENDEV
	ret = -EPERM;
#else
	ret = proc_dointvec(ctl, write, filp, buffer, lenp);
	if (!ret) {
		ret = -EFAULT;
		if (ctl && ctl->de && ctl->de->parent) {
			struct dev_behaviour_t *p;

			ret = -ESRCH;
			if (strncmp(ctl->de->parent->name, "all", 3)) {
				struct net_device *dev;

				dev = dev_get_by_name(ctl->de->parent->name);
				if (dev) {
					for (p = get_first_behav(); p; p = p->next) {
						if (p->ifindex == dev->ifindex) {
							ret = closeipsec_update_behav(p, closeipsec_var);
							break;
						}
					}

					goto out;
				}

				printk(KERN_DEBUG
				       "closeipsec_sysctl_strategy: unknown dev %s.\n",
					   ctl->de->parent->name);
				goto out;
			}
				
			for (p = get_first_behav(); p; p = p->next)
				closeipsec_update_behav(p, closeipsec_var);

			ret = 0;
		}
	}
#endif

out:
	return ret;
}


int in_dev_forward(struct in_device *in_dev, struct sk_buff *skb)
{
#ifdef CONFIG_CLOSEIPSEC_IN
	int ret;

	ret = in_dev->cnf.forwarding;

	if (ret) {
		struct dev_behaviour_t *p;

		p = get_behav(in_dev->dev);
		if (p)
			ret = ((p->behaviour & DEV_OPEN) || skb->ipsec_mark) ? 1 : 0;
	}

	return ret;
#else
	return in_dev->cnf.forwarding;
#endif
}

#endif /* CONFIG_IPSEC_NF */


static int close_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct dev_behaviour_t *p;

#ifdef CONFIG_CLOSEIPSEC
#ifndef CONFIG_CLOSEIPSEC_IN
        if (icmp_bypass) {
          if (skb->protocol == htons(ETH_P_IP)) {
            if (skb->nh.iph->protocol == IPPROTO_ICMP) {
                skb->ipsec_mark = 1;
            }
          }
        }
#endif
#endif

	p = get_behav(dev);
	if (p) {
		if (skb->protocol == htons(ETH_P_ARP))
			return p->hard_start_xmit(skb, dev);

		if (skb->ipsec_mark)
			return p->hard_start_xmit(skb, dev);

		if (p->behaviour & DEV_OPEN)
			return p->hard_start_xmit(skb, dev);

		printk("Outgoing packet dropped by CloseIPSec\n");
		p->out_dropped++;
	}

	p = get_default_behav();
	p->out_dropped++;

	kfree_skb(skb);
	return 0;
}


int closeipsec_register_dev(struct dev_behaviour_t *behav)
{
	int err;

	err = -EFAULT;
	if (behav) {
		struct net_device *dev;

		dev = behav->dev;
		if (dev && dev->hard_start_xmit) {
			atomic_set(&behav->set, 0);

			behav->in_dropped = 0;
			behav->out_dropped = 0;
			/* beh->behaviour is already initialized */

			behav->hard_start_xmit = dev->hard_start_xmit;
			dev->hard_start_xmit = close_hard_start_xmit;

			err = 0;
		}
	}

	return err;
}


int closeipsec_unregister_dev(struct dev_behaviour_t *behav)
{
	int err;

	err = -EFAULT;
	if (behav) {
		struct net_device *dev;

		dev = behav->dev;
		if (dev) {
			dev->hard_start_xmit = behav->hard_start_xmit;
			err = 0;
		}
	}

	return err;
}


#ifdef CONFIG_CLOSEIPSEC_IN

extern struct inet_protocol *inet_protocol_base;
extern struct inet_protocol *inet_protos[MAX_INET_PROTOS];
extern int ip_rcv(struct sk_buff *skb, struct net_device *dev,
                  struct packet_type *pt);

static int allow_udp = 0;

static int (*orig_udp_rcv) (struct sk_buff *);
static int (*orig_ip_rcv) (struct sk_buff *, struct net_device *,
                           struct packet_type *);


static int close_ip_rcv(struct sk_buff *skb, struct net_device *dev,
                        struct packet_type *pt)
{
	struct dev_behaviour_t *p;

	if (!skb)
		return -EFAULT;

	/* the packet was already processed by the ipsec function */
	if (skb->ipsec_mark)
		goto accept;

	/* pass packets by protocol */
	if (skb->nh.iph) {
		__u8 prot = skb->nh.iph->protocol;
	
		if (allow_udp && prot == IPPROTO_UDP)
			goto accept;

#ifndef CONFIG_IPSEC_NF
		if (prot == IPPROTO_AH || prot == IPPROTO_ESP)
			goto accept;
#endif
	}

	/*
	 * pass packet by dev behaviour, this is the slowest test and has to be the
	 * last one
	 */
	p = get_behav(dev);
	if (p) {
#ifdef CONFIG_IPSEC_OPENDEV
		if (p->behaviour & DEV_OPEN)
			goto accept;
#endif

#ifdef CONFIG_IPSEC_NF
 		if (p->behaviour & DEV_CRYPT)
			goto accept;
#endif

		/* well, we have no policy to accept this packet, drop it */
		p->in_dropped++;
	}

	/* count this dropped packet also in all-entry */
	p = get_default_behav();
	p->in_dropped++;

	kfree_skb(skb);
	return 0;

accept:
	return orig_ip_rcv(skb, dev, pt);
}


static int close_udp_rcv(struct sk_buff *skb)
{
	if (skb && !skb->ipsec_mark) {
		struct dev_behaviour_t *p;

		/* accept packets to ike-daemon */
		if (skb->h.uh && skb->len > 4) {
			if (skb->h.uh->dest == htons(500)) {
				goto accept;
			}
		}

		if (skb->dev) {
			p = get_behav(skb->dev);
			if (p) {
#ifdef CONFIG_IPSEC_NF
 				if (p->behaviour & DEV_CRYPT)
					goto accept;
#endif

#ifdef CONFIG_IPSEC_OPENDEV
				if (p->behaviour & DEV_OPEN)
					goto accept;
#endif

				/* well, we have no policy to accept this packet, drop it */
				p->in_dropped++;
			} else {
				struct net_device *dev;

				dev = skb->dev;
				if (dev->name) {
					if (!strncmp(dev->name, "lo", 2))
						goto accept;

					printk(KERN_DEBUG "close_udp_rcv: "
					                  "no behaviour found for dev %s.\n",
					       dev->name);
				}
			}
		} else
			printk(KERN_DEBUG "close_udp_rcv: no dev given.\n");

		p = get_default_behav();
		p->in_dropped++;


		/* broken or not an ike packet and not processed by ipsec -> drop it */
		kfree_skb(skb);
		return 0;
	}

accept:
	return orig_udp_rcv(skb);
}

#endif /* CONFIG_CLOSEIPSEC_IN */


int close_ip_proto(struct packet_type *ptype)
{
#ifdef CONFIG_CLOSEIPSEC_IN
	int ret;

	ret = -EFAULT;
	if (ptype) {
		ret = -EINVAL;
		if (ptype->func == ip_rcv) {
			struct inet_protocol *p;

			printk("Close-IPSec: replace receive functions.\n");

			orig_ip_rcv = ptype->func;
			ptype->func = close_ip_rcv;

			p = inet_protos[IPPROTO_UDP & (MAX_INET_PROTOS - 1)]; 
			if (p && p->protocol == IPPROTO_UDP) {
				orig_udp_rcv = p->handler;
				p->handler = close_udp_rcv;

				allow_udp = 1;
			} else
				for (p = inet_protocol_base; p; p = p->next) {
					if (p->protocol == IPPROTO_UDP) {
						orig_udp_rcv = p->handler;
						p->handler = close_udp_rcv;

						allow_udp = 1;
						break;
					}
				}

			ret = 0;
		}
	}

	return ret;
#else
	return 0;
#endif
}


int __init closeipsec_init(void)
{
	struct dev_behaviour_t *def_behav;
	int err;

	printk("Initializing Close-IPSec subsystem\n");

#ifndef CONFIG_CLOSEIPSEC_IN
	// open ICMP Bypass by default (will be changed!!)
        icmp_bypass = 0;
#endif
	// set Closed User Group by default
  	cug = 1;
	
	def_behav = init_dev_behaviour();
	if (!IS_ERR(def_behav)) {
		atomic_set(&def_behav->set, 0);
		def_behav->in_dropped = 0;
		def_behav->out_dropped = 0;
		def_behav->hard_start_xmit = NULL;

#ifdef CONFIG_IPSEC_NF
		ipsec_sysctl_init(def_behav);
#endif

		err = 0;
	} else
		err = PTR_ERR(def_behav);

	return err;
}


__initcall(closeipsec_init);

