/*
 * declarations relevant to encapsulation-like operations
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: ipsec_encap.h,v 1.5 2001/08/27 14:54:14 kai Exp $
 */

#define SENT_IP4	0x0008		/* data is two struct in_addr */

#define SEN_HDRLEN	(2*sizeof(__u8)+sizeof(__u16))

#define SEN_IP4_SRCOFF	(0)
#define SEN_IP4_DSTOFF (sizeof (struct in_addr))
#define SEN_IP4_OPTOFF	(2 * sizeof (struct in_addr))

#define SEN_IP4_LEN	(SENT_HDRLEN + SENT_IP4_OPTOFF)

#ifdef CONFIG_IPSEC_DEBUG
#define DB_ER_PROCFS	0x0001
#define DB_SP_PROCFS	0x0001
#endif /* CONFIG_IPSEC_DEBUG */

struct sockaddr_encap
{
	__u8	sen_len;		/* length */
	__u8	sen_family;		/* AF_ENCAP */
	__u16	sen_type;		/* see SENT_* */
	union
	{
		struct			/* SENT_IP4 */
		{
			struct in_addr Src;
			struct in_addr Dst;
		} Sip4;
	} Sen;

	__u8    sen_proto;		/* upper protocol */
	__u16   Src_port;
	__u16   Dst_port;
	__u16   align1;
};

#define sen_ip_src	Sen.Sip4.Src
#define sen_ip_dst	Sen.Sip4.Dst
#define sen_ip_sport    Src_port
#define sen_ip_dport    Dst_port


#ifndef AF_ENCAP
#define AF_ENCAP 26
#endif /* AF_ENCAP */

/*
 * The "type" is really part of the address as far as the routing
 * system is concerned. By using only one bit in the type field
 * for each type, we sort-of make sure that different types of
 * encapsulation addresses won't be matched against the wrong type.
 */

#ifdef __KERNEL__
/*
 * An entry in the radix tree 
 */

struct rjtentry
{
	struct	radij_node rd_nodes[2];	/* tree glue, and other values */
#define	rd_key(r)	((struct sockaddr_encap *)((r)->rd_nodes->rj_key))
#define	rd_mask(r)	((struct sockaddr_encap *)((r)->rd_nodes->rj_mask))
	short	rd_flags;
	short	rd_count;
};

/*
 * An encapsulation route consists of a pointer to a 
 * radix tree entry and a SAID (a destination_address/SPI/protocol triple).
 */

struct
eroute
{
	struct rjtentry er_rjt;
	struct sa_id er_said;
	uint32_t er_count;
	uint64_t er_lasttime;
	struct sockaddr_encap er_eaddr;
	struct sockaddr_encap er_emask;
};

#define er_dst er_said.dst
#define er_spi er_said.spi
#define er_proto er_said.proto
#ifdef CONFIG_IPSEC_DEBUG
extern int debug_eroute;
extern int debug_spi;
#endif /* CONFIG_IPSEC_DEBUG */

#ifdef NETDEV_23
#define device net_device
#define ipsec_dev_get __dev_get_by_name
#else
#define ipsec_dev_get dev_get
#endif /* NETDEV_23 */
#endif /* __KERNEL__ */

