/*
 * Initialization code, and /proc file system interface code.
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

char ipsec_init_c_version[] = "RCSID $Id: ipsec_init.c,v 1.12 2002/08/07 14:57:09 kai Exp $";

#include <linux/config.h>
#include <linux/version.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/in.h>          /* struct sockaddr_in */
#include <linux/skbuff.h>
#include <freeswan.h>
#ifdef SPINLOCK
#ifdef SPINLOCK_23
#include <linux/spinlock.h> /* *lock* */
#else /* SPINLOCK_23 */
#include <asm/spinlock.h> /* *lock* */
#endif /* SPINLOCK_23 */
#endif /* SPINLOCK */
#ifdef NET_21
#include <asm/uaccess.h>
#include <linux/in6.h>
#endif /* NET_21 */
#include <asm/checksum.h>
#include <net/ip.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif /* CONFIG_PROC_FS */
#ifdef NETLINK_SOCK
#include <linux/netlink.h>
#else
#include <net/netlink.h>
#endif

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_tunnel.h"

#include "version.c"

#include "ipsec_rcv.h"
#include "ipsec_ah.h"
#include "ipsec_esp.h"

#include <pfkeyv2.h>
#include <pfkey.h>

#ifdef CONFIG_IPSECTHREAD
#include "ipsec_threads/hsips.h"
#endif

#include <linux/cryptdev.h>

#ifdef CONFIG_IPSEC_DEBUG
int debug_eroute = 0;
int debug_spi = 0;
#endif /* CONFIG_IPSEC_DEBUG */

#ifdef CONFIG_PROC_FS
#ifndef PROC_FS_2325
static
#endif /* PROC_FS_2325 */
int
ipsec_eroute_get_info(char *buffer, char **start, off_t offset, int length
#ifndef PROC_NO_DUMMY
, int dummy
#endif /* !PROC_NO_DUMMY */
)
{
	struct wsbuf w = {buffer, length, offset, 0, 0, 0, 0};    

#ifdef CONFIG_IPSEC_DEBUG
	if (debug_radij & DB_RJ_DUMPTREES)
	  rj_dumptrees();			/* XXXXXXXXX */
#endif /* CONFIG_IPSEC_DEBUG */

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_eroute_get_info: buffer=0x%p,"
		    " *start=0x%x, offset=%d, length=%d\n",
		    buffer, (u_int)*start, (int)offset, length);

	spin_lock_bh(&eroute_lock);

	rj_walktree(rnh, ipsec_rj_walker_procprint, &w);
/*	rj_walktree(mask_rjhead, ipsec_rj_walker_procprint, &w); */

	spin_unlock_bh(&eroute_lock);

	*start = buffer + (offset - w.begin);	/* Start of wanted data */
	w.len -= (offset - w.begin);			/* Start slop */
	if (w.len > length)
		w.len = length;
	return w.len;
}

#ifndef PROC_FS_2325
static
#endif /* PROC_FS_2325 */
int
ipsec_spi_get_info(char *buffer, char **start, off_t offset, int length
#ifndef  PROC_NO_DUMMY
, int dummy
#endif /* !PROC_NO_DUMMY */
)
{
	int len = 0;
	off_t pos = 0, begin = 0;
	int i;
	struct tdb *tdbp;
	char sa[SATOA_BUF];
	char buf_s[ADDRTOA_BUF], buf_d[ADDRTOA_BUF];
	size_t sa_len;

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_spi_get_info: buffer=0x%p,"
		    "*start=0x%x, offset=%d, length=%d\n",
		    buffer, (u_int)*start, (int)offset, length);
	
	KLIPS_PRINT(debug_netlink, "ipsec_spip_get_info: before write_lock\n");
	// write_lock_bh(&sync_lock);
	
	local_bh_disable();
	while (!write_trylock(&sync_lock)) {
		local_bh_enable();
		sleep_on(&sync_lock_wq);
		local_bh_disable();
	}

	KLIPS_PRINT(debug_netlink, "ipsec_spip_get_info: after write_lock\n");
	
	for (i = 0; i < TDB_HASHMOD; i++) {
		for (tdbp = tdbh[i]; tdbp; tdbp = tdbp->tdb_hnext) {
			sa_len = satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
			len += sprintf(buffer + len, "%s ", sa);
			len += sprintf(buffer + len, "%s%s%s", TDB_XFORM_NAME(tdbp));
			len += sprintf(buffer + len, ": dir=%s",
				       (tdbp->tdb_flags & EMT_INBOUND) ?
				       "in " : "out");
			if(tdbp->tdb_iv_bits) {
				int j;
				len += sprintf(buffer + len, " iv_bits=%dbits iv=0x",
					       tdbp->tdb_iv_bits);
				for(j = 0; j < tdbp->tdb_iv_bits / 8; j++) {
					len += sprintf(buffer + len, "%02x",
						       (__u32)((__u8*)(tdbp->tdb_iv))[j]);
				}
			}
			if(tdbp->tdb_encalg || tdbp->tdb_authalg) {
				if(tdbp->tdb_replaywin) {
					len += sprintf(buffer + len, " ooowin=%d",
						       tdbp->tdb_replaywin);
				}
				if(tdbp->tdb_replaywin_errs) {
					len += sprintf(buffer + len, " ooo_errs=%d",
						       tdbp->tdb_replaywin_errs);
				}
				if(tdbp->tdb_replaywin_lastseq) {
                                       len += sprintf(buffer + len, " seq=%d",
						      tdbp->tdb_replaywin_lastseq);
				}
				if(tdbp->tdb_replaywin_bitmap) {
					len += sprintf(buffer + len, " bit=0x%Lx",
						       tdbp->tdb_replaywin_bitmap);
				}
				if(tdbp->tdb_replaywin_maxdiff) {
					len += sprintf(buffer + len, " max_seq_diff=%d",
						       tdbp->tdb_replaywin_maxdiff);
				}
			}
			if(tdbp->tdb_flags & ~EMT_INBOUND) {
				len += sprintf(buffer + len, " flags=0x%x",
					       tdbp->tdb_flags & ~EMT_INBOUND);
				len += sprintf(buffer + len, "<");
				/* flag printing goes here */
				len += sprintf(buffer + len, ">");
			}
			if(tdbp->tdb_auth_bits) {
				len += sprintf(buffer + len, " alen=%d",
					       tdbp->tdb_auth_bits);
			}
			if(tdbp->tdb_key_bits_a) {
				len += sprintf(buffer + len, " aklen=%d",
					       tdbp->tdb_key_bits_a);
			}
			if(tdbp->tdb_auth_errs) {
				len += sprintf(buffer + len, " auth_errs=%d",
					       tdbp->tdb_auth_errs);
			}
			if(tdbp->tdb_key_bits_e) {
				len += sprintf(buffer + len, " eklen=%d",
					       tdbp->tdb_key_bits_e);
			}
			if(tdbp->tdb_encsize_errs) {
				len += sprintf(buffer + len, " encr_size_errs=%d",
					       tdbp->tdb_encsize_errs);
			}
			if(tdbp->tdb_encpad_errs) {
				len += sprintf(buffer + len, " encr_pad_errs=%d",
					       tdbp->tdb_encpad_errs);
			}
			
			if((tdbp->tdb_said.proto == IPPROTO_IPIP) && tdbp->tdb_addr_s && tdbp->tdb_addr_d) {
				addrtoa(((struct sockaddr_in*)(tdbp->tdb_addr_s))->sin_addr,
					0, buf_s, sizeof(buf_s));
				addrtoa(((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr,
					0, buf_d, sizeof(buf_d));
				len += sprintf(buffer + len, " %s -> %s",
					       buf_s, buf_d);
			}

			len += sprintf(buffer + len, " life(c,s,h)=");
			if(tdbp->tdb_lifetime_allocations_c > 1 || 
			   tdbp->tdb_lifetime_allocations_s ||
			   tdbp->tdb_lifetime_allocations_h) {
				len += sprintf(buffer + len, "alloc(%d,%d,%d)",
					       (int)(jiffies - tdbp->tdb_lifetime_allocations_c),
					       tdbp->tdb_lifetime_allocations_s,
					       tdbp->tdb_lifetime_allocations_h);
			}
			if(tdbp->tdb_lifetime_bytes_c ||
			   tdbp->tdb_lifetime_bytes_s ||
			   tdbp->tdb_lifetime_bytes_h) {
				len += sprintf(buffer + len, "bytes(%Ld,",
					       tdbp->tdb_lifetime_bytes_c);
				len += sprintf(buffer + len, "%Ld,",
					       tdbp->tdb_lifetime_bytes_s);
				len += sprintf(buffer + len, "%Ld)",
					       tdbp->tdb_lifetime_bytes_h);
			}
			if(tdbp->tdb_lifetime_addtime_c ||
			   tdbp->tdb_lifetime_addtime_s ||
			   tdbp->tdb_lifetime_addtime_h) {
				len += sprintf(buffer + len, "add(%Ld,%Ld,%Ld)",
					       jiffies / HZ - tdbp->tdb_lifetime_addtime_c,
					       tdbp->tdb_lifetime_addtime_s ?
						tdbp->tdb_lifetime_addtime_s : 0,
					       tdbp->tdb_lifetime_addtime_h ?
						tdbp->tdb_lifetime_addtime_h : 0);
			}
			if(tdbp->tdb_lifetime_usetime_c ||
			   tdbp->tdb_lifetime_usetime_s ||
			   tdbp->tdb_lifetime_usetime_h) {
				len += sprintf(buffer + len, "use(%Ld,%Ld,%Ld)",
					       tdbp->tdb_lifetime_usetime_c ?
						jiffies / HZ - tdbp->tdb_lifetime_usetime_c : 0,
					       tdbp->tdb_lifetime_usetime_s ? 
						tdbp->tdb_lifetime_usetime_s : 0,
					       tdbp->tdb_lifetime_usetime_h ? 
						tdbp->tdb_lifetime_usetime_h : 0);
			}
			if(tdbp->tdb_lifetime_packets_c ||
			   tdbp->tdb_lifetime_packets_s ||
			   tdbp->tdb_lifetime_packets_h) {
				len += sprintf(buffer + len, "packets(%Ld,%Ld,%Ld)",
					       tdbp->tdb_lifetime_packets_c,
					       tdbp->tdb_lifetime_packets_s,
					       tdbp->tdb_lifetime_packets_h);
			}

			if(tdbp->tdb_lifetime_usetime_c) {
				len += sprintf(buffer + len, " idle=%Ld",
					       jiffies / HZ - tdbp->tdb_lifetime_usetime_l);
			}

			len += sprintf(buffer + len, " use_count=%d del_flag=%d"
					       , atomic_read(&tdbp->use_count)
					       , atomic_read(&tdbp->del_flag));
			len += sprintf(buffer + len, "\n");

			pos = begin + len;
			if(pos < offset) {
				len = 0;
				begin = pos;
			}
			if (pos > offset + length) {
				goto done_spi_i;
			}
		}
	}

 done_spi_i:	
	KLIPS_PRINT(debug_netlink, "ipsec_spip_get_info: before write_unlock\n");
	write_unlock_bh(&sync_lock);
	KLIPS_PRINT(debug_netlink, "ipsec_spip_get_info: after write_unlock\n");

	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;
}

#ifndef PROC_FS_2325
static
#endif /* PROC_FS_2325 */
int
ipsec_spigrp_get_info(char *buffer, char **start, off_t offset, int length
#ifndef PROC_NO_DUMMY
, int dummy
#endif /* !PROC_NO_DUMMY */
)
{
	int len = 0;
	off_t pos = 0, begin = 0;
	int i;
	struct tdb *tdbp, *tdbp2;
	char sa[SATOA_BUF];
	size_t sa_len;

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_spigrp_get_info: buffer=0x%p,"
		    " *start=0x%x, offset=%d, length=%d\n",
		    buffer, (u_int)*start, (int)offset, length);

	KLIPS_PRINT(debug_netlink, "ipsec_spigrp_get_info: before write_lock\n");
	// write_lock_bh(&sync_lock);
	local_bh_disable();
	while (!write_trylock(&sync_lock)) {
		local_bh_enable();
		sleep_on(&sync_lock_wq);
		local_bh_disable();
	}

	KLIPS_PRINT(debug_netlink, "ipsec_spigrp_get_info: after write_lock\n");
	
	for (i = 0; i < TDB_HASHMOD; i++) {
		for (tdbp = tdbh[i]; tdbp; tdbp = tdbp->tdb_hnext)
		{
			if(!tdbp->tdb_inext)
			{
				tdbp2 = tdbp;
				while(tdbp2) {
					sa_len = satoa(tdbp2->tdb_said, 0, sa, SATOA_BUF);
					len += sprintf(buffer + len, "%s ",
						       sa);
					tdbp2 = tdbp2->tdb_onext;
				}
				len += sprintf(buffer + len, "\n");
				pos = begin + len;
				if(pos < offset) {
					len = 0;
					begin = pos;
				}
				if (pos > offset + length) {
					goto done_spigrp_i;
				}
			}
		}
	}

 done_spigrp_i:	
	KLIPS_PRINT(debug_netlink, "ipsec_spigrp_get_info: before write_unlock\n");
	write_unlock_bh(&sync_lock);
	KLIPS_PRINT(debug_netlink, "ipsec_spigrp_get_info: after write_unlock\n");

	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;
}

#ifndef PROC_FS_2325
static
#endif /* PROC_FS_2325 */
int
ipsec_tncfg_get_info(char *buffer, char **start, off_t offset, int length
#ifndef PROC_NO_DUMMY
, int dummy
#endif /* !PROC_NO_DUMMY */
)
{
	int len = 0;
	off_t pos = 0, begin = 0;
	int i;
	char name[9];
	struct device *dev, *privdev;
	struct ipsecpriv *priv;

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_tncfg_get_info: buffer=0x%p,"
		    "*start=0x%x, offset=%d, length=%d\n",
		    buffer, (u_int)*start, (int)offset, length);

	for(i = 0; i < IPSEC_NUM_IF; i++) {
		sprintf(name, "ipsec%d", i);
		dev = ipsec_dev_get(name);
		if(dev) {
			priv = (struct ipsecpriv *)(dev->priv);
			len += sprintf(buffer + len, "%s",
				       dev->name);
			if(priv) {
				privdev = (struct device *)(priv->dev);
				len += sprintf(buffer + len, " -> %s",
					       privdev ? privdev->name : "NULL");
				len += sprintf(buffer + len, " mtu=%d -> %d",
					       /* priv */ dev->mtu, privdev ? privdev->mtu : 0);
			} else {
				KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
					    "klips_debug:ipsec_tncfg_get_info: device '%s' has no private data space!\n",
					    dev->name);
			}
			len += sprintf(buffer + len, "\n");

			pos = begin + len;
			if(pos < offset) {
				len = 0;
				begin = pos;
			}
			else if (pos > offset + length)	{
				break;
			}
		}
	}
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;
}

#ifndef PROC_FS_2325
static
#endif /* PROC_FS_2325 */
int
ipsec_version_get_info(char *buffer, char **start, off_t offset, int length
#ifndef PROC_NO_DUMMY
, int dummy
#endif /* !PROC_NO_DUMMY */
)
{

	int len = 0;
	off_t begin = 0;

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_version_get_info: buffer=0x%p,"
		    "*start=0x%x, offset=%d, length=%d\n",
		    buffer, (u_int)*start, (int)offset, length);

	len += sprintf(buffer + len, "SINA version: %s\n", sinaipsec_version);

#ifdef CONFIG_CLOSEIPSEC_IN
	len += sprintf(buffer + len, "IPSec Bypass Disabled\n");
#else
	len += sprintf(buffer + len, "IPSec Bypass Enabled\n");
#endif /* CONFIG_CLOSEIPSEC_IN */

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_init: ipsec_init version: %s\n",
		    ipsec_init_c_version);
	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_init: ipsec_tunnel version: %s\n",
		    ipsec_tunnel_c_version);

	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;

}

#ifdef CONFIG_IPSEC_DEBUG
#ifndef PROC_FS_2325
static
#endif /* PROC_FS_2325 */
int
ipsec_klipsdebug_get_info(char *buffer, char **start, off_t offset, int length
#ifndef PROC_NO_DUMMY
, int dummy
#endif /* !PROC_NO_DUMMY */
)
{

	int len = 0;
	off_t begin = 0;

	KLIPS_PRINT(debug_tunnel & DB_TN_PROCFS,
		    "klips_debug:ipsec_klipsdebug_get_info: buffer=0x%p,"
		    "*start=0x%x, offset=%d, length=%d\n",
		    buffer, (u_int)*start, (int)offset, length);

	len += sprintf(buffer + len, "debug_tunnel=%08x.\n", debug_tunnel);
	len += sprintf(buffer + len, "debug_netlink=%08x.\n", debug_netlink);
	len += sprintf(buffer + len, "debug_xform=%08x.\n", debug_xform);
	len += sprintf(buffer + len, "debug_eroute=%08x.\n", debug_eroute);
	len += sprintf(buffer + len, "debug_spi=%08x.\n", debug_spi);
	len += sprintf(buffer + len, "debug_radij=%08x.\n", debug_radij);
	len += sprintf(buffer + len, "debug_esp=%08x.\n", debug_esp);
	len += sprintf(buffer + len, "debug_ah=%08x.\n", debug_ah);
	len += sprintf(buffer + len, "debug_rcv=%08x.\n", debug_rcv);
	len += sprintf(buffer + len, "debug_pfkey=%08x.\n", debug_pfkey);

	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);			/* Start slop */
	if (len > length)
		len = length;
	return len;

}
#endif /* CONFIG_IPSEC_DEBUG */

#ifndef PROC_FS_2325
struct proc_dir_entry ipsec_eroute =
{
	0,
	12, "ipsec_eroute",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_eroute_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_spi =
{
	0,
	9, "ipsec_spi",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_spi_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_spigrp =
{
	0,
	12, "ipsec_spigrp",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_spigrp_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_tncfg =
{
	0,
	11, "ipsec_tncfg",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_tncfg_get_info,
	NULL, NULL, NULL, NULL, NULL
};

struct proc_dir_entry ipsec_version =
{
	0,
	13, "ipsec_version",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_version_get_info,
	NULL, NULL, NULL, NULL, NULL
};

#ifdef CONFIG_IPSEC_DEBUG
struct proc_dir_entry ipsec_klipsdebug =
{
	0,
	16, "ipsec_klipsdebug",
	S_IFREG | S_IRUGO, 1, 0, 0, 0,
	&proc_net_inode_operations,
	ipsec_klipsdebug_get_info,
	NULL, NULL, NULL, NULL, NULL
};
#endif /* CONFIG_IPSEC_DEBUG */
#endif /* !PROC_FS_2325 */
#endif /* CONFIG_PROC_FS */

int ipsec_device_event(struct notifier_block *dnot, unsigned long event, void *ptr);
/*
 * the following structure is required so that we receive
 * event notifications when network devices are enabled and
 * disabled (ifconfig up and down).
 */
static struct notifier_block ipsec_dev_notifier={
	ipsec_device_event,
	NULL,
	0
};

#ifndef CONFIG_CLOSEIPSEC_IN

extern int icmp_bypass;

static int proc_read_icmpbypass(char *page, char **start,
                            off_t off, int count,
                            int *eof, void *data)
{
        int len;
        len = sprintf(page, "%d\n",icmp_bypass);
        return len;
}


static int proc_write_icmpbypass(struct file *file,
                             const char *buffer,
                             unsigned long count,
                             void *data)
{
        int len;
        len = count;
        if (len > 0) {
          if (buffer[0] == '1') {
            icmp_bypass = 1;
          }
          if (buffer[0] == '0') {
            icmp_bypass = 0;
          }
        }
        return len;
}

#endif // !CLOSIPSEC_IN

extern int cug;
extern int cug_trap;

static int proc_read_cug(char *page, char **start,
                            off_t off, int count,
                            int *eof, void *data)
{
        int len;
        len = sprintf(page, "%d\n",cug);
        return len;
}


static int proc_write_cug(struct file *file,
                             const char *buffer,
                             unsigned long count,
                             void *data)
{
        int len;
        len = count;
        if (cug_trap) {
        	printk("CUG already set\n"); 
        	return len;
        }  
        
        if (len > 0) {
       		if (buffer[0] == '1') {
            		cug = 1;
            		cug_trap = 1;
	       	}
          
        	if (buffer[0] == '0') {
            		cug = 0;
            		cug_trap = 1;
          	}
        }
        return len;
}


/* void */
int
ipsec_init(void)
{
        struct proc_dir_entry *ipsec_icmpbypass;
        struct proc_dir_entry *ipsec_cug;
	int error = 0;

#ifdef CONFIG_IPSECTHREAD
       error = hsips_init();
       if (error)
               return error;
#endif

#ifndef CONFIG_CLOSEIPSEC_IN
       // create proc entry icmp_bypass
       ipsec_icmpbypass = create_proc_entry("ipsec_icmpbypass", 0644, proc_net);
       ipsec_icmpbypass->nlink      = 1;
       ipsec_icmpbypass->read_proc  = proc_read_icmpbypass;
       ipsec_icmpbypass->write_proc = proc_write_icmpbypass;
#endif //!CLOSEIPSEC_IN

       // create proc entry cug
       ipsec_cug = create_proc_entry("ipsec_cug", 0644, proc_net);
       ipsec_cug->nlink      = 1;
       ipsec_cug->read_proc  = proc_read_cug;
       ipsec_cug->write_proc = proc_write_cug;

#ifdef CONFIG_PROC_FS
#  ifndef PROC_FS_2325
#    ifdef PROC_FS_21
	proc_register(proc_net, &ipsec_eroute);
	proc_register(proc_net, &ipsec_spi);
	proc_register(proc_net, &ipsec_spigrp);
	proc_register(proc_net, &ipsec_tncfg);
	proc_register(proc_net, &ipsec_version);
#      ifdef CONFIG_IPSEC_DEBUG
	proc_register(proc_net, &ipsec_klipsdebug);
#      endif /* CONFIG_IPSEC_DEBUG */
#    else /* PROC_FS_21 */
	proc_register_dynamic(&proc_net, &ipsec_eroute);
	proc_register_dynamic(&proc_net, &ipsec_spi);
	proc_register_dynamic(&proc_net, &ipsec_spigrp);
	proc_register_dynamic(&proc_net, &ipsec_tncfg);
	proc_register_dynamic(&proc_net, &ipsec_version);
#      ifdef CONFIG_IPSEC_DEBUG
	proc_register_dynamic(&proc_net, &ipsec_klipsdebug);
#      endif /* CONFIG_IPSEC_DEBUG */
#    endif /* PROC_FS_21 */
#  else /* !PROC_FS_2325 */
	proc_net_create ("ipsec_eroute", 0, ipsec_eroute_get_info);
	proc_net_create ("ipsec_spi", 0, ipsec_spi_get_info);
	proc_net_create ("ipsec_spigrp", 0, ipsec_spigrp_get_info);
	proc_net_create ("ipsec_tncfg", 0, ipsec_tncfg_get_info);
	proc_net_create ("ipsec_version", 0, ipsec_version_get_info);
#    ifdef CONFIG_IPSEC_DEBUG
	proc_net_create ("ipsec_klipsdebug", 0, ipsec_klipsdebug_get_info);
#    endif /* CONFIG_IPSEC_DEBUG */
#  endif /* !PROC_FS_2325 */
#endif          /* CONFIG_PROC_FS */

	printk("klips_debug:ipsec_init: ipsec module loading.\nSINA version %s\n",
	       sinaipsec_version);
	       
#ifdef CONFIG_CLOSEIPSEC_IN	       
	printk("IPSec bypass Disabled\n");
#else
	printk("IPSec bypass Enabled\n");
#endif /* CONFIG_CLOSEIPSEC_IN */	       

#ifndef SPINLOCK
	sync_lock.lock = 0;
	eroute_lock.lock = 0;
#endif /* !SPINLOCK */

	error |= ipsec_tdbinit();
	error |= ipsec_radijinit();

	error |= pfkey_init();

	register_netdevice_notifier(&ipsec_dev_notifier);

#ifdef CONFIG_IPSEC_ESP
	inet_add_protocol(&esp_protocol);
#endif /* CONFIG_IPSEC_ESP */

#ifdef CONFIG_IPSEC_AH
	inet_add_protocol(&ah_protocol);
#endif /* CONFIG_IPSEC_AH */

	error |= ipsec_tunnel_init_devices();

    return error;
}	


/* void */
int
ipsec_cleanup(void)
{
	int error = 0;

	KLIPS_PRINT(debug_netlink, /* debug_tunnel & DB_TN_INIT, */
		    "klips_debug:ipsec_cleanup: calling ipsec_tunnel_cleanup_devices.\n");
	error |= ipsec_tunnel_cleanup_devices();

#ifdef CONFIG_IPSEC_AH
	if ( inet_del_protocol(&ah_protocol) < 0 )
		printk(KERN_INFO "klips_debug:ipsec_cleanup:ah close: can't remove protocol\n");
#endif /* CONFIG_IPSEC_AH */
#ifdef CONFIG_IPSEC_ESP
	if ( inet_del_protocol(&esp_protocol) < 0 )
		printk(KERN_INFO "klips_debug:ipsec_cleanup:esp close: can't remove protocol\n");
#endif /* CONFIG_IPSEC_ESP */

	unregister_netdevice_notifier(&ipsec_dev_notifier);

	KLIPS_PRINT(debug_netlink, /* debug_tunnel & DB_TN_INIT, */
		    "klips_debug:ipsec_cleanup: calling ipsec_tdbcleanup.\n");
	error |= ipsec_tdbcleanup(0);
	KLIPS_PRINT(debug_netlink, /* debug_tunnel & DB_TN_INIT, */
		    "klips_debug:ipsec_cleanup: calling ipsec_radijcleanup.\n");
	error |= ipsec_radijcleanup();
	
	KLIPS_PRINT(debug_pfkey, /* debug_tunnel & DB_TN_INIT, */
		    "klips_debug:ipsec_cleanup: calling pfkey_cleanup.\n");
	error |= pfkey_cleanup();

#ifdef CONFIG_PROC_FS
#  ifndef PROC_FS_2325
#    ifdef CONFIG_IPSEC_DEBUG
	if (proc_net_unregister(ipsec_klipsdebug.low_ino) != 0)
		printk("klips_debug:ipsec_cleanup: cannot unregister /proc/net/ipsec_klipsdebug\n");
#    endif /* CONFIG_IPSEC_DEBUG */
	if (proc_net_unregister(ipsec_version.low_ino) != 0)
		printk("klips_debug:ipsec_cleanup: cannot unregister /proc/net/ipsec_version\n");
	if (proc_net_unregister(ipsec_eroute.low_ino) != 0)
		printk("klips_debug:ipsec_cleanup: cannot unregister /proc/net/ipsec_eroute\n");
	if (proc_net_unregister(ipsec_spi.low_ino) != 0)
		printk("klips_debug:ipsec_cleanup: cannot unregister /proc/net/ipsec_spi\n");
	if (proc_net_unregister(ipsec_spigrp.low_ino) != 0)
		printk("klips_debug:ipsec_cleanup: cannot unregister /proc/net/ipsec_spigrp\n");
	if (proc_net_unregister(ipsec_tncfg.low_ino) != 0)
		printk("klips_debug:ipsec_cleanup: cannot unregister /proc/net/ipsec_tncfg\n");
#  else /* !PROC_FS_2325 */
#    ifdef CONFIG_IPSEC_DEBUG
	proc_net_remove ("ipsec_klipsdebug");
#    endif /* CONFIG_IPSEC_DEBUG */
	proc_net_remove ("ipsec_eroute");
	proc_net_remove ("ipsec_spi");
	proc_net_remove ("ipsec_spigrp");
	proc_net_remove ("ipsec_tncfg");
	proc_net_remove ("ipsec_version");
#  endif /* !PROC_FS_2325 */
#endif          /* CONFIG_PROC_FS */

        remove_proc_entry("icmp_bypass",proc_net);
        remove_proc_entry("cug",proc_net);

#ifdef CONFIG_IPSECTHREAD
       error = hsips_cleanup();
#endif

	return error;
}

#ifdef MODULE
int
init_module(void)
{
	int error = 0;

	error |= ipsec_init();

	return error;
}

int
cleanup_module(void)
{
	int error = 0;

	KLIPS_PRINT(debug_netlink, /* debug_tunnel & DB_TN_INIT, */
		    "klips_debug:cleanup_module: calling ipsec_cleanup.\n");

	error |= ipsec_cleanup();

	KLIPS_PRINT(1, "klips_debug:cleanup_module: ipsec module unloaded.\n");

	return error;
}
#endif /* MODULE */

