/*
 * receive code
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

char ipsec_rcv_c_version[] = "RCSID $Id: ipsec_rcv.c,v 1.27 2002/09/03 16:42:41 kai Exp $";
  
#include <linux/config.h>
#include <linux/version.h>

#define __NO_VERSION__
#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/skbuff.h>
#include <freeswan.h>
#ifdef SPINLOCK
#ifdef SPINLOCK_23
#include <linux/spinlock.h> /* *lock* */
#else /* SPINLOCK_23 */
#include <asm/spinlock.h> /* *lock* */
#endif /* SPINLOCK_23 */
#endif /* SPINLOCK */
#ifdef NET_21
#include <asm/uaccess.h>
#include <linux/in6.h>
#define proto_priv cb
#endif /* NET21 */
#include <asm/checksum.h>
#include <net/ip.h>

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_tunnel.h"
#include "ipsec_rcv.h"
#if defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH)
#include "ipsec_ah.h"
#endif /* defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH) */
#ifdef CONFIG_IPSEC_ESP
#include "ipsec_esp.h"
#endif /* CONFIG_IPSEC_ESP */

#include <pfkeyv2.h>
#include <pfkey.h>

extern int cug;

#ifdef CONFIG_IPSECTHREAD
#  include "ipsec_threads/hsips.h"
#endif

#include <linux/cryptdev.h>

int sysctl_ipsec_inbound_policy_check = 0;

#ifdef CONFIG_DEBUG_CRYPT
extern void dmp(char *s, caddr_t bb, int len);
#endif

#ifdef CONFIG_IPSEC_DEBUG
int debug_ah = 0;
int debug_esp = 0;
int debug_rcv = 0;
#endif /* CONFIG_IPSEC_DEBUG */

#ifdef CONFIG_IPSEC_ESP
extern void des_ede3_cbc_encrypt(caddr_t, caddr_t, int, caddr_t, caddr_t, caddr_t, caddr_t, int);
#endif /* !CONFIG_IPSEC_ESP */
#if defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH)
__u32 zeroes[AH_AMAX];
#endif /* defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH) */

#include "pepp_header.h"

int s_to_i(char *p)
#if LEN_PEPP_LEN == 8
{ int ret=0;
  int i,c=0;
  while(++c<8) {
    switch(p[c]) {
     case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	     i=10+p[c]-'a';
	     break;
     default:
	     i=p[c]-'0';
    }
    ret *=16;
    ret +=i;
  }
  return ret;
}
#elif LEN_PEPP_LEN == 2
{ unsigned char *u=(unsigned char *)p;
  return (u[0]+u[1]*256);
}
#else
#error LEN_PEPP_LEN must be 8 or 2 
#endif


/*
 * Check-replay-window routine, adapted from the original 
 * by J. Hughes, from draft-ietf-ipsec-esp-des-md5-03.txt
 *
 *  This is a routine that implements a 64 packet window. This is intend-
 *  ed on being an implementation sample.
 */

static int
ipsec_checkreplaywindow(struct tdb*tdbp, __u32 seq)
{
	__u32 diff;
	
	if (tdbp->tdb_replaywin == 0)	/* replay shut off */
		return 1;
	if (seq == 0) 
		return 0;		/* first == 0 or wrapped */

	/* new larger sequence number */
	if (seq > tdbp->tdb_replaywin_lastseq) {
		return 1;		/* larger is good */
	}
	diff = tdbp->tdb_replaywin_lastseq - seq;

	/* too old or wrapped */ /* if wrapped, kill off SA? */
	if (diff >= tdbp->tdb_replaywin) {
		return 0;
	}
	/* this packet already seen */
	if (tdbp->tdb_replaywin_bitmap & (1 << diff))
		return 0;
	return 1;			/* out of order but good */
}

static int
ipsec_updatereplaywindow(struct tdb*tdbp, __u32 seq)
{
	__u32 diff;
	
	if (tdbp->tdb_replaywin == 0)	/* replay shut off */
		return 1;
	if (seq == 0) 
		return 0;		/* first == 0 or wrapped */

	/* new larger sequence number */
	if (seq > tdbp->tdb_replaywin_lastseq) {
		diff = seq - tdbp->tdb_replaywin_lastseq;

		/* In win, set bit for this pkt */
		if (diff < tdbp->tdb_replaywin)
			tdbp->tdb_replaywin_bitmap =
				(tdbp->tdb_replaywin_bitmap << diff) | 1;
		else
			/* This packet has way larger seq num */
			tdbp->tdb_replaywin_bitmap = 1;

		if(seq - tdbp->tdb_replaywin_lastseq - 1 > tdbp->tdb_replaywin_maxdiff) {
			tdbp->tdb_replaywin_maxdiff = seq - tdbp->tdb_replaywin_lastseq - 1;
		}
		tdbp->tdb_replaywin_lastseq = seq;
		return 1;		/* larger is good */
	}
	diff = tdbp->tdb_replaywin_lastseq - seq;

	/* this packet already seen */
	if (tdbp->tdb_replaywin_bitmap & (1 << diff))
		return 0;
	tdbp->tdb_replaywin_bitmap |= (1 << diff);	/* mark as seen */
	return 1;			/* out of order but good */
}

int
#ifdef NET_21
ipsec_rcv(struct sk_buff *skb)
#else /* NET_21 */
ipsec_rcv(struct sk_buff *skb, struct device *dev, struct options *opt, 
		__u32 daddr, unsigned short xlen, __u32 saddr,
                                   int redo, struct inet_protocol *protocol)
#endif /* NET_21 */
{
#ifdef NET_21
#ifdef CONFIG_IPSEC_DEBUG
	struct device *dev = skb->dev;
#endif /* CONFIG_IPSEC_DEBUG */
#endif /* NET_21 */
	unsigned char protoc;
	struct iphdr *ipp;
#ifdef CONFIG_IPSEC_ESP
	struct esp *espp = NULL;
	int esphlen = 0, authlen = 0;
	__u32 iv[4];
	unsigned char *ivp;
#endif /* CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_AH
	struct ah *ahp = NULL;
	int ahhlen = 0;
	struct iphdr ipo;
#endif /* CONFIG_IPSEC_AH */
#if defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH)
	CRYPTO_STATUS cs = CRYPTO_OK;
	unsigned char *authenticator = NULL;
#endif /* defined(CONFIG_IPSEC_ESP) || defined(CONFIG_IPSEC_AH) */

	int hard_header_len;
	int iphlen;
	unsigned char *dat;
	struct tdb *tdbp, *tdbs;
	struct sa_id said;
	struct net_device_stats *stats = NULL;		/* This device's statistics */
	struct device *ipsecdev = NULL, *prvdev;
	struct ipsecpriv *prv;
	char name[9];
	char sa[SATOA_BUF];
	size_t sa_len;
	char ipaddr_txt[ADDRTOA_BUF];
	int i;
	int ike_packet = 0;
	struct in_addr ipaddr;
	__u8 next_header = 0;
	__u8 proto;
	
	int pad = 0, ilen, padlen = 0, len, replay = 0;
	__u8 *idat;
	struct tdb* tdbprev = NULL;
	struct tdb* tdbnext = NULL;

	unsigned int hid;

	tdbs = NULL;
	hid = 0;

	/* Don't unlink in the middle of a turnaround */
	MOD_INC_USE_COUNT;
	
	KLIPS_PRINT(debug_netlink, "ipsec_rcv: before read_lock\n");
	read_lock(&sync_lock);
	KLIPS_PRINT(debug_netlink, "ipsec_rcv: after read_lock\n");

	if (skb == NULL) {
		KLIPS_PRINT(debug_rcv, 
			    "klips_debug:ipsec_rcv: "
			    "NULL skb passed in.\n");
		goto rcvleave;
	}
		
#ifdef CONFIG_SMP
	hid = smp_processor_id();
#endif
#ifdef CONFIG_IPSECTHREAD
	memcpy((char *) &hid, skb->cb, sizeof(unsigned int));
#endif

	if (skb->data == NULL) {
		KLIPS_PRINT(debug_rcv,
			    "klips_debug:ipsec_rcv: "
			    "NULL skb->data passed in, packet is bogus, dropping.\n");
		goto rcvleave;
	}
	
	if (skb->nh.iph) { 
	        int ihl = skb->nh.iph->ihl * 4;

		skb->len += ihl;
		skb->data -= ihl;

		skb->nh.raw = skb->data;
	} else {
		KLIPS_PRINT(debug_rcv, "klips_debug:ipsec_rcv: "
		                       "can't see the ip-header.\n");
		goto rcvleave;
	}

	/* dev->hard_header_len is unreliable and should not be used */
	hard_header_len = skb->mac.raw ? (skb->data - skb->mac.raw) : 0;
	if((hard_header_len < 0) || (hard_header_len > skb_headroom(skb)))
		hard_header_len = 0;

#ifdef NET_21
	/* if skb was cloned (most likely due to a packet sniffer such as
	   tcpdump being momentarily attached to the interface), make
	   a copy of our own to modify */
	if(skb_cloned(skb)) {
		/* include any mac header while copying.. */
		if(skb_headroom(skb) < hard_header_len) {
			printk(KERN_WARNING "klips_error:ipsec_rcv: "
			       "tried to skb_push hhlen=%d, %d available.  "
			       "This should never happen, please report.\n",
			       hard_header_len, skb_headroom(skb));
			goto rcvleave;
		}
		skb_push(skb, hard_header_len);

		if ( skb_cow(skb, skb_headroom(skb)) ) {
			goto rcvleave;
		}
		if(skb->len < hard_header_len) {
			printk(KERN_WARNING "klips_error:ipsec_rcv: "
			       "tried to skb_pull hhlen=%d, %d available.  "
			       "This should never happen, please report.\n",
			       hard_header_len, skb->len);
			goto rcvleave;
		}
		skb_pull(skb, hard_header_len);
	}
	
#endif /* NET_21 */
		
	KLIPS_PRINT(debug_rcv, 
		    "klips_debug:ipsec_rcv: Info -- ");
	KLIPS_PRINTMORE(debug_rcv && skb->dev, "sdb->dev=%s ",
		    skb->dev->name ? skb->dev->name : "NULL");
	KLIPS_PRINTMORE(debug_rcv && dev, "dev=%s ",
		    dev->name ? dev->name : "NULL");
	KLIPS_PRINTMORE(debug_rcv, "\n");

	KLIPS_PRINT(debug_rcv && !(skb->dev && dev && (skb->dev == dev)),
		    "klips_debug:ipsec_rcv: Informational -- "
		    "**if this happens, find out why** "
		    "skb->dev:%s is not equal to dev:%s\n",
		    skb->dev ? (skb->dev->name ? skb->dev->name : "NULL") : "NULL",
		    dev ? (dev->name ? dev->name : "NULL") : "NULL");

#ifdef NET_21
	protoc = ((struct iphdr *)skb->data)->protocol;
#else /* NET_21 */
	if(protocol) {
		protoc = protocol->protocol;
	} else {
		KLIPS_PRINT(debug_rcv & DB_RX_TDB,
			    "klips_debug:ipsec_rcv: "
			    "protocol arg is NULL, this is odd, using value in packet.\n");
		protoc = ((struct iphdr *)skb->data)->protocol;
	}
#endif /* NET_21 */

	if( (protoc != IPPROTO_AH) &&
	    (protoc != IPPROTO_ESP) ) {
		KLIPS_PRINT(debug_rcv & DB_RX_TDB,
			    "klips_debug:ipsec_rcv: Why the hell is someone "
			    "passing me a non-ipsec packet? -- dropped.\n");
		goto rcvleave;
	}

	if(skb->dev) {
		for(i = 0; i < IPSEC_NUM_IF; i++) {
			sprintf(name, "ipsec%d", i);
			if(!strcmp(name, skb->dev->name)) {
				prv = (struct ipsecpriv *)(skb->dev->priv);
				if(prv) {
					stats = (struct net_device_stats *) &(prv->mystats);
				}
				ipsecdev = skb->dev;
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: Info -- pkt "
					    "already proc'ed a group of ipsec headers, "
					    "processing next group of ipsec headers.\n");
				break;
			}
			if((ipsecdev = ipsec_dev_get(name)) == NULL) {
				KLIPS_PRINT(debug_rcv,
					    "klips_error:ipsec_rcv: device %s does "
					    "not exist\n", name);
			}
			prv = ipsecdev ? (struct ipsecpriv *)(ipsecdev->priv) : NULL;
			prvdev = prv ? (struct device *)(prv->dev) : NULL;
			
			KLIPS_PRINT(debug_rcv && prvdev, 
				    "klips_debug:ipsec_rcv: physical device"
				    " for device %s is %s\n",
				    name, prvdev->name);
			if(prvdev && skb->dev &&
			   !strcmp(prvdev->name, skb->dev->name)) {
				stats = prv ? ((struct net_device_stats *) &(prv->mystats)) : NULL;
				skb->dev = ipsecdev;
				if(stats) {
					stats->rx_packets++;
				}
				break;
			}
		}
	} else {
		KLIPS_PRINT(debug_rcv, 
			    "klips_debug:ipsec_rcv: device supplied"
			    " with skb is NULL\n");
	}
			
	if(!stats) {
		ipsecdev = NULL;
	}
	KLIPS_PRINT((debug_rcv && !stats),
		    "klips_error:ipsec_rcv: packet received from physical "
		    "I/F (%s) not connected to ipsec I/F.  Cannot record "
		    "stats.\n"
		    "klips_error          May not have SA for decoding.  "
		    "Is IPSEC traffic expected on this I/F?  "
		    "Check routing.\n",
		    skb->dev ? (skb->dev->name ? skb->dev->name : "NULL") : "NULL");

#ifdef CONFIG_IPSEC_DEBUG
	if (debug_rcv) {
		struct timeval tv;
		do_gettimeofday(&tv);
		printk("klips_debug:ipsec_rcv: ts=%02ld.%04ld %s packet received,"
		       " dev = %s\n", 
		       (unsigned long int)tv.tv_sec % 60,
		       (unsigned long int)tv.tv_usec / 100,
		       protoc == IPPROTO_AH ? "AH" : "ESP",
		       ipsecdev ? ipsecdev->name : "NULL");
	}
#endif /* CONFIG_IPSEC_DEBUG */

	/* begin decapsulating loop here */
	do {
#ifdef CONFIG_IPSEC_ESP
		espp = NULL;
		esphlen = 0;
		authlen = 0;
#endif /* !CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_AH
		ahp = NULL;
		ahhlen = 0;
#endif /* CONFIG_IPSEC_AH */

		len = skb->len;
		dat = skb->data;
		ipp = (struct iphdr *)skb->data;
		proto = ipp->protocol;
		ipaddr.s_addr = ipp->saddr;
		addrtoa(ipaddr, 0, ipaddr_txt, sizeof(ipaddr_txt));
		
		KLIPS_PRINT(debug_rcv, "klips_debug:ipsec_rcv: old\n");
		KLIPS_IP_PRINT(debug_rcv, ipp);
		
		iphlen = ipp->ihl << 2;
		ipp->check = 0;			/* we know the sum is good */
		
#ifdef CONFIG_IPSEC_ESP
		if ((proto == IPPROTO_ESP) && ((len - iphlen) % 4)) {
			printk("klips_error:ipsec_rcv: got packet with content length = %d from "
			       "%s -- should be on 4 octet boundary, packet dropped\n",
			       len - iphlen, ipaddr_txt);	/* XXX this will need to be 8 for IPv6 */
			if(stats) {
				stats->rx_errors++;
			}
			goto rcvleave;
		}
#endif /* !CONFIG_IPSEC_ESP */
		
		/*
		 * Find tunnel control block and (indirectly) call the appropriate
		 * tranform routine. The resulting sk_buf is a valid
		 * IP packet ready to go through input processing.
		 */
		
		said.dst.s_addr = ipp->daddr;
		switch(proto) {
#ifdef CONFIG_IPSEC_ESP
		case IPPROTO_ESP:
			espp = (struct esp *)(skb->data + iphlen);
			said.spi = espp->esp_spi;
			break;
#endif /* !CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_AH
		case IPPROTO_AH:
			ahp = (struct ah *)(skb->data + iphlen);
			said.spi = ahp->ah_spi;
			break;
#endif /* CONFIG_IPSEC_AH */
		default:
			if(stats) {
				stats->rx_errors++;
			}
			goto rcvleave;
		}
		said.proto = proto;
		
#ifdef CONFIG_IPSEC_AH
		if(proto == IPPROTO_AH) {
			ahhlen = (ahp->ah_hl << 2) +
				((caddr_t)&(ahp->ah_rpl) - (caddr_t)ahp);
			next_header = ahp->ah_nh;
			if (ahhlen != sizeof(struct ah)) {
				KLIPS_PRINT(debug_rcv & DB_RX_INAU,
					    "klips_debug:ipsec_rcv: bad "
					    "authenticator length %d, expected "
					    "%d from %s\n",
					    ahhlen - ((caddr_t)(ahp->ah_data) -
						      (caddr_t)ahp),
					    AHHMAC_HASHLEN,
					    ipaddr_txt);
				if(stats) {
					stats->rx_errors++;
				}
				goto rcvleave;
			}
			
		}
#endif /* CONFIG_IPSEC_AH */
		
		if (cug) {
			/* check for special SPI 0x1 -> this is IKE traffic */
			if (ntohl(said.spi) == 1) {
			  said.dst.s_addr = 0;
			  ike_packet = 1;
			}
		}

		sa_len = satoa(said, 0, sa, SATOA_BUF);

		/*
		 * The spinlock is to prevent any other process from accessing or deleting
		 * the structure while we are using and updating it.
		 */
		tdbs = tdbp = __find_tdb(&said);

		sa_len = satoa(said, 0, sa, SATOA_BUF);
		if (tdbp == NULL || atomic_read(&tdbp->del_flag)) {		     
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: no Tunnel Descriptor "
				    "Block for SA:%s: incoming packet with no SA "
				    "dropped\n", sa);
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
	
		if (sysctl_ipsec_inbound_policy_check) {
			if(ipp->saddr != ((struct sockaddr_in*)(tdbp->tdb_addr_s))->sin_addr.s_addr) {
				ipaddr.s_addr = ((struct sockaddr_in*)(tdbp->tdb_addr_s))->sin_addr.s_addr;
				addrtoa(ipaddr, 0, ipaddr_txt, sizeof(ipaddr_txt));
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "tdb src addr=%s", ipaddr_txt);
				ipaddr.s_addr = ipp->saddr;
				addrtoa(ipaddr, 0, ipaddr_txt, sizeof(ipaddr_txt));
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s, src=%s of pkt does not agree with expected SA source address policy.\n",
					    sa, ipaddr_txt);
					     
				if(stats) {
					stats->rx_dropped++;
				}
				goto rcvleave;
			}
			{
				ipaddr.s_addr = ipp->saddr;
				addrtoa(ipaddr, 0, ipaddr_txt, sizeof(ipaddr_txt));
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s, src=%s of pkt agrees with expected SA source address policy.\n",
					    sa, ipaddr_txt);
			}
			if(tdbnext) {
				if(tdbnext != tdbp) {
					KLIPS_PRINT(debug_rcv,
						    "klips_debug:ipsec_rcv: "
						    "unexpected SA:%s: does not agree with tdb->inext policy, dropped\n",
						    sa);
					if(stats) {
						stats->rx_dropped++;
					}
					goto rcvleave;
				}
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s grouping from previous SA is OK.\n",
					    sa);
			} else {
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s First SA in group.\n",
					    sa);
			}
			
			if(tdbp->tdb_onext) {
				if(tdbprev != tdbp->tdb_onext) {
					KLIPS_PRINT(debug_rcv,
						    "klips_debug:ipsec_rcv: "
						    "unexpected SA:%s: does not agree with tdb->onext policy, dropped.\n",
						    sa);
					if(stats) {
						stats->rx_dropped++;
					}
					goto rcvleave;
				} else {
					KLIPS_PRINT(debug_rcv,
						    "klips_debug:ipsec_rcv: "
						    "SA:%s grouping to previous SA is OK.\n",
						    sa);
				}
			} else {
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s No previous backlink in group.\n",
					    sa);
			}
		}
		
		/* If it is in larval state, drop the packet, we cannot process yet. */
		if(tdbp->tdb_state == SADB_SASTATE_LARVAL) {
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "TDB in larval state, cannot be used yet, "
				    "dropping packet.\n");
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
		
		if(tdbp->tdb_state == SADB_SASTATE_DEAD) {
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "TDB in dead state, cannot be used any more, "
				    "dropping packet.\n");
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
		
		if(tdbp->tdb_lifetime_bytes_h &&
		   (tdbp->tdb_lifetime_bytes_c > tdbp->tdb_lifetime_bytes_h)) {
			pfkey_expire(tdbp, 1);
			markdel_tdbchain(tdbp);
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "hard bytes lifetime of SA:%s has been reached, "
				    "SA expired, incoming packet dropped.\n", sa);
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
		if(tdbp->tdb_lifetime_bytes_s &&
		   (tdbp->tdb_lifetime_bytes_c > tdbp->tdb_lifetime_bytes_s)) {
		   	if (tdbp->tdb_state != SADB_SASTATE_DYING)
		   	{
			tdbp->tdb_state = SADB_SASTATE_DYING;
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "soft bytes lifetime of SA:%s has been reached, "
				    "SA expiring, soft expire message sent up, "
				    "incoming packet still processed.\n", sa);
			pfkey_expire(tdbp, 0);
			}
		}
		
		if(tdbp->tdb_lifetime_addtime_h &&
		   ((jiffies / HZ) - tdbp->tdb_lifetime_addtime_c >
		    tdbp->tdb_lifetime_addtime_h)) {
			pfkey_expire(tdbp, 1);
			markdel_tdbchain(tdbp);
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "hard addtime lifetime of SA:%s has been reached, "
				    "SA expired, incoming packet dropped.\n", sa);
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
		if(tdbp->tdb_lifetime_addtime_s &&
		   ((jiffies / HZ) - tdbp->tdb_lifetime_addtime_c >
		    tdbp->tdb_lifetime_addtime_s)) {
		       	if (tdbp->tdb_state != SADB_SASTATE_DYING)
		   	{
			tdbp->tdb_state = SADB_SASTATE_DYING;
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "soft addtime lifetime of SA:%s has been reached, "
				    "SA expiring, soft expire message sent up, "
				    "incoming packet still processed.\n", sa);
			pfkey_expire(tdbp, 0);
			}
		}
		
		if(tdbp->tdb_lifetime_usetime_c) {
			if(tdbp->tdb_lifetime_usetime_h &&
			   ((jiffies / HZ) - tdbp->tdb_lifetime_usetime_c >
			    tdbp->tdb_lifetime_usetime_h)) {
				pfkey_expire(tdbp, 1);
				markdel_tdbchain(tdbp);
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "hard usetime lifetime of SA:%s has been reached, "
					    "SA expired, incoming packet dropped.\n", sa);
				if(stats) {
					stats->rx_dropped++;
				}
				goto rcvleave;
			}
			if(tdbp->tdb_lifetime_usetime_s &&
			   ((jiffies / HZ) - tdbp->tdb_lifetime_usetime_c >
			    tdbp->tdb_lifetime_usetime_s)) {
			       	if (tdbp->tdb_state != SADB_SASTATE_DYING)
		   		{
				tdbp->tdb_state = SADB_SASTATE_DYING;
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "soft usetime lifetime of SA:%s has been reached, "
					    "SA expiring, soft expire message sent up, "
					    "incoming packet still processed.\n", sa);
				pfkey_expire(tdbp, 0);
				}
			}
		}
		
		if(tdbp->tdb_lifetime_packets_h &&
		   (tdbp->tdb_lifetime_packets_c > tdbp->tdb_lifetime_packets_h)) {
			pfkey_expire(tdbp, 1);
			markdel_tdbchain(tdbp);
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "hard packets lifetime of SA:%s has been reached, "
				    "SA expired, incoming packet dropped.\n", sa);
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
		if(tdbp->tdb_lifetime_packets_s &&
		   (tdbp->tdb_lifetime_packets_c > tdbp->tdb_lifetime_packets_s)) {
		      	if (tdbp->tdb_state != SADB_SASTATE_DYING)
	   		{
			tdbp->tdb_state = SADB_SASTATE_DYING;
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "soft packets lifetime of SA:%s has been reached, "
				    "SA expiring, soft expire message sent up, "
				    "incoming packet still processed.\n", sa);
			pfkey_expire(tdbp, 0);
			}
		}
		
		/* authenticate, if required */
		idat = dat + iphlen;

		switch(tdbp->tdb_authalg) {
		
		case AH_NONE:
			authlen = 0;
			break;
		
		default:
			authlen = AHHMAC_HASHLEN;
			break;
		}
		
		ilen = len - iphlen - authlen;

		ivp = (__u8 *)espp+sizeof(struct esp);
		
#ifdef CONFIG_IPSEC_ESP
		KLIPS_PRINT(proto == IPPROTO_ESP && debug_rcv, 
			    "klips_debug:ipsec_rcv: packet from %s received with"
			    " seq=%d (iv)=0x%08x%08x... iplen=%d esplen=%d sa=%s\n",
			    ipaddr_txt,
			    (__u32)ntohl(espp->esp_rpl),
			    (__u32)ntohl(*((__u32 *)ivp    )),
			    (__u32)ntohl(*((__u32 *)ivp + 1)),
			    len, ilen, sa);
#endif /* !CONFIG_IPSEC_ESP */
		
		switch(proto) {
#ifdef CONFIG_IPSEC_ESP
		case IPPROTO_ESP:
			replay = ntohl(espp->esp_rpl);
			authenticator = &(dat[len - authlen]);
			break;
#endif /* !CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_AH
		case IPPROTO_AH:
			replay = ntohl(ahp->ah_rpl);
			authenticator = ahp->ah_data;
			break;
#endif /* CONFIG_IPSEC_AH */
		}

		if (!ipsec_checkreplaywindow(tdbp, replay)) {
			KLIPS_PRINT(debug_rcv & DB_RX_REPLAY,
				    "klips_debug:ipsec_rcv: duplicate frame from %s,"
				    " packet dropped\n",
				    ipaddr_txt);
			if(stats) {
				stats->rx_dropped++;
			}
			tdbp->tdb_replaywin_errs += 1;
			goto rcvleave;
		}
		
		/*
		 * verify authenticator in AH (ESP-Auth is done later)
		 */
		
		KLIPS_PRINT(debug_rcv,
			    "klips_debug:ipsec_rcv: encalg = %d, authalg = %d.\n",
			    tdbp->tdb_encalg, tdbp->tdb_authalg);

 	    	if(!authenticator) {
			if(stats) {
				stats->rx_dropped++;
			}
			tdbp->tdb_auth_errs += 1;
			goto rcvleave;
		}		

#ifdef CONFIG_IPSEC_AH		
		if( tdbp->tdb_authalg && (proto == IPPROTO_AH))
		{
			ipo = *ipp;
			ipo.tos = 0;	/* mutable RFC 2402 3.3.3.1.1.1 */
			ipo.frag_off = 0;
			ipo.ttl = 0;
			ipo.check = 0;
					
			{	
			  SData In, Out;

			  In.nKEK = Out.nKEK = 0;
			  In.nDataLength = sizeof(struct iphdr);
			  In.pData = &ipo;

			  cs = crypto_transform(tdbp->tdb_authalg_v_handle[hid], &In, NULL, NULL);
			  if (cs != CRYPTO_OK) {
		      		KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
      		                goto reinit;
       			    }

			  In.nDataLength = ahhlen - AHHMAC_HASHLEN;
			  In.pData = ahp;
			  cs = crypto_transform(tdbp->tdb_authalg_v_handle[hid], &In, NULL, NULL);
			  if (cs != CRYPTO_OK) {
		      		KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
       		                goto reinit;
      			    }

			  In.nDataLength =  AHHMAC_HASHLEN;
			  In.pData = zeroes;
			  cs = crypto_transform(tdbp->tdb_authalg_v_handle[hid], &In, NULL, NULL);
			  if (cs != CRYPTO_OK) {
		      		KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
        	                goto reinit;
        		    }

			  In.nDataLength = len - iphlen - ahhlen;
			  In.pData = dat + iphlen + ahhlen;

			  // output carries the value to be verified 
			  Out.nDataLength = authlen;
			  Out.pData = authenticator;
				  
			  cs = crypto_transform(tdbp->tdb_authalg_v_handle[hid], &In, &Out, NULL);
			  if (cs != CRYPTO_OK) {
		      		KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
       		                goto reinit;
       			    }
reinit:	        			  
       			  // reinitialize AUth
       			  crypto_reopen(tdbp->tdb_authalg_v_handle[hid], NULL);

			}
			
			if (cs == CRYPTO_ERR_VERIFICATION_FAILED)
             		{
	 			KLIPS_PRINT(debug_rcv & DB_RX_INAU,
				    "klips_debug:ipsec_rcv: auth failed on incoming packet from %s, dropped\n", ipaddr_txt);

				printk("klips_debug:ipsec_rcv: auth failed on incoming packet dropped\n");

				if(stats) {
					stats->rx_dropped++;
				}
				tdbp->tdb_auth_errs += 1;
				goto rcvleave;
			}
			else if (cs != CRYPTO_OK) 
			{
		      		KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
       		                goto reinit;
  			}
		 	else 
			{
				KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: authentication successful.\n");
			}
		}
#endif /* CONFIG_IPSEC_AH */		

#ifdef CONFIG_IPSEC_ESP
		if (proto == IPPROTO_ESP)
		{
		    	SData In, In_h, Out_e, Out_a, mi, **ppOut, *aOut[2];
			In.nKEK = In_h.nKEK = \
			Out_e.nKEK = Out_a.nKEK = mi.nKEK = 0;
		    	
			memcpy(iv, ivp, tdbp->tdb_iv_size);

#ifdef CONFIG_DEBUG_CRYPT	
			if (!ike_packet)
				dmp("IV used:", (caddr_t)iv, tdbp->tdb_iv_size);
#endif

			esphlen = sizeof(struct esp) + tdbp->tdb_iv_size;
			idat += esphlen;
			ilen -= esphlen;
			
			if ((ilen) % 8) {
				printk("klips_error:ipsec_rcv: got packet with esplen = %d from "
				       "%s -- should be on 8 octet boundary, packet dropped\n",
				       ilen, ipaddr_txt);
				if(stats) {
					stats->rx_errors++;
				}
				tdbp->tdb_encsize_errs += 1;
				goto rcvleave;
			}

		    	In.nDataLength = ilen;
			In.pData = idat;

#ifdef CONFIG_DEBUG_CRYPT
			if (!ike_packet)
				dmp("Packet to decrypt:", In.pData, In.nDataLength);
#endif
			    
			Out_e.nDataLength = ilen;
			Out_e.pData = kmalloc(ilen, GFP_ATOMIC);
			if (Out_e.pData == NULL)  {
			      printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: no MEM");
			      goto rcvleave;
			}
				    				
			mi.pData = iv;
			mi.nDataLength = tdbp->tdb_iv_size;

			if (tdbp->tdb_authalg)
			{
				In_h.nDataLength = sizeof(struct esp) + tdbp->tdb_iv_size;
				In_h.pData = (caddr_t)espp;				  
				    
				Out_a.nDataLength = authlen;
				Out_a.pData = authenticator;
				
				aOut[0] = &Out_a;
				aOut[1] = &Out_e;
				ppOut = (SData **) &aOut;
				
				if (!tdbp->tdb_i_combined_handle[hid]) {
				      KLIPS_PRINT(debug_tunnel, "need a combined handle here\n");
		              	      kfree(Out_e.pData);
		              	      goto rcvleave;
				}

#ifdef CONFIG_DEBUG_CRYPT
				if (!ike_packet)
					dmp("Received (H)MAC:", authenticator, authlen);
#endif

				if (!tdbp->tdb_is_pepp) {

#ifdef CONFIG_DEBUG_CRYPT
				  if (!ike_packet)
					dmp("Additional Header to (H)MAC:", In_h.pData, In_h.nDataLength);
#endif
				  // first HMAC the ESP HEADER and IV
				  cs = crypto_transform(tdbp->tdb_authalg_v_handle[hid], &In_h, NULL, NULL);
				  if (cs != CRYPTO_OK) {
				        KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
				        cs = crypto_reopen(tdbp->tdb_authalg_v_handle[hid], NULL);
		              	        kfree(Out_e.pData);
		              	        goto rcvleave;
				  }
				}	

				cs = crypto_c_transform(tdbp->tdb_i_combined_handle[hid], &In, (SData **) ppOut, 2, &mi);
				
				// reopen must happen in any case...
				crypto_reopen(tdbp->tdb_authalg_v_handle[hid], NULL);
				
				if (cs == CRYPTO_ERR_VERIFICATION_FAILED)
	             		{
	 				KLIPS_PRINT(debug_rcv & DB_RX_INAU,
					    "klips_debug:ipsec_rcv: auth failed on incoming packet from %s, dropped\n", ipaddr_txt);
					
					printk("klips_debug:ipsec_rcv: auth failed on incoming packet dropped\n");

					if(stats) {
						stats->rx_dropped++;
					}
					tdbp->tdb_auth_errs += 1;
					kfree(Out_e.pData);
					goto rcvleave;
				} 
				else if (cs != CRYPTO_OK) 
				{
				      KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
		              	      kfree(Out_e.pData);
		              	      goto rcvleave;
				}
				else {
					KLIPS_PRINT(debug_rcv,
				    	"klips_debug:ipsec_rcv: authentication successful.\n");
				}

			}
			else  // only decryption transform, no auth
			{	
			    cs = crypto_transform(tdbp->tdb_encalg_d_handle[hid], &In, &Out_e, &mi);
			    if (cs != CRYPTO_OK) {
			      KLIPS_PRINT(debug_tunnel, "error crypto transform #%d\n",  cs);
			      kfree(Out_e.pData);
			      goto rcvleave;
			    }
			}

#ifdef CONFIG_DEBUG_CRYPT
			if (!ike_packet)
				dmp("Decrypted Packet:", Out_e.pData, Out_e.nDataLength);
#endif
			if (tdbp->tdb_is_pepp) {
			  int failed=0;
			  if (Out_e.nDataLength<=PEPP_HEADER_FULL_SIZE)
			    failed=1;
			  else 
 			    if (memcmp(espp,Out_e.pData,PEPP_HEADER_SHORT_SIZE)!=0) {
			 	failed=1;
			    } else {
			      int len=s_to_i((caddr_t)Out_e.pData+PEPP_HEADER_SHORT_SIZE);
			      if (len!=Out_e.nDataLength-LEN_PEPP_LEN) 
			  	failed=1;
			    } 

			  if (failed) {
				KLIPS_PRINT(debug_rcv & DB_RX_INAU,
				    "klips_debug:ipsec_rcv: auth failed on incoming packet from %s, dropped\n", ipaddr_txt);
				
				printk("klips_debug:ipsec_rcv: auth failed on incoming packet dropped\n");

				if(stats) {
					stats->rx_dropped++;
				}
				tdbp->tdb_auth_errs += 1;
				kfree(Out_e.pData);
				goto rcvleave;
			  }
			  ilen=Out_e.nDataLength-PEPP_HEADER_FULL_SIZE;
			  memcpy(idat,(caddr_t)Out_e.pData+PEPP_HEADER_FULL_SIZE,
				ilen);
			  skb_trim(skb,skb->len-PEPP_HEADER_FULL_SIZE);
			} else {
			  memcpy(idat, Out_e.pData, Out_e.nDataLength);
			}
			kfree(Out_e.pData);
			    
			next_header = idat[ilen - 1];
			padlen = idat[ilen - 2];
			pad = padlen + 2 + authlen;
			{
			        int badpad = 0;
				
				KLIPS_PRINT(debug_rcv & DB_RX_IPAD,
					    "klips_debug:ipsec_rcv: "
					    "padlen=%d, contents: 0x<offset>: 0x<value> 0x<value> ...\n",
					    padlen);
				
				for (i = 1; i <= padlen; i++) {
					if((i % 16) == 1) {
						KLIPS_PRINT(debug_rcv & DB_RX_IPAD,
							    "klips_debug:           %02x:",
							    i - 1);
					}
					KLIPS_PRINTMORE(debug_rcv & DB_RX_IPAD,
						    " %02x",
						    idat[ilen - 2 - padlen + i - 1]);
					if(i != idat[ilen - 2 - padlen + i - 1]) {
					        badpad = 1;
					} 
					if((i % 16) == 0) {
						KLIPS_PRINTMORE(debug_rcv & DB_RX_IPAD,
							    "\n");
					}
				}
				if((i % 16) != 1) {
					KLIPS_PRINTMORE(debug_rcv & DB_RX_IPAD,
							"\n");
				}
				if(badpad) {
					KLIPS_PRINT(debug_rcv & DB_RX_IPAD,
						    "klips_debug:ipsec_rcv: "
						    "warning, decrypted packet from %s has bad padding\n",
						    ipaddr_txt);
					KLIPS_PRINT(debug_rcv & DB_RX_IPAD,
						    "klips_debug:ipsec_rcv: "
						    "...may be bad decryption -- not dropped\n");
					tdbp->tdb_encpad_errs += 1;
				}
				
				KLIPS_PRINT(debug_rcv & DB_RX_IPAD,
					    "klips_debug:ipsec_rcv: "
					    "packet decrypted from %s: next_header = %d, padding = %d\n",
					    ipaddr_txt,
					    next_header,
					    pad - 2 - authlen);
			}
#endif /* !CONFIG_IPSEC_ESP */
		}

		/* If the sequence number == 0, expire SA, it had rolled */
		if(tdbp->tdb_replaywin && !replay /* !tdbp->tdb_replaywin_lastseq */) {
			markdel_tdbchain(tdbp);
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "replay window counter rolled, expiring SA.\n");
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}

		if (!ipsec_updatereplaywindow(tdbp, replay)) {
			KLIPS_PRINT(debug_rcv & DB_RX_REPLAY,
				    "klips_debug:ipsec_rcv: duplicate frame from %s,"
				    " packet dropped\n",
				    ipaddr_txt);
			if(stats) {
				stats->rx_dropped++;
			}
			tdbp->tdb_replaywin_errs += 1;
			goto rcvleave;
		}
		
	       
		tdbp->tdb_lifetime_bytes_c += ilen;
		if(!tdbp->tdb_lifetime_usetime_c) {
			tdbp->tdb_lifetime_usetime_c = jiffies / HZ;
		}
		tdbp->tdb_lifetime_usetime_l = jiffies / HZ;
		tdbp->tdb_lifetime_packets_c += 1;
		

		/*
		 *	Discard the original IP header
		 */
		
		KLIPS_PRINT(debug_rcv & DB_RX_PKTRX,
			    "klips_debug:ipsec_rcv: proto = %d, protoc = %d, "
			    "len = %d.\n",
			    proto, protoc, len);

		ipp->protocol = next_header;
		
		switch(proto) {
#ifdef CONFIG_IPSEC_ESP
		case IPPROTO_ESP:
			ipp->tot_len = htons(ntohs(ipp->tot_len) - (esphlen + pad));
			memmove((void *)(skb->data + esphlen),
				(void *)(skb->data), iphlen);
			if(skb->len < esphlen) {
				printk(KERN_WARNING "klips_error:ipsec_rcv: "
				       "tried to skb_pull esphlen=%d, %d available.  "
				       "This should never happen, please report.\n",
				       esphlen, (int)(skb->len));
				goto rcvleave;
			}
			skb_pull(skb, esphlen);
		break;
#endif /* CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_AH
		case IPPROTO_AH:
			ipp->tot_len = htons(ntohs(ipp->tot_len) - ahhlen);
			memmove((void *)(skb->data + ahhlen),
				(void *)(skb->data), iphlen);
			if(skb->len < ahhlen) {
				printk(KERN_WARNING "klips_error:ipsec_rcv: "
				       "tried to skb_pull ahhlen=%d, %d available.  "
				       "This should never happen, please report.\n",
				       ahhlen, (int)(skb->len));
				goto rcvleave;
			}
			skb_pull(skb, ahhlen);
			break;
#endif /* CONFIG_IPSEC_AH */
		}

#ifdef CONFIG_IPSEC_ESP
		if(proto == IPPROTO_ESP) {
			KLIPS_PRINT(debug_rcv & DB_RX_PKTRX,
				    "klips_debug:ipsec_rcv: trimming to %d.\n",
				    len - esphlen - pad);
			if(pad + esphlen <= len) {
			        skb_trim(skb, len - esphlen - pad);
			} else {
			        KLIPS_PRINT(debug_rcv & DB_RX_PKTRX,
				            "klips_debug:ipsec_rcv: bogus packet, size is zero or negative, dropping.\n");
				goto rcvleave;
			}
		}
#endif /* CONFIG_IPSEC_ESP */
		
		/*
		 *	Adjust pointers
		 */
		
		len = skb->len;
		dat = skb->data;
		
#ifdef NET_21
		skb->nh.raw = skb->data;
		skb->h.raw = skb->nh.raw + (skb->nh.iph->ihl << 2);
		
		memset(&(IPCB(skb)->opt), 0, sizeof(struct ip_options));
#else /* NET_21 */
		skb->h.iph=(struct iphdr *)skb->data;
		skb->ip_hdr=(struct iphdr *)skb->data;
		memset(skb->proto_priv, 0, sizeof(struct options));
#endif /* NET_21 */
		
		ipp = (struct iphdr *)dat;
		ipp->check = 0;
		ipp->check = ip_fast_csum((unsigned char *)dat, iphlen >> 2);

		KLIPS_PRINT(debug_rcv & DB_RX_PKTRX,
			    "klips_debug:ipsec_rcv: after <%s%s%s>, SA:%s:\n",
					    TDB_XFORM_NAME(tdbp), sa);
		KLIPS_IP_PRINT(debug_rcv & DB_RX_PKTRX, ipp);
		
		skb->protocol = htons(ETH_P_IP);
		skb->ip_summed = 0;

		tdbprev = tdbp;
		tdbnext = tdbp->tdb_inext;

		if(sysctl_ipsec_inbound_policy_check) {
			if(tdbnext) {
				if(tdbnext->tdb_onext != tdbp) {
					KLIPS_PRINT(debug_rcv,
						    "klips_debug:ipsec_rcv: "
						    "SA:%s, backpolicy does not agree with fwdpolicy.\n",
						    sa);
					if(stats) {
						stats->rx_dropped++;
					}
					goto rcvleave;
				}
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s, backpolicy agrees with fwdpolicy.\n",
					    sa);
				if(!(   (ipp->protocol == IPPROTO_AH  )
					|| (ipp->protocol == IPPROTO_ESP )
					|| (ipp->protocol == IPPROTO_IPIP)
					)) {
					KLIPS_PRINT(debug_rcv,
						    "klips_debug:ipsec_rcv: "
						    "packet with incomplete policy dropped, last successful SA:%s.\n",
						    sa);
					if(stats) {
						stats->rx_dropped++;
					}
					goto rcvleave;
				}
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "SA:%s, Another IPSEC header to process.\n",
					    sa);
			} else {
				KLIPS_PRINT(debug_rcv,
					    "klips_debug:ipsec_rcv: "
					    "No tdb_inext from this SA:%s.\n",
					    sa);
			}
		}

	} while((ipp->protocol == IPPROTO_ESP) ||
		(ipp->protocol == IPPROTO_AH )
#ifdef CONFIG_IPSEC_IPCOMP
		|| (ipp->protocol == IPPROTO_COMP)
#endif /* CONFIG_IPSEC_IPCOMP */
		);
	/* end decapsulation loop here */
	

	/*
	 * XXX this needs to be locked from when it was first looked
	 * up in the decapsulation loop.  Perhaps it is better to put
	 * the IPIP decap inside the loop.
	 */
	if(tdbnext) {
		tdbp = tdbnext;
		sa_len = satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
		if(ipp->protocol != IPPROTO_IPIP) {
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "SA:%s, Hey!  How did this get through?  Dropped.\n",
				    sa);
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
		if(sysctl_ipsec_inbound_policy_check && (tdbnext = tdbp->tdb_inext)) {
			char sa2[SATOA_BUF];
			sa_len = satoa(tdbnext->tdb_said, 0, sa2, SATOA_BUF);
			KLIPS_PRINT(debug_rcv,
				    "klips_debug:ipsec_rcv: "
				    "unexpected SA:%s after IPIP SA:%s\n",
				    sa2, sa);
			if(stats) {
				stats->rx_dropped++;
			}
			goto rcvleave;
		}
	}

	if(ipp->protocol == IPPROTO_IPIP) {
		if(skb->len < iphlen) {
			printk(KERN_WARNING "klips_debug:ipsec_rcv: "
			       "tried to skb_pull iphlen=%d, %d available.  "
			       "This should never happen, please report.\n",
			       iphlen, (int)(skb->len));
			goto rcvleave;
		}
		skb_pull(skb, iphlen);

#ifdef NET_21
		ipp = (struct iphdr *)skb->nh.raw = skb->data;
		skb->h.raw = skb->nh.raw + (skb->nh.iph->ihl << 2);
		
		memset(&(IPCB(skb)->opt), 0, sizeof(struct ip_options));
#else /* NET_21 */
		ipp = skb->ip_hdr = skb->h.iph = (struct iphdr *)skb->data;

		memset(skb->proto_priv, 0, sizeof(struct options));
#endif /* NET_21 */

		skb->protocol = htons(ETH_P_IP);
		skb->ip_summed = 0;
		KLIPS_PRINT(debug_rcv & DB_RX_PKTRX,
			    "klips_debug:ipsec_rcv: IPIP tunnel stripped.\n");
		KLIPS_IP_PRINT(debug_rcv & DB_RX_PKTRX, ipp);
	}

#ifdef NET_21
	if(stats) {
		stats->rx_bytes += skb->len;
	}
	if(skb->dst) {
		dst_release(skb->dst);
		skb->dst = NULL;
	}
	skb->pkt_type = PACKET_HOST;
	if(hard_header_len &&
	   (skb->mac.raw != (skb->data - hard_header_len)) &&
	   (hard_header_len <= skb_headroom(skb))) {
		/* copy back original MAC header */
		memmove(skb->data - hard_header_len, skb->mac.raw, hard_header_len);
		skb->mac.raw = skb->data - hard_header_len;
	}
#endif /* NET_21 */
#ifdef SKB_RESET_NFCT
        nf_conntrack_put(skb->nfct);
        skb->nfct = NULL;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif /* CONFIG_NETFILTER_DEBUG */
#endif /* SKB_RESET_NFCT */

#ifdef CONFIG_CLOSEIPSEC
	skb->ipsec_mark = 1;
#endif

	KLIPS_PRINT(debug_rcv & DB_RX_PKTRX,
		    "klips_debug:ipsec_rcv: netif_rx() called.\n");

	netif_rx(skb);

	MOD_DEC_USE_COUNT;
	KLIPS_PRINT(debug_netlink, "ipsec_rcv: before read_unlock\n");
	read_unlock(&sync_lock);
	wake_up(&sync_lock_wq);
	KLIPS_PRINT(debug_netlink, "ipsec_rcv: after read_unlock\n");

	return(0);
	
 rcvleave:
 	if(skb) {
#ifdef NET_21
                kfree_skb(skb);
#else /* NET_21 */
                kfree_skb(skb, FREE_WRITE);
#endif /* NET_21 */
	}

	MOD_DEC_USE_COUNT;
	KLIPS_PRINT(debug_netlink, "ipsec_rcv: before read_unlock\n");
	read_unlock(&sync_lock);
	wake_up(&sync_lock_wq);
	KLIPS_PRINT(debug_netlink, "ipsec_rcv: after read_unlock\n");

	return(0);
}

struct inet_protocol ah_protocol =
{
#ifdef CONFIG_IPSECTHREAD
        hsips_rcv,
#else
	ipsec_rcv,				/* AH handler */
#endif
	NULL,				/* TUNNEL error control */
	0,				/* next */
	IPPROTO_AH,			/* protocol ID */
	0,				/* copy */
	NULL,				/* data */
	"AH"				/* name */
};

struct inet_protocol esp_protocol = 
{
#ifdef CONFIG_IPSECTHREAD
        hsips_rcv,
#else
	ipsec_rcv,			/* ESP handler          */
#endif
	NULL,				/* TUNNEL error control */
	0,				/* next */
	IPPROTO_ESP,			/* protocol ID */
	0,				/* copy */
	NULL,				/* data */
	"ESP"				/* name */
};

