/*
 * filename: hsips.c
 *
 * written by: Oliver Stecklina <stecklina@secunet.de>
 *
 * last edit: 07.05.2001
 *
 */

#include <asm/errno.h>

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>

#ifdef CONFIG_IPSECTHREAD_DEV
#  include "thread_dev.h"
#endif

#include <linux/proc_fs.h>

#include "hsips.h"

/* respawn process status informations */
struct respawn_status respawn_thread;
extern struct ipst_status *hsips_threads;

int max_threads = CONFIG_IPSECTHREAD_MAX_THREADS;

#define HWPEPP_MODNAME	"hwpepp"

extern struct module *module_list;

struct module *
find_hwpepp_mod(void) 
{
	struct module *mod;

	for (mod = module_list; mod; mod = mod->next) {
		if (mod->flags & MOD_DELETED)
			continue;
		if (!strcmp(mod->name, HWPEPP_MODNAME))
			break;
	}
	
	return mod;
}


int hsips_get_thread_cnt(void)
{
	int cnt = 0;

	if (respawn_thread.num)
		cnt = respawn_thread.num;
	else {
		if ( find_hwpepp_mod() == NULL ) {
			cnt = smp_num_cpus + 1;
			printk( KERN_DEBUG "hwpepp not found.\n" );
		} else {
			cnt = 15 + 1; // PEPP HACK
			printk( KERN_DEBUG "hwpepp found.\n" );
		}
	}

	if (max_threads > 0 && cnt > max_threads)
		cnt = max_threads;

	/* we need at least 2 threads ... even on non-smp! (ike-thread) */
	if (cnt < 2)
		cnt = 2;

	return cnt;
}

int
ipsec_threads_get_info(char *buffer, char **start, off_t offset, int length)
{
        int len = 0;
        off_t begin = 0;

	if (!respawn_thread.task) {
        	len += sprintf(buffer + len, "no status info available!\n");
		goto out;
	}
	
	if (hsips_threads) {
		int i;
		for (i = 0; i < respawn_thread.num; i++) {
			if (hsips_threads[i].task) {
				spin_lock(&hsips_threads[i].lock);
				len += sprintf(buffer + len, 
				"#%d(%d): pid=%d rx=%d tx=%d cnt=%d up=%d dn=%d sig=%d\n"
				, hsips_threads[i].num
				, hsips_threads[i].task->processor
				, hsips_threads[i].pid
				, hsips_threads[i].rx_buf ? 
				    atomic_read(&hsips_threads[i].rx_buf->size) 
				    : 0
				, hsips_threads[i].tx_buf ? 
				    atomic_read(&hsips_threads[i].tx_buf->size)
				    : 0
				, atomic_read(&hsips_threads[i].count)
				, atomic_read(&hsips_threads[i].up)
				, atomic_read(&hsips_threads[i].dn)
				, atomic_read(&hsips_threads[i].sig)
				);
							       
				spin_unlock(&hsips_threads[i].lock);
			}
		}
	}

	len += sprintf(buffer + len, "rx_err=%d rx_drop=%d\n"
			, atomic_read(&respawn_thread.rx_err)
			, atomic_read(&respawn_thread.rx_drop));
	len += sprintf(buffer + len, "tx_err=%d tx_drop=%d\n"
			, atomic_read(&respawn_thread.tx_err)
			, atomic_read(&respawn_thread.tx_drop));
	len += sprintf(buffer + len, "qsize=%d\n", hsips_get_qsize());

out:	
        *start = buffer + (offset - begin);     /* Start of wanted data */
        len -= (offset - begin);                        /* Start slop */
        if (len > length)
		len = length;
        return len;
}

int hsips_init(void)
{
	int retval = -ENOSYS;

	printk(KERN_NOTICE "hsips_init: starting with MAX_THREADS=%d\n"
			, max_threads);

#ifdef CONFIG_IPSECTHREAD_DEV
	retval = register_hsipsthreaddev();
#else

	printk(KERN_DEBUG "hsips_init: starting respawn threads ... \n");

	memset((char *) &respawn_thread, 0, sizeof(struct respawn_status));
	retval = kernel_thread(hsips_respawn, (void *) &respawn_thread, 0);
	if (retval < 0)
		printk(KERN_NOTICE "hsips_init: create respawn thread failed.\n");
	else {
		retval = 0;
		printk(KERN_DEBUG "hsips_init:                         successful \n");
	}

	proc_net_create("ipsec_threads", 0, ipsec_threads_get_info);

#endif

	return retval;
}


int hsips_cleanup(void)
{
	int retval;

#ifndef CONFIG_IPSECTHREAD_DEV
	proc_net_remove("ipsec_threads");
#endif
	
	retval = hsips_destroy(&respawn_thread);

#ifdef CONFIG_IPSECTHREAD_DEV
	retval |= unregister_hsipsthreaddev();
#endif

	return retval;
}

