/*
 * filename: hsips.h
 *
 * written by: Oliver Stecklina <stecklina@secunet.de>
 *
 * last edit: 07.05.2001
 *
 */

#ifndef _SINA_HSIPS_H
#define _SINA_HSIPS_H

#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <asm/semaphore.h>

#define IPST_UNDEF	10
#define IPST_INIT	11
#define IPST_RUN	12


struct ipst_buf {
	struct sk_buff *first;
	struct sk_buff *cur;

	struct timeval  *stamp;
	atomic_t       size;

	spinlock_t     lock;
};


struct ipst_status {
	struct task_struct *task;
	pid_t              pid;
	unsigned int       num;

	atomic_t           state;
	atomic_t           count;

	spinlock_t         lock;
	wait_queue_head_t  wq;
	struct semaphore   sem;
	atomic_t           up;
	atomic_t           dn;
	atomic_t           sig;

	struct ipst_buf    *rx_buf;
	struct ipst_buf    *tx_buf;

	void               *cu;

	int (*init) (void *);
	int (*ctl) (struct ipst_status *, void *);
	int (*destroy) (struct ipst_status *);
};


struct respawn_status {
	struct task_struct *task;
	pid_t              pid;

        unsigned int       num;   /* number of worker threads */

	atomic_t	   rx_drop;
	atomic_t	   tx_drop;
	atomic_t	   rx_err;
	atomic_t	   tx_err;
};


/* hsips.c */
extern int hsips_init(void);
extern int hsips_cleanup(void);
extern int hsips_get_thread_cnt(void);

/* dispatcher.c */
extern int hsips_start_xmit(struct sk_buff *, struct net_device *);
extern int hsips_rcv(struct sk_buff *);

/* respawn.c */
extern int hsips_respawn(void *status_p);
extern int hsips_destroy(struct respawn_status *status);
extern int hsips_get_qsize(void);

#define IPST_LOCK_BUFFERS_IRQSAVE(status, rx_flags, tx_flags) { \
	spin_lock_irqsave(&status->rx_buf->lock, rx_flags); \
	spin_lock_irqsave(&status->tx_buf->lock, tx_flags); \
}

#define IPST_LOCK_BUFFERS(status) { \
	spin_lock_bh(&status->rx_buf->lock); \
	spin_lock_bh(&status->tx_buf->lock); \
}

#define TIME_CHECK(t0, t1) \
	(!t0) ? 1 : ((!t1) ? 0 : ((t0->tv_sec > t1->tv_sec) ? 0 : \
		((t0->tv_sec == t1->tv_sec && t0->tv_usec > t1->tv_usec) ? 0 : 1)))

#endif
