/*
 * filename: thread_dev.c
 *
 * written by: Oliver Stecklina <stecklina@secunet.de>
 *
 * last edit: 13.12.2000
 *
 */

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/devfs_fs_kernel.h>


#include "thread_dev.h"
#include "hsips.h"

static int dev_busy = 0;

extern struct respawn_status respawn_thread;
extern struct ipst_status *hsips_threads;
#if 0
extern int max_threads;
#endif
static int ipst_dev_open(struct inode *inode, struct file *file)
{
	if (dev_busy)
		return -EBUSY;
	
	dev_busy = 1;

	printk(KERN_DEBUG "ipst_dev: pid %d open thread_dev \n", current->pid);

	return 0;
}

static int ipst_dev_release(struct inode *inode, struct file *file)
{
	if (!dev_busy)
		return -ESRCH;

	dev_busy = 0;

	return 0;
}

static int ipst_dev_ioctl(struct inode *inode, struct file *file,
                          unsigned int cmd, unsigned long arg)
{
       	pid_t pid = 0;
	int retval = 0;

	if (!dev_busy)
		return -ESRCH;
	
	switch (cmd) {
		case IPST_START: {
			/* first check if respawn thread is already running ... */
			if ( !respawn_thread.task ) {
			  printk(KERN_DEBUG "hsips_init: switching to respawn threads ...\n");

			  if (arg > 0) 
			    respawn_thread.num = (int) arg;
			  else
			    respawn_thread.num = 1;
#if 0
			  if (max_threads > 0 && 
			      respawn_thread.num > max_threads)
		                respawn_thread.num = max_threads;
#endif	
			  
			  pid = kernel_thread(hsips_respawn, (void *) &respawn_thread, 0);
			  if (pid < 0) {
				printk(KERN_NOTICE "hsips_init: failed.\n");
				retval = pid;
			  }
			}
			else
			  printk(KERN_DEBUG "hsips_init: respawn thread already running ...\n");
			    
			break;
		}
		case IPST_STOP: {
			retval = hsips_destroy(&respawn_thread);
			memset((char *) &respawn_thread, 0, sizeof(struct respawn_status));

			break;
		}
		case IPST_STATUS: {
		        if (!respawn_thread.task) {
			  printk(KERN_NOTICE "ipst_dev: no status info available!\n");
			  break;
		        }

			printk(KERN_NOTICE "ipst_dev: status info for all running threads:\n");
			printk(KERN_NOTICE "ipst_dev: number of threads: %d\n", 
			       respawn_thread.num);

			if (hsips_threads) {
			  int i;

			  for (i = 0; i < respawn_thread.num; i++) {
			    if (hsips_threads[i].task) {

			      spin_lock(&hsips_threads[i].lock);
			      printk(KERN_NOTICE "ipst_dev:  #%d: num=%d pid=%d rx=%d tx=%d cnt=%d up=%d dn=%d sig=%d cpu(%d)\n"
				     , i
				     , hsips_threads[i].num
				     , hsips_threads[i].pid
				     , hsips_threads[i].rx_buf ? atomic_read(&hsips_threads[i].rx_buf->size) : 0
				     , hsips_threads[i].tx_buf ? atomic_read(&hsips_threads[i].tx_buf->size) : 0
				     , atomic_read(&hsips_threads[i].count)
				     , atomic_read(&hsips_threads[i].up)
				     , atomic_read(&hsips_threads[i].dn)
				     , atomic_read(&hsips_threads[i].sig)
				     , hsips_threads[i].task->processor
				     );

			      spin_unlock(&hsips_threads[i].lock);
			    }
			  }
			  printk(KERN_NOTICE "ipst_dev: qsize=%d\n"
					  , hsips_get_qsize()
					  );
			}
			break;
		}
		default: {
			retval = -EINVAL;
			break;
		}
	}
	return retval;
}


static struct file_operations ipst_fops = {
	open:		ipst_dev_open,
	release:	ipst_dev_release,
	ioctl:		ipst_dev_ioctl,
};


int register_hsipsthreaddev(void)
{
	if (devfs_register_chrdev(HSIPS_THREADDEV_NO, "ipsec_thread", &ipst_fops)) {
		printk(KERN_NOTICE "ipsec_thread: unable to get major %d\n",
		       HSIPS_THREADDEV_NO);
		return -EIO;
	}

	return 0;
}


int unregister_hsipsthreaddev(void)
{
	return devfs_unregister_chrdev(HSIPS_THREADDEV_NO, "ipsec_thread");
}


