/*
 * IPSEC Tunneling code. Heavily based on drivers/net/new_tunnel.c
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

char ipsec_tunnel_c_version[] = "RCSID $Id: ipsec_tunnel.c,v 1.45 2003/03/06 16:18:54 br Exp $";
  
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/config.h>	/* for CONFIG_IP_FORWARD */
#include <linux/version.h>

#include <asm/uaccess.h>	/* copy from user */

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/tcp.h>         /* struct tcphdr */
#include <linux/udp.h>         /* struct udphdr */
#include <linux/skbuff.h>
#include <freeswan.h>
#ifdef SPINLOCK
 #ifdef SPINLOCK_23
  #include <linux/spinlock.h> /* *lock* */
 #else /* SPINLOCK_23 */
  #include <asm/spinlock.h> /* *lock* */
 #endif /* SPINLOCK_23 */
#endif /* SPINLOCK */
#ifdef NET_21
 #define MSS_HACK_		/* experimental */
 #include <asm/uaccess.h>
 #include <linux/in6.h>
 #define ip_chk_addr inet_addr_type
 #define IS_MYADDR RTN_LOCAL
 #include <net/dst.h>
 #undef dev_kfree_skb
 #define dev_kfree_skb(a,b) kfree_skb(a)
 #define proto_priv cb
 #define PHYSDEV_TYPE
 #define DEV_QUEUE_XMIT(skb, device, pri) {\
	skb->dev = device; \
	neigh_compat_output(skb); \
	/* skb->dst->output(skb); */ \
 }
 #define ICMP_SEND(skb_in, type, code, info, dev) \
	icmp_send(skb_in, type, code, htonl(info))
 #define IP_SEND(skb, dev) \
	ip_send(skb);
#else /* NET_21 */
 #define DEV_QUEUE_XMIT(skb, device, pri) {\
	dev_queue_xmit(skb, device, pri); \
 }
 #define ICMP_SEND(skb_in, type, code, info, dev) \
	icmp_send(skb_in, type, code, info, dev)
 #define IP_SEND(skb, dev) \
	if(ntohs(iph->tot_len) > physmtu) { \
		ip_fragment(NULL, skb, dev, 0); \
		dev_kfree_skb(skb, FREE_WRITE); \
	} else { \
		dev_queue_xmit(skb, dev, SOPRI_NORMAL); \
	}
#endif /* NET_21 */
#include <asm/checksum.h>
#include <net/icmp.h>		/* icmp_send() */
#include <net/ip.h>
#ifdef NETDEV_23
#include <linux/netfilter_ipv4.h>
#endif

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_tunnel.h"
#include "ipsec_ipe4.h"
#include "ipsec_ah.h"
#include "ipsec_esp.h"

#include <pfkeyv2.h>
#include <pfkey.h>

#include <net/ip.h>
#include <linux/if_arp.h>
#ifdef MSS_HACK
#include <net/tcp.h>		/* TCP options */
#endif	/* MSS_HACK */

#include <linux/cryptdev.h>

#ifdef CONFIG_CLOSEIPSEC
#include "dev_behaviour.h"
#include "close_ipsec.h"
#endif

#ifdef CONFIG_IPSECTHREAD
#  include "ipsec_threads/hsips.h"
#endif

static __u32 zeroes[64];


#include "pepp_header.h"

void i_to_s(unsigned int l,char *c)
#if LEN_PEPP_LEN == 8	
{
  const char hex[]="0123456789abcdef";
  char *p=c+8;
  while (l>0) {
   *(--p)=hex[l%16];
   l/=16;
   if (p==c)
     return;	   
  }
  while (p>c)
  *(--p)='0';
}
#elif LEN_PEPP_LEN == 2
{
  c[0]=l%256;
  c[1]=l/256;
}
#else
#error LEN_PEPP_LEN must be 2 or 8
#endif
	

#ifdef CONFIG_IPSEC_DEBUG
int sysctl_ipsec_debug_verbose = 0;
int debug_tunnel = 0;
#endif /* CONFIG_IPSEC_DEBUG */

#ifdef CONFIG_DEBUG_CRYPT
DEBUG_NO_STATIC void
dmp(char *s, caddr_t bb, int len)
{
	int i;
	unsigned char *b = bb;
  
	{
		printk(KERN_INFO "klips_debug:ipsec_tunnel_:at %s, len=%d:", s, len);
		for (i=0; i < len; i++) {
			if(!(i%16)){
				printk("\nklips_debug:  ");
			}
			printk(" %02x", *b++);
		}
		printk("\n");
	}
}
#endif

#ifndef SKB_COPY_EXPAND
/*
 *	This is mostly skbuff.c:skb_copy().
 */
struct sk_buff *
skb_copy_expand(struct sk_buff *skb, int headroom, int tailroom, int priority)
{
	struct sk_buff *n;
	unsigned long offset;

	/*
	 *	Do sanity checking
	 */
	if((headroom < 0) || (tailroom < 0) || ((headroom+tailroom) < 0)) {
		printk(KERN_WARNING "klips_error:skb_copy_expand: "
		       "Illegal negative head/tailroom %d, %d\n",
		       headroom, tailroom);
		return NULL;
	}
	/*
	 *	Allocate the copy buffer
	 */
	 
#ifndef NET_21
	IS_SKB(skb);
#endif /* !NET_21 */


	n=alloc_skb(skb->end - skb->head + headroom + tailroom, priority);

	KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
		    KERN_INFO "klips_debug:skb_copy_expand: "
		    "head=%p data=%p tail=%p end=%p end-head=%d\n",
		    skb->head,
		    skb->data,
		    skb->tail,
		    skb->end,
		    skb->end - skb->head);

	if(n==NULL)
		return NULL;

	/*
	 *	Shift between the two data areas in bytes
	 */
	 
	/* offset=n->head-skb->head; */ /* moved down a few lines */

	/* Set the data pointer */
	skb_reserve(n,skb->data-skb->head+headroom);
	/* Set the tail pointer and length */
	if(skb_tailroom(n) < skb->len) {
		printk(KERN_WARNING "klips_error:skb_copy_expand: "
		       "tried to skb_put %ld, %d available.  "
		       "This should never happen, please report.\n",
		       (unsigned long int)skb->len, skb_tailroom(n));
		dev_kfree_skb(n, FREE_WRITE);
		return NULL;
	}
	skb_put(n,skb->len);

	offset=n->head + headroom - skb->head;

	/* Copy the bytes */
	memcpy(n->head + headroom, skb->head,skb->end-skb->head);
#ifdef NET_21
	n->csum=skb->csum;
	n->priority=skb->priority;
	n->protocol=skb->protocol;
	n->dst=dst_clone(skb->dst);
	if(skb->nh.raw)
		n->nh.raw=skb->nh.raw+offset;
	n->is_clone=0;
	atomic_set(&n->users, 1);
	n->destructor = NULL;
	n->security=skb->security;
#else /* NET_21 */
	n->link3=NULL;
	n->when=skb->when;
	if(skb->ip_hdr)
	        n->ip_hdr=(struct iphdr *)(((char *)skb->ip_hdr)+offset);
	n->saddr=skb->saddr;
	n->daddr=skb->daddr;
	n->raddr=skb->raddr;
	n->seq=skb->seq;
	n->end_seq=skb->end_seq;
	n->ack_seq=skb->ack_seq;
	n->acked=skb->acked;
	n->free=1;
	n->arp=skb->arp;
	n->tries=0;
	n->lock=0;
	n->users=0;
#endif /* NET_21 */
	n->list=NULL;
	n->sk=NULL;
	n->dev=skb->dev;
	if(skb->h.raw)
		n->h.raw=skb->h.raw+offset;
	if(skb->mac.raw) 
		n->mac.raw=skb->mac.raw+offset;
	memcpy(n->proto_priv, skb->proto_priv, sizeof(skb->proto_priv));
	n->used=skb->used;
	n->pkt_type=skb->pkt_type;
	n->stamp=skb->stamp;
	
#ifndef NET_21
	IS_SKB(n);
#endif /* !NET_21 */
	return n;
}
#endif /* !SKB_COPY_EXPAND */

#ifdef CONFIG_IPSEC_DEBUG
void
ipsec_print_ip(struct iphdr *ip)
{
	char buf[ADDRTOA_BUF];

	printk(KERN_INFO "klips_debug:   IP:");
	printk(" ihl:%d", ip->ihl*4);
	printk(" ver:%d", ip->version);
	printk(" tos:%d", ip->tos);
	printk(" tlen:%d", ntohs(ip->tot_len));
	printk(" id:%d", ip->id);
	printk(" frag_off:%d", ip->frag_off);
	printk(" ttl:%d", ip->ttl);
	printk(" proto:%d", ip->protocol);
	printk(" chk:%d", ip->check);
	addrtoa(*((struct in_addr*)(&ip->saddr)), 0, buf, sizeof(buf));
	printk(" saddr:%s", buf);
	addrtoa(*((struct in_addr*)(&ip->daddr)), 0, buf, sizeof(buf));
	printk(" daddr:%s", buf);
	printk("\n");

#if 0
	{
		__u8 *c;
		int i;
		
		c = ((__u8*)ip) + ip->ihl*4;
		for(i = 0; i < ntohs(ip->tot_len) - ip->ihl*4; i++, c++) {
			if(!(i % 16)) {
				printk(KERN_INFO "klips_debug:   contents:");
			}
			printk("%02x ", *c);
			if(!((i + 1) % 16)) {
				printk("\n");
			}
		}
		if(i % 16) {
			printk("\n");
		}
	}
#endif
}
#endif /* CONFIG_IPSEC_DEBUG */

#ifdef REAL_LOCKING_P
/*
 *	Locking
 */
 
DEBUG_NO_STATIC int
ipsec_tunnel_lock(struct ipsecpriv *prv)
{
	unsigned long flags;
	save_flags(flags);
	cli();
	/*
	 *	Lock in an interrupt may fail
	 */
	if(prv->locked && in_interrupt()) {
		restore_flags(flags);
		return 0;
	}
	while(prv->locked)
		sleep_on(&prv->wait_queue);
	prv->locked=1;
	restore_flags(flags);
	return 1;
}

DEBUG_NO_STATIC void
ipsec_tunnel_unlock(struct ipsecpriv *prv)
{
	prv->locked=0;
	wake_up(&prv->wait_queue);
}
#endif /* REAL_LOCKING_P */

DEBUG_NO_STATIC int
ipsec_tunnel_open(struct device *dev)
{
	struct ipsecpriv *prv = dev->priv;
	
	/*
	 * Can't open until attached.
	 */

	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_open: "
		    "dev = %s, prv->dev = %s\n",
		    dev->name, prv->dev?prv->dev->name:"NONE");

	if (prv->dev == NULL)
		return -ENODEV;
	
	MOD_INC_USE_COUNT;
	return 0;
}

DEBUG_NO_STATIC int
ipsec_tunnel_close(struct device *dev)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

#ifdef MSS_HACK
/*
 * Recaclulate checksum using differences between changed datum, 
 * borrowed from netfilter.
 */
DEBUG_NO_STATIC u_int16_t 
ipsec_fast_csum(u_int32_t oldvalinv, u_int32_t newval, u_int16_t oldcheck)
{
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs),
	oldcheck^0xFFFF));
}

/*
 * Determine effective MSS.
 *
 * Note that we assume that there is always an MSS option for our own
 * SYN segments, which is mentioned in tcp_syn_build_options(), kernel 2.2.x.
 * This could change, and we should probably parse TCP options instead.
 *
 */
DEBUG_NO_STATIC u_int8_t
ipsec_adjust_mss(struct sk_buff *skb, struct tcphdr *tcph, u_int16_t mtu)
{
	u_int16_t oldmss, newmss;
	u_int32_t *mssp;
	struct sock *sk = skb->sk;
	
	newmss = tcp_sync_mss(sk, mtu);
	printk(KERN_INFO "klips: setting mss to %u\n", newmss);
	mssp = (u_int32_t *)tcph + sizeof(struct tcphdr) / sizeof(u_int32_t);
	oldmss = ntohl(*mssp) & 0x0000FFFF;
	*mssp = htonl((TCPOPT_MSS << 24) | (TCPOLEN_MSS << 16) | newmss);
	tcph->check = ipsec_fast_csum(htons(~oldmss), 
	                              htons(newmss), tcph->check);
	return 1;
}
#endif	/* MSS_HACK */
                                                        
#ifdef NETDEV_23
static inline int ipsec_tunnel_xmit2(struct sk_buff *skb)
{
	return ip_send(skb);
}
#endif

/*
 *	This function assumes it is being called from dev_queue_xmit()
 *	and that skb is filled properly by that function.
 */

int
ipsec_tunnel_start_xmit(struct sk_buff *skb, struct device *dev)
{
	struct ipsecpriv *prv;		/* Our device' private space */
	struct sk_buff *oskb = NULL;	/* Original skb pointer */
	struct net_device_stats *stats;	/* This device's statistics */
	struct iphdr  *iph;		/* Our new IP header */
	__u32   newdst;			/* The other SG's IP address */
	__u32	orgdst;			/* Original IP destination address */
	__u32	orgedst;		/* 1st SG's IP address */
	__u32   newsrc;			/* The new source SG's IP address */
	__u32	orgsrc;			/* Original IP source address */
	__u32	innersrc;		/* Innermost IP source address */
	int	iphlen;			/* IP header length */
	int	pyldsz;			/* upper protocol payload size */
	int	headroom;	
	int	tailroom;
	int     max_headroom = 0;	/* The extra header space needed */
	int	max_tailroom = 0;	/* The extra stuffing needed */
	int     ll_headroom;		/* The extra link layer hard_header space needed */
	int     tot_headroom = 0;	/* The total header space needed */
	int	tot_tailroom = 0;	/* The totalstuffing needed */
	__u8	*saved_header = NULL;	/* saved copy of the hard header */
	int i;
	
	struct sockaddr_encap matcher;	/* eroute search key */
	struct eroute *er = NULL;
	unsigned char ike_packet = 0; 
	unsigned char ike_packet_processed = 0; 
	struct tdb *tdbp, *tdbq, *tdbs;	/* Tunnel Descriptor Block pointers */
	char sa[SATOA_BUF];
	size_t sa_len;
	int hard_header_stripped = 0;	/* has the hard header been removed yet? */
	int hard_header_len = 0;
	struct device *physdev;
/*	struct device *virtdev; */
	short physmtu;
	short mtudiff;
#ifdef NET_21
	struct rtable *rt = NULL;
	int pass = 0;
#endif /* NET_21 */

	struct sa_id outgoing_said;
	unsigned int hid;

	tdbs = NULL;

	KLIPS_PRINT(debug_netlink, "ipsec_tunnel: before read_lock\n");
	read_lock(&sync_lock);
	KLIPS_PRINT(debug_netlink, "ipsec_tunnel: after read_lock\n");

	/*
	 *	Return if there is nothing to do.  (Does this ever happen?) XXX
	 */
	if (skb == NULL) {
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    KERN_INFO "klips_error:tunnel: "
			    "Nothing to do!\n" );
		goto cleanup;
	}
	if (dev == NULL) {
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    KERN_INFO "klips_error:tunnel: "
			    "No device associated with skb!\n" );
		goto cleanup;
	}

        hid = 0;
#ifdef CONFIG_SMP
	hid = smp_processor_id();
#endif
#ifdef CONFIG_IPSECTHREAD
	memcpy((char *) &hid, skb->cb, sizeof(unsigned int));
#endif
	prv = dev->priv;
	if (prv == NULL) {
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    KERN_INFO "klips_error:tunnel: "
			    "Device has no private structure!\n" );
		goto cleanup;
	}

	physdev = prv->dev;
	if (physdev == NULL) {
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    KERN_INFO "klips_error:tunnel: "
			    "Device is not attached to physical device!\n" );
		goto cleanup;
	}

	physmtu = physdev->mtu;

	stats = (struct net_device_stats *) &(prv->mystats);

#ifdef NET_21
	/* if skb was cloned (most likely due to a packet sniffer such as
	   tcpdump being momentarily attached to the interface), make
	   a copy of our own to modify */
	if(skb_cloned(skb)) {
		if ( skb_cow(skb, skb_headroom(skb)) ) {
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
		        	    KERN_INFO "klips_error:tunnel: "
				    "skb_cow failed to allocate buffer, dropping.\n" );
			goto cleanup;
		}
	}
#endif /* NET_21 */

#ifdef NET_21
	iph = skb->nh.iph;
#else /* NET_21 */
	iph = skb->ip_hdr;
#endif /* NET_21 */

	/* physdev->hard_header_len is unreliable and should not be used */
	hard_header_len = (unsigned char *)iph - skb->data;

	if(hard_header_len < 0) {
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    KERN_INFO "klips_error:tunnel: "
			    "Negative hard_header_len (%d)?!\n", hard_header_len);
		stats->tx_dropped++;
		goto cleanup;
	}

	if(hard_header_len == 0) { /* no hard header present */
		hard_header_stripped = 1;
	}

#ifdef CONFIG_IPSEC_DEBUG
	if (debug_tunnel & DB_TN_XMIT) {
		int i;
		char c;
		
		printk(KERN_INFO "klips_debug:ipsec_tunnel_start_xmit: "
		       ">>> skb->len=%ld hard_header_len:%d",
		       (unsigned long int)skb->len, hard_header_len);
		c = ' ';
		for (i=0; i < hard_header_len; i++) {
			printk("%c%02x", c, skb->data[i]);
			c = ':';
		}
		printk(" \n");
	}
#endif /* CONFIG_IPSEC_DEBUG */

	KLIPS_IP_PRINT(debug_tunnel & DB_TN_XMIT, iph);

	/*
	 * Sanity checks
	 */

	if ((iph->ihl << 2) != sizeof (struct iphdr)) {
		KLIPS_PRINT(debug_tunnel,
			    KERN_INFO "klips_debug:ipsec_tunnel_start_xmit: "
			    "cannot process IP header options yet.  "
			    "May be mal-formed packet.\n"); /* XXX */
		KLIPS_IP_PRINT(debug_tunnel, iph);
		stats->tx_dropped++;
		goto cleanup;
	}
	
#ifndef NET_21
	/* TTL decrement code (on the way out!) borrowed from ip_forward.c */
	{
		unsigned long checksum = iph->check;
		iph->ttl--;
	/*
	 *	Re-compute the IP header checksum.
	 *	This is efficient. We know what has happened to the header
	 *	and can thus adjust the checksum as Phil Karn does in KA9Q
	 *	except we do this in "network byte order".
	 */
		checksum += htons(0x0100);
		/* carry overflow? */
		checksum += checksum >> 16;
		iph->check = checksum;
	}
	if (iph->ttl <= 0) {
		/* Tell the sender its packet died... */
		ICMP_SEND(skb, ICMP_TIME_EXCEEDED, ICMP_EXC_TTL, 0, physdev);

		KLIPS_PRINT(debug_tunnel, "klips_debug:ipsec_tunnel_start_xmit: "
			    "TTL=0, too many hops!\n");
		stats->tx_dropped++;
		goto cleanup;
	}
#endif /* !NET_21 */
	
	/*
	 * Are we udp/500? If so, let it through (or encrypt with CUG key)
	 * without interference since it is most likely an IKE packet.
	 */

	if(ip_chk_addr((unsigned long)iph->saddr) == IS_MYADDR) 
	{
	  if(iph->protocol == IPPROTO_UDP) 
	  {
	    struct udphdr *udph = (struct udphdr*)((caddr_t)iph + (iph->ihl << 2));	    
	    if(ntohs(udph->dest) == 500) 
	    	ike_packet = 1;
	  }
	}

	/*
	 * First things first -- look us up in the erouting tables.
	 */
	memset(&matcher, 0, sizeof(struct sockaddr_encap));
	matcher.sen_len = sizeof (struct sockaddr_encap);
	matcher.sen_family = AF_ENCAP;
	matcher.sen_type = SENT_IP4;

	if (ike_packet && cug) 
	{
	    matcher.sen_ip_src.s_addr = 0;
	    matcher.sen_ip_dst.s_addr = 0;
	}
	else
	{
	    matcher.sen_ip_src.s_addr = iph->saddr;
	    matcher.sen_ip_dst.s_addr = iph->daddr;
	}
	matcher.sen_proto = iph->protocol;
	matcher.sen_ip_sport = ntohs((iph->protocol == IPPROTO_UDP
			        ? ((struct udphdr*)((caddr_t)iph + (iph->ihl << 2)))->source
				: (iph->protocol == IPPROTO_TCP
				? ((struct tcphdr*)((caddr_t)iph + (iph->ihl << 2)))->source
				: 0)));
	matcher.sen_ip_dport = 	ntohs((iph->protocol == IPPROTO_UDP
				? ((struct udphdr*)((caddr_t)iph + (iph->ihl << 2)))->dest
				: (iph->protocol == IPPROTO_TCP
				? ((struct tcphdr*)((caddr_t)iph + (iph->ihl << 2)))->dest
				: 0)));
	
	
	/*
	 * The spinlock is to prevent any other process from accessing or deleting
	 * the eroute while we are using and updating it.
	 */
	spin_lock(&eroute_lock);
	
	if (!er)
		er = ipsec_findroute(&matcher);
	if(er) {
		outgoing_said = er->er_said;
		er->er_count++;
		er->er_lasttime = jiffies/HZ;
	}

	// special treatment for IKE packets
	
    	if (!er && ike_packet) {	      
		if (cug) {
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    "klips_debug:ipsec_tunnel_start_xmit: "
			    "udp/500 IKE packet - IKE CUG SA not yet initialized\n"); 
		}
		else {
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    "klips_debug:ipsec_tunnel_start_xmit: "
			    "udp/500 IKE packet, sending unprocessed, "
			    "calling dev_queue_xmit\n"); 
			spin_unlock(&eroute_lock);
			goto bypass;
	      }
	}

	spin_unlock(&eroute_lock);

	KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
		    "klips_debug:ipsec_tunnel_start_xmit: "
		    "Original head/tailroom: %d, %d\n",
		    skb_headroom(skb), skb_tailroom(skb));

	innersrc = iph->saddr;
	/* start encapsulation loop here XXX */
	do {

                int icmp_type = 0;
                int icmp_code = 0;

		newdst = orgdst = iph->daddr;
		newsrc = orgsrc = iph->saddr;
		orgedst = outgoing_said.dst.s_addr;
		iphlen = iph->ihl << 2;
		pyldsz = ntohs(iph->tot_len) - iphlen;
		max_headroom = max_tailroom = 0;

#ifdef CONFIG_CLOSEIPSEC
#ifndef CONFIG_CLOSEIPSEC_IN
		/* ICMP Bypass */
 	        if ( (iph->protocol == IPPROTO_ICMP) && (icmp_bypass) ) {
                  icmp_type = ((struct icmphdr*)((caddr_t)iph + (iph->ihl << 2)))->type;
                  icmp_code = ((struct icmphdr*)((caddr_t)iph + (iph->ihl << 2)))->code;
                  if ( icmp_type == 11 )  {    // icmp time exceeded
#ifdef NET_21
                    pass = 1;
#endif
                    goto bypass;
                  }
                }
#endif
#endif

		/* SPI_TRAP/HOLD stuff for better handling of PFK_ACQUIRE ... */

		/* no eroute found (can't happen anymore?) or SPI_DROP set 
		 * --> DROP
		 */
		if(er == NULL || (outgoing_said.proto==IPPROTO_INT
                                  && outgoing_said.spi==htonl(SPI_DROP))) {
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				"klips_debug:ipsec_tunnel_start_xmit: "
				"shunt SA of DROP or no eroute: dropping.\n");
			stats->tx_dropped++;
			goto cleanup;
		}
		
		/* SPI_REJECT set
		 * --> DROP
		 */
		if(outgoing_said.proto==IPPROTO_INT
                   && outgoing_said.spi==htonl(SPI_REJECT)) {
                        KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				"klips_debug:ipsec_tunnel_start_xmit: "
				"shunt SA of REJECT: notifying when coded... and dropping.\n");
                        stats->tx_dropped++;
                        goto cleanup;
                }
	
#ifndef CONFIG_CLOSEIPSEC_IN
		/* SPI_PASS ... we really should avoid this one in SINA, right? 
		 * --> BYPASS ipsec 
		 */
		if(outgoing_said.proto==IPPROTO_INT
                   && outgoing_said.spi==htonl(SPI_PASS)) {
#ifdef NET_21
			pass = 1;
#endif /* NET_21 */
                        KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				"klips_debug:ipsec_tunnel_start_xmit: "
				"shunt SA of PASS: calling dev_queue_xmit\n");
                        goto bypass;
                }
#endif /* CONFIG_CLOSEIPSEC_IN */
		
		/* SPI_HOLD ... packets should be queued, but are dropped for now! */
		if(outgoing_said.proto==IPPROTO_INT
                   && outgoing_said.spi==htonl(SPI_HOLD)) {
                        KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				"klips_debug:ipsec_tunnel_start_xmit: "
				"shunt SA of HOLD: will stash when coded..., dropping for now.\n");
                        stats->tx_dropped++;
                        goto cleanup;
                }
	
		/* SPI_TRAP ... send PFKEY_ACQUIRE to KMd and replace it with SPI_HOLD */
		if(outgoing_said.proto==IPPROTO_INT
                   && outgoing_said.spi==htonl(SPI_TRAP)) {
                        struct tdb tdb;
                        struct sockaddr_in src, dst;
#ifdef CONFIG_IPSEC_DEBUG
                        char bufsrc[ADDRTOA_BUF], bufdst[ADDRTOA_BUF];
#endif /* CONFIG_IPSEC_DEBUG */
                        struct eroute hold_eroute;
                        struct sa_id hold_said;
			int error = 0;
 			
			tdb.tdb_said.proto = iph->protocol;
			src.sin_family = AF_INET;
			dst.sin_family = AF_INET;
			src.sin_addr.s_addr = iph->saddr;
			dst.sin_addr.s_addr = iph->daddr;
			src.sin_port = 
			  (iph->protocol == IPPROTO_UDP
			   ? ((struct udphdr*)((caddr_t)iph + (iph->ihl << 2)))->source
			   : (iph->protocol == IPPROTO_TCP
			      ? ((struct tcphdr*)((caddr_t)iph + (iph->ihl << 2)))->source
			      : 0));
			dst.sin_port = 
			  (iph->protocol == IPPROTO_UDP
			   ? ((struct udphdr*)((caddr_t)iph + (iph->ihl << 2)))->dest
			   : (iph->protocol == IPPROTO_TCP
			      ? ((struct tcphdr*)((caddr_t)iph + (iph->ihl << 2)))->dest
			      : 0));
			
			
			for(i = 0;
			    i < sizeof(struct sockaddr_in)
			      - offsetof(struct sockaddr_in, sin_zero);
			    i++) {
			  src.sin_zero[i] = 0;
			  dst.sin_zero[i] = 0;
			}
			
			tdb.tdb_addr_s = (struct sockaddr *)&src;
			tdb.tdb_addr_d = (struct sockaddr *)&dst;

                        KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				"klips_debug:ipsec_tunnel_start_xmit: "
				"SADB_ACQUIRE sent with src=%s:%d, dst=%s:%d, proto=%d.\n",
				addrtoa(((struct sockaddr_in*)(tdb.tdb_addr_s))->sin_addr, 0, bufsrc, sizeof(bufsrc)) 
					<= ADDRTOA_BUF ? bufsrc : "BAD_ADDR",
				((struct sockaddr_in*)(tdb.tdb_addr_s))->sin_port,
				addrtoa(((struct sockaddr_in*)(tdb.tdb_addr_d))->sin_addr, 0, bufdst, sizeof(bufdst)) 
					<= ADDRTOA_BUF ? bufdst : "BAD_ADDR",
				((struct sockaddr_in*)(tdb.tdb_addr_d))->sin_port,
				tdb.tdb_said.proto);
			 
			pfkey_acquire(&tdb);

                        /* install HOLD eroute */
			memset((caddr_t)&hold_eroute, 0, sizeof(hold_eroute));
			memset((caddr_t)&hold_said, 0, sizeof(hold_said));

			hold_said.proto = IPPROTO_INT;
			hold_said.spi = htonl(SPI_HOLD);
			hold_said.dst.s_addr = 0L;
			
			hold_eroute.er_eaddr.sen_len = sizeof(struct sockaddr_encap);
			hold_eroute.er_emask.sen_len = sizeof(struct sockaddr_encap);
			hold_eroute.er_eaddr.sen_family = AF_ENCAP;
			hold_eroute.er_emask.sen_family = AF_ENCAP;
			hold_eroute.er_eaddr.sen_type = er->er_eaddr.sen_type;
			hold_eroute.er_emask.sen_type = er->er_emask.sen_type;
			
			hold_eroute.er_eaddr.sen_ip_src.s_addr = er->er_eaddr.sen_ip_src.s_addr;
			hold_eroute.er_eaddr.sen_ip_dst.s_addr = er->er_eaddr.sen_ip_dst.s_addr;
			hold_eroute.er_emask.sen_ip_src.s_addr = er->er_emask.sen_ip_src.s_addr;
			hold_eroute.er_emask.sen_ip_dst.s_addr = er->er_emask.sen_ip_dst.s_addr;
			
			/* port and protocol granularity */
			hold_eroute.er_eaddr.sen_proto = er->er_eaddr.sen_proto;
			hold_eroute.er_eaddr.sen_ip_sport = er->er_eaddr.sen_ip_sport;
			hold_eroute.er_eaddr.sen_ip_dport = er->er_eaddr.sen_ip_dport;
			
			hold_eroute.er_count = 0;
			hold_eroute.er_lasttime = jiffies/HZ;

#if 0
#ifdef IPSEC_CONFIG_FULL_SELECTOR_LIST
                        /* These are ficticious.  Don't uncomment these until
			 * proto, sport and dport exist in the SPDB */
                        hold_eroute.er_proto = iph->protocol;
                        hold_eroute.er_src_port =
                                (iph->protocol == IPPROTO_UDP
                                 ? ((struct udphdr*) (((caddr_t)iph) + (iph->ihl << 2)))->source
                                 : (iph->protocol == IPPROTO_TCP
                                    ? ((struct tcphdr*)((caddr_t)iph + (iph->ihl << 2)))->source
                                    : 0));
                        hold_eroute.er_dst_port =
                                (iph->protocol == IPPROTO_UDP
                                 ? ((struct udphdr*) (((caddr_t)iph) + (iph->ihl << 2)))->dest
                                 : (iph->protocol == IPPROTO_TCP
                                    ? ((struct tcphdr*)((caddr_t)iph + (iph->ihl << 2)))->dest
                                    : 0));
#endif /* IPSEC_CONFIG_FULL_SELECTOR_LIST */
#endif												
                        if (!(ipsec_breakroute(&(hold_eroute.er_eaddr),
                                               &(hold_eroute.er_emask)) == EINVAL)) {
                                KLIPS_PRINT(debug_pfkey,
                                            "klips_debug:ipsec_tunnel_start_xmit: "
                                            "breakeroute should have failed.\n");
                                /* SENDERR(-error); */
                        } else {
                                KLIPS_PRINT(debug_pfkey,
                                            "klips_debug:ipsec_tunnel_start_xmit: "
                                            "HOLD breakeroute found nothing as expected.\n");
                        }

                        if ((error = ipsec_makeroute(&(hold_eroute.er_eaddr),
                                                     &(hold_eroute.er_emask),
                                                     hold_said))) {
                                KLIPS_PRINT(debug_pfkey,
                                            "klips_debug:ipsec_tunnel_start_xmit: "
                                            "HOLD makeroute returned %d, failed.\n", error);
                                /* SENDERR(-error); */
                        } else {
                                KLIPS_PRINT(debug_pfkey,
                                            "klips_debug:ipsec_tunnel_start_xmit: "
                                            "HOLD makeroute call successful.\n");
                        }
			
			/* save packets to send them later ??? */
			goto cleanup;
		}

		/*
		 * The spinlock is to prevent any other process from accessing or deleting
		 * the tdb while we are using and updating it.
		 */
		tdbs = tdbp = __find_tdb(&outgoing_said);

		sa_len = satoa(outgoing_said, 0, sa, SATOA_BUF);

		if (tdbp == NULL || atomic_read(&tdbp->del_flag)) {
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "no Tunnel Descriptor Block for SA%s: "
				    "outgoing packet with no SA, dropped.\n", sa);
			stats->tx_dropped++;
			goto cleanup;
		}
		
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    "klips_debug:ipsec_tunnel_start_xmit: "
			    "found Tunnel Descriptor Block -- SA:<%s%s%s> %s\n",
			    TDB_XFORM_NAME(tdbp), sa);
		
		/*
		 * How much headroom do we need to be able to apply
		 * all the grouped transforms?
		 */
		tdbq = tdbp;	/* save the head of the tdb chain */
		while (tdbp)	{
			sa_len = satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);

			/* If it is in larval state, drop the packet, we cannot process yet. */
			if(tdbp->tdb_state == SADB_SASTATE_LARVAL) {
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "TDB in larval state for SA:<%s%s%s> %s, "
					    "cannot be used yet, dropping packet.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				stats->tx_errors++;
				goto cleanup;
			}

			if(tdbp->tdb_state == SADB_SASTATE_DEAD) {
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "TDB in dead state for SA:<%s%s%s> %s, "
					    "can no longer be used, dropping packet.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				stats->tx_errors++;
				goto cleanup;
			}

			/* If the replay window counter == -1, expire SA, it will roll */
			if(tdbp->tdb_replaywin && tdbp->tdb_replaywin_lastseq == -1) {
				pfkey_expire(tdbp, 1);
				markdel_tdbchain(tdbp);
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "replay window counter rolled for SA:<%s%s%s> %s, "
					    "packet dropped, expiring SA.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				stats->tx_errors++;
				goto cleanup;
			}

			/* If any of the lifetime counters have overflowed, expire the SA(s). */
			if(tdbp->tdb_lifetime_bytes_h &&
			   (tdbp->tdb_lifetime_bytes_c > tdbp->tdb_lifetime_bytes_h)) {
				pfkey_expire(tdbp, 1);
				markdel_tdbchain(tdbp);
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "hard bytes lifetime of SA:<%s%s%s> %s has been reached, "
					    "SA expired, outgoing packet dropped.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				stats->tx_errors++;
				goto cleanup;
			}
			if(tdbp->tdb_lifetime_bytes_s &&
			   (tdbp->tdb_lifetime_bytes_c > tdbp->tdb_lifetime_bytes_s)) {
				if (tdbp->tdb_state != SADB_SASTATE_DYING)
				{
				pfkey_expire(tdbp, 0);
				tdbp->tdb_state = SADB_SASTATE_DYING;
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "soft bytes lifetime of SA:<%s%s%s> %s has been reached, "
					    "SA expiring, soft expire message sent up, "
					    "outgoing packet still processed.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				}
			}

			if(tdbp->tdb_lifetime_addtime_h &&
			   ((jiffies / HZ) - tdbp->tdb_lifetime_addtime_c >
			    tdbp->tdb_lifetime_addtime_h)) {
				pfkey_expire(tdbp, 1);
				markdel_tdbchain(tdbp);
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "hard addtime lifetime of SA:<%s%s%s> %s has been reached, "
					    "SA expired, outgoing packet dropped.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				stats->tx_errors++;
				goto cleanup;
			}
			if(tdbp->tdb_lifetime_addtime_s &&
			   ((jiffies / HZ) - tdbp->tdb_lifetime_addtime_c >
			    tdbp->tdb_lifetime_addtime_s)) {
				if (tdbp->tdb_state != SADB_SASTATE_DYING)
				{
				pfkey_expire(tdbp, 0);
				tdbp->tdb_state = SADB_SASTATE_DYING;
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "soft addtime lifetime of SA:<%s%s%s> %s has been reached, "
					    "SA expiring, soft expire message sent up, "
					    "outgoing packet still processed.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				}
			}

			if(tdbp->tdb_lifetime_usetime_c) {
				if(tdbp->tdb_lifetime_usetime_h &&
				   ((jiffies / HZ) - tdbp->tdb_lifetime_usetime_c >
				    tdbp->tdb_lifetime_usetime_h)) {
					pfkey_expire(tdbp, 1);
					markdel_tdbchain(tdbp);
					KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
						    "klips_debug:ipsec_tunnel_start_xmit: "
						    "hard usetime lifetime of SA:<%s%s%s> %s has been reached, "
						    "SA expired, outgoing packet dropped.\n",
						    TDB_XFORM_NAME(tdbp), sa);
					stats->tx_errors++;
					goto cleanup;
				}
				if(tdbp->tdb_lifetime_usetime_s &&
				   ((jiffies / HZ) - tdbp->tdb_lifetime_usetime_c >
				    tdbp->tdb_lifetime_usetime_s)) {
				    if (tdbp->tdb_state != SADB_SASTATE_DYING)
					{
					pfkey_expire(tdbp, 0);
					tdbp->tdb_state = SADB_SASTATE_DYING;
					KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
						    "klips_debug:ipsec_tunnel_start_xmit: "
						    "soft usetime lifetime of SA:<%s%s%s> %s has been reached, "
						    "SA expiring, soft expire message sent up, "
						    "outgoing packet still processed.\n",
						    TDB_XFORM_NAME(tdbp), sa);
					}
				}
			}

			if(tdbp->tdb_lifetime_packets_h &&
			   (tdbp->tdb_lifetime_packets_c > tdbp->tdb_lifetime_packets_h)) {
				pfkey_expire(tdbp, 1);
				markdel_tdbchain(tdbp);
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "hard packets lifetime of SA:<%s%s%s> %s has been reached, "
					    "SA expired, outgoing packet dropped.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				stats->tx_errors++;
				goto cleanup;
			}
			if(tdbp->tdb_lifetime_packets_s &&
			   (tdbp->tdb_lifetime_packets_c > tdbp->tdb_lifetime_packets_s)) {
			   if (tdbp->tdb_state != SADB_SASTATE_DYING)
				{
				pfkey_expire(tdbp, 0);
				tdbp->tdb_state = SADB_SASTATE_DYING;
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "soft packets lifetime of SA:<%s%s%s> %s has been reached, "
					    "SA expiring, soft expire message sent up, "
					    "outgoing packet still processed.\n",
					    TDB_XFORM_NAME(tdbp), sa);
				}
			}

			headroom = tailroom = 0;
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "calling room for <%s%s%s>, SA:%s\n", 
				    TDB_XFORM_NAME(tdbp), sa);
			switch(tdbp->tdb_said.proto) {
#ifdef CONFIG_IPSEC_AH
			case IPPROTO_AH:
				headroom += sizeof(struct ah);
				break;
#endif /* CONFIG_IPSEC_AH */
#ifdef CONFIG_IPSEC_ESP
			case IPPROTO_ESP:
				switch(tdbp->tdb_encalg) {

 				default: 
					headroom += sizeof(struct esp) + tdbp->tdb_iv_size;
					break;

#if 0 // KM *** SINA DOESN'T LIKE ESP_NULL
				case ESP_NULL:
					headroom += offsetof(struct esp, esp_iv);
					break;
#endif // KM ************
				}

				switch(tdbp->tdb_authalg) {
				
				case AH_NONE:
					break;

				default:
					tailroom += AHHMAC_HASHLEN;
					break;
				}			

				if (tdbp->tdb_is_pepp) {
					headroom += PEPP_HEADER_FULL_SIZE;
					tailroom += ((tdbp->tdb_e_bs - ((PEPP_HEADER_FULL_SIZE+pyldsz + 2 * sizeof(unsigned char)) % tdbp->tdb_e_bs)) % tdbp->tdb_e_bs) + 2;
				} else
					tailroom += ((tdbp->tdb_e_bs - ((pyldsz + 2 * sizeof(unsigned char)) % tdbp->tdb_e_bs)) % tdbp->tdb_e_bs) + 2;
				break;

#endif /* CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_IPIP
			case IPPROTO_IPIP:
				headroom += sizeof(struct iphdr);
				break;
#endif /* CONFIG_IPSEC_IPIP */
			default:
				stats->tx_errors++;
				goto cleanup;
			}
			tdbp = tdbp->tdb_onext;
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "Required head/tailroom: %d,%d\n", 
				    headroom, tailroom);
			max_headroom += headroom;
			max_tailroom += tailroom;
			pyldsz += (headroom + tailroom);
		}
		tdbp = tdbq;	/* restore the head of the tdb chain */
		
		KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
			    "klips_debug:ipsec_tunnel_start_xmit: "
			    "existing head/tailroom: %d, %d"
			    " before applying xforms with %d, %d head/tailroom.\n",
			    skb_headroom(skb), skb_tailroom(skb),
			    max_headroom, max_tailroom);
		
		tot_headroom += max_headroom;
		tot_tailroom += max_tailroom;
		
		mtudiff = prv->mtu + tot_headroom + tot_tailroom - physmtu;

		KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
			    "klips_debug:ipsec_tunnel_start_xmit: mtu:%d physmtu:%d "
			    "tothr:%d tottr:%d mtudiff:%d ippkttotlen:%d\n",
			    prv->mtu, physmtu,
			    tot_headroom, tot_tailroom, mtudiff, ntohs(iph->tot_len));
		if(mtudiff > 0) {
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_info:ipsec_tunnel_start_xmit: "
				    "dev %s mtu of %d decreased by %d\n",
				    dev->name,
				    prv->mtu,
				    prv->mtu - (physmtu - tot_headroom - tot_tailroom));
			prv->mtu = physmtu - (tot_headroom + tot_tailroom + 7);	/* add some slop? */
#ifdef NET_21
#if 0
			skb->dst->pmtu = prv->mtu; /* RGB */
#endif /* 0 */
#else /* NET_21 */
#if 0
			dev->mtu = prv->mtu; /* RGB */
#endif /* 0 */
#endif /* NET_21 */
		}
		
#ifdef MSS_HACK
		/*
		 * If this is a transport mode TCP packet with
		 * SYN set, determine an effective MSS based on 
		 * AH/ESP overheads determined above.
		 */
		if (iph->protocol == IPPROTO_TCP 
		    && outgoing_said.proto != IPPROTO_IPIP) {
			struct tcphdr *tcph = skb->h.th;
			if (tcph->syn && !tcph->ack) {
				if(!ipsec_adjust_mss(skb, tcph, prv->mtu)) {
					printk(KERN_WARNING "klips: "
					       "ipsec_adjust_mss() failed\n");
					stats->tx_errors++;
					goto cleanup;
				}
			}
		}
#endif /* MSS_HACK */
		/* spin_unlock(&eroute_lock, eroute_flags); */
		
		if(!hard_header_stripped) {
			if((saved_header = kmalloc(hard_header_len, GFP_ATOMIC)) == NULL) {
				printk(KERN_WARNING "klips_debug:ipsec_tunnel_start_xmit: Failed, "
				       "tried to allocate %d bytes for temp hard_header.\n", 
				       hard_header_len);
				stats->tx_errors++;
				goto cleanup;
			}
			for (i = 0; i < hard_header_len; i++) {
				saved_header[i] = skb->data[i];
			}
			if(skb->len < hard_header_len) {
				printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: "
				       "tried to skb_pull hhlen=%d, %d available.  "
				       "This should never happen, please report.\n",
				       hard_header_len, (int)(skb->len));
				stats->tx_errors++;
				goto cleanup;
			}
			skb_pull(skb, hard_header_len);
			hard_header_stripped = 1;
			
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "head/tailroom: %d, %d after hard_header stripped.\n",
				    skb_headroom(skb), skb_tailroom(skb));
			KLIPS_IP_PRINT(debug_tunnel & DB_TN_CROUT, iph);
		} else {
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "hard header already stripped.\n");
		}
		
		ll_headroom = (hard_header_len + 15) & ~15;

		if ((skb_headroom(skb) >= max_headroom + 2 * ll_headroom) && 
		    (skb_tailroom(skb) >= max_tailroom)
#ifndef NET_21
			&& skb->free
#endif /* !NET_21 */
			) {
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "data fits in existing skb\n");
		} else {
			struct sk_buff* tskb = skb;

			if(!oskb) {
				oskb = skb;
			}

			tskb = skb_copy_expand(skb,
			/* The reason for 2 * link layer length here still baffles me...RGB */
					       max_headroom + 2 * ll_headroom,
					       max_tailroom,
					       GFP_ATOMIC);
#ifdef NET_21
			if(tskb && skb->sk) {
				skb_set_owner_w(tskb, skb->sk);
			}
#endif /* NET_21 */
			if(!(skb == oskb) ) {
				dev_kfree_skb(skb, FREE_WRITE);
			}
			skb = tskb;
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "head/tailroom: %d, %d after allocation\n",
				    skb_headroom(skb), skb_tailroom(skb));
			if (!skb) {
				printk(KERN_WARNING "klips_debug:ipsec_tunnel_start_xmit: Failed, "
				       "tried to allocate %d head and %d tailroom\n", 
				       max_headroom, max_tailroom);
				stats->tx_errors++;
				goto cleanup;
			}
		}
		
		/*
		 * Apply grouped transforms to packet
		 */
		while (tdbp) {
#ifdef CONFIG_IPSEC_ESP
			struct esp *espp;
			__u32 iv[4];
			unsigned char *idat, *pad;
			int authlen = 0, padlen = 0, i;
#endif /* CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_AH
			struct iphdr ipo;
			struct ah *ahp;
#endif /* CONFIG_IPSEC_AH */

			__u8 hash[AH_AMAX];
			int headroom = 0, tailroom = 0, ilen = 0, len = 0;
			unsigned char *dat;
			
			iphlen = iph->ihl << 2;
			pyldsz = ntohs(iph->tot_len) - iphlen;
			sa_len = satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);
			KLIPS_PRINT(debug_tunnel & DB_TN_OXFS,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "calling output for <%s%s%s>, SA:%s\n", 
				    TDB_XFORM_NAME(tdbp), sa);
			
			switch(tdbp->tdb_said.proto) {
#ifdef CONFIG_IPSEC_AH
			case IPPROTO_AH:
				headroom += sizeof(struct ah);
				break;
#endif /* CONFIG_IPSEC_AH */
#ifdef CONFIG_IPSEC_ESP
			case IPPROTO_ESP:
				switch(tdbp->tdb_encalg) {
				
				default:
					headroom += sizeof(struct esp) + tdbp->tdb_iv_size;
					break;

				switch(tdbp->tdb_authalg) {
				case AH_NONE:
					break;

				default:
					authlen = AHHMAC_HASHLEN;
					break;
				}		
				
				if (tdbp->tdb_is_pepp) {
					headroom += PEPP_HEADER_FULL_SIZE;
					tailroom += ((tdbp->tdb_e_bs - ((PEPP_HEADER_FULL_SIZE+pyldsz + 2 * sizeof(unsigned char)) % tdbp->tdb_e_bs)) % tdbp->tdb_e_bs) + 2;
				} else
					tailroom += ((tdbp->tdb_e_bs - ((pyldsz + 2 * sizeof(unsigned char)) % tdbp->tdb_e_bs)) % tdbp->tdb_e_bs) + 2;

				tailroom += authlen;
				break;
#endif /* !CONFIG_IPSEC_ESP */
#ifdef CONFIG_IPSEC_IPIP
			case IPPROTO_IPIP:
				headroom += sizeof(struct iphdr);
				break;
#endif /* !CONFIG_IPSEC_IPIP */
			default:
				stats->tx_errors++;
				goto cleanup;
			}
			
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "pushing %d bytes, putting %d, proto %d.\n", 
				    headroom, tailroom, tdbp->tdb_said.proto);
			if(skb_headroom(skb) < headroom) {
				printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: "
				       "tried to skb_push headroom=%d, %d available.  "
				       "This should never happen, please report.\n",
				       headroom, skb_headroom(skb));
				stats->tx_errors++;
				goto cleanup;
			}
			dat = skb_push(skb, headroom);
			ilen = skb->len - tailroom;
			if(skb_tailroom(skb) < tailroom) {
				printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: "
				       "tried to skb_put %d, %d available.  "
				       "This should never happen, please report.\n",
				       tailroom, skb_tailroom(skb));
				stats->tx_errors++;
				goto cleanup;
			}
			skb_put(skb, tailroom);
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "head/tailroom: %d, %d before xform.\n",
				    skb_headroom(skb), skb_tailroom(skb));
			len = skb->len;
			if(len > 0xfff0) {
				printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: "
				       "tot_len (%d) > 65520.  "
				       "This should never happen, please report.\n",
				       len);
				stats->tx_errors++;
				goto cleanup;
			}
			memmove((void *)dat, (void *)(dat + headroom), iphlen);
			iph = (struct iphdr *)dat;
			iph->tot_len = htons(skb->len);
			
			KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "packet contents before xform:\n");
			KLIPS_IP_PRINT(debug_tunnel & DB_TN_CROUT, iph);
			
			switch(tdbp->tdb_said.proto) {
#ifdef CONFIG_IPSEC_ESP
			case IPPROTO_ESP:
			{
			    	SData In, In_h, Out_e, Out_a, mi, 
			    		  **ppOut = NULL, 
				    	  *aOut[2] ;
			    	CRYPTO_STATUS cs = CRYPTO_OK;

				espp = (struct esp *)(dat + iphlen);
				espp->esp_spi = tdbp->tdb_said.spi;
				espp->esp_rpl = htonl(++(tdbp->tdb_replaywin_lastseq));
				
				/* assume, we always have an IV, 
			           and always known Algs (checked, when we
			           add the tdb 
			        */

				//******* case ESP_WHATEVER for IV-Alg's 
				spin_lock(&tdbp->lock);
				memcpy(iv, tdbp->tdb_iv, tdbp->tdb_iv_size); 
				memcpy((__u8 *)espp+sizeof(struct esp), iv, tdbp->tdb_iv_size); 
				spin_unlock(&tdbp->lock);
				
				idat = dat + iphlen + headroom;
				ilen = len - (iphlen + headroom + authlen);
				
				/* Self-describing padding */
				pad = &dat[len - tailroom];
				padlen = tailroom - 2 - authlen;
				for (i = 0; i < padlen; i++) {
					pad[i] = i + 1; 
				}
				dat[len - authlen - 2] = padlen;
				dat[len - authlen - 1] = iph->protocol;
				iph->protocol = IPPROTO_ESP;
				
				KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
					    "klips_debug:ipsec_tunnel_start_xmit: "
					    "before <%s%s%s>, SA:%s:\n",
					    TDB_XFORM_NAME(tdbp), sa);
				KLIPS_IP_PRINT(debug_tunnel & DB_TN_XMIT, iph);

				/* ESP Encryption (no NULL Encryption allowed !!) */				

				In.nKEK = In_h.nKEK = \
				Out_e.nKEK = Out_a.nKEK = mi.nKEK = 0;
			
				if (tdbp->tdb_is_pepp) {
					In.nDataLength = ilen+PEPP_HEADER_FULL_SIZE;

					In.pData= idat - PEPP_HEADER_FULL_SIZE; 

					memcpy(In.pData,(caddr_t)espp,PEPP_HEADER_SHORT_SIZE);
					i_to_s(In.nDataLength-LEN_PEPP_LEN,(caddr_t)In.pData+PEPP_HEADER_SHORT_SIZE);

				} else {
					In.nDataLength = ilen;
					In.pData = idat;
				}

#ifdef CONFIG_DEBUG_CRYPT
				if (!ike_packet) 
				    dmp("Packet to encrypt:", idat, ilen);
#endif
				    
				Out_e.nDataLength = In.nDataLength; 
				Out_e.pData = kmalloc(In.nDataLength, GFP_ATOMIC);
				if (Out_e.pData == NULL)  {
				      printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: no MEM");
				      goto cleanup;
				}

#ifdef CONFIG_DEBUG_CRYPT
				if (!ike_packet) 
				    dmp("IV used:", (caddr_t)iv, tdbp->tdb_iv_size);
#endif
				    				
				mi.pData = iv;
				mi.nDataLength = tdbp->tdb_iv_size; 


				/* ESP Authentication, if required */
				
				if (tdbp->tdb_authalg) 
				{

					  Out_a.nDataLength = AH_AMAX;
					  Out_a.pData = hash;
					  
					  In_h.pData = (caddr_t)espp;
					  In_h.nDataLength = sizeof(struct esp) + tdbp->tdb_iv_size;
					  
					  aOut[0] = &Out_e;
					  aOut[1] = &Out_a;
					  ppOut = (SData **) &aOut;
					  
					  if (!tdbp->tdb_o_combined_handle[hid])
					  {
					      printk(KERN_WARNING "klips_error: no combined crypto handle");
					      goto cleanup;
				    	  }

#ifdef CONFIG_DEBUG_CRYPT				    	  
					  if (!ike_packet) 
					    dmp("Additional Header to HMAC:", In_h.pData, In_h.nDataLength);
#endif
				    	  
					  // first, HMAC the ESP header
					  if (!tdbp->tdb_is_pepp) {
 					      cs = crypto_transform(tdbp->tdb_authalg_g_handle[hid], &In_h, NULL, NULL);
      					      if (cs != CRYPTO_OK) 
	      			    	      {
				     		 KLIPS_PRINT(debug_tunnel, "error crypto transform auth header#%lu\n",  cs);
						 crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);	
		 			         goto cleanup;
					   	      }
					  }

					  cs = crypto_c_transform(tdbp->tdb_o_combined_handle[hid], &In, (SData **) ppOut, 2, &mi);
	      			    	  if (cs != CRYPTO_OK) 
	      			    	  {
				     		 KLIPS_PRINT(debug_tunnel, "error crypto combined transform #%lu\n",  cs);
					      	 crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);	
		 			         goto cleanup;
				   	  }

#ifdef CONFIG_DEBUG_CRYPT
					  if (!ike_packet) 
					    dmp("HMAC (96 bit in Packet):", hash, authlen);
#endif

					  crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);	
					  memcpy(&(dat[len - authlen]), hash, authlen);
				}  
				else
				{	
				    	  cs = crypto_transform(tdbp->tdb_encalg_e_handle[hid], &In, &Out_e, &mi);
				    	  if (cs != CRYPTO_OK) {
					      KLIPS_PRINT(debug_tunnel, "error crypto transform (encryption) #%lu\n",  cs);
					      goto cleanup;
					  }
				}

#ifdef CONFIG_DEBUG_CRYPT
			        if (!ike_packet) 
				    dmp("Encrypted Packet):",Out_e.pData, Out_e.nDataLength);
#endif
				if (tdbp->tdb_is_pepp) {
					memcpy(idat-PEPP_HEADER_FULL_SIZE, 
						Out_e.pData, Out_e.nDataLength);
				} else {
					memcpy(idat,Out_e.pData,Out_e.nDataLength);
				}
				/* lock the saving of the next iv */
				spin_lock(&tdbp->lock);
				memcpy(tdbp->tdb_iv, mi.pData, tdbp->tdb_iv_size);
				spin_unlock(&tdbp->lock);

				kfree(Out_e.pData);
#ifdef NET_21
				skb->h.raw = (unsigned char*)espp;
#endif /* NET_21 */
				break;
			}
#endif // CONFIG_IPSEC_ESP

#ifdef CONFIG_IPSEC_AH
			case IPPROTO_AH:
				ahp = (struct ah *)(dat + iphlen);
				ahp->ah_spi = tdbp->tdb_said.spi;
				ahp->ah_rpl = htonl(++(tdbp->tdb_replaywin_lastseq));
				ahp->ah_rv = 0;
				ahp->ah_nh = iph->protocol;
				ahp->ah_hl = (headroom >> 2) - sizeof(__u64)/sizeof(__u32);
				iph->protocol = IPPROTO_AH;
				
				ipo = *iph;
				ipo.tos = 0;
				ipo.frag_off = 0;
				ipo.ttl = 0;
				ipo.check = 0;

			        {
				    SData In, Out;
			 	    CRYPTO_STATUS cs = CRYPTO_OK;
		  
				    In.nKEK = 0;
				    In.nDataLength = sizeof (struct iphdr);
				    In.pData = &ipo;
				    	
				    cs = crypto_transform(tdbp->tdb_authalg_g_handle[hid], &In, NULL, NULL);
 				    if (cs != CRYPTO_OK) {
					      KLIPS_PRINT(debug_tunnel, "error crypto AH transform #%lu\n",  cs);
					      cs = crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);
					      stats->tx_errors++;
					      goto cleanup;
					    }

				    In.nDataLength =  headroom - sizeof(ahp->ah_data);
				    In.pData = ahp;
				    
				    cs = crypto_transform(tdbp->tdb_authalg_g_handle[hid], &In, NULL, NULL);
 				    if (cs != CRYPTO_OK) {
					      KLIPS_PRINT(debug_tunnel, "error crypto AH transform #%lu\n",  cs);
					      cs = crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);
					      stats->tx_errors++;
					      goto cleanup;
					    }

				    In.nDataLength = AHHMAC_HASHLEN;
				    In.pData = zeroes;
				    
				    cs = crypto_transform(tdbp->tdb_authalg_g_handle[hid], &In, NULL, NULL);
 				    if (cs != CRYPTO_OK) {
					      KLIPS_PRINT(debug_tunnel, "error crypto AH transform #%lu\n",  cs);
					      cs = crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);
					      stats->tx_errors++;
					      goto cleanup;
					    }

				    In.pData = dat + iphlen + headroom;
				    In.nDataLength =  len - iphlen - headroom;
				 
				    Out.nKEK = 0;
				    Out.nDataLength = AH_AMAX;
				    Out.pData = hash;
				    
				    cs = crypto_transform(tdbp->tdb_authalg_g_handle[hid], &In, &Out, NULL);
			            if (cs != CRYPTO_OK) {
					      KLIPS_PRINT(debug_tunnel, "error crypto transform #%lu\n",  cs);
					      cs = crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);
					      stats->tx_errors++;
					      goto cleanup;
				    }

				    cs = crypto_reopen(tdbp->tdb_authalg_g_handle[hid], NULL);
				    if (cs != CRYPTO_OK) {
					      KLIPS_PRINT(debug_tunnel, "error crypto reopen AH transform #%lu\n",  cs);
					      stats->tx_errors++;
					      goto cleanup;
					    }

				    memcpy(ahp->ah_data, hash, AHHMAC_HASHLEN);
				}
#ifdef NET_21
				skb->h.raw = (unsigned char*)ahp;
#endif /* NET_21 */
				break;
#endif /* CONFIG_IPSEC_AH */
#ifdef CONFIG_IPSEC_IPIP
			case IPPROTO_IPIP:
				iph->version  = 4; /* const should be used here */
#ifdef NET_21
				iph->tos      = skb->nh.iph->tos; /* XXX is this right? */
#else /* NET_21 */
				iph->tos      = skb->ip_hdr->tos; /* XXX is this right? */
#endif /* NET_21 */
				iph->ttl      = 64; /* ip_statistics.IpDefaultTTL; */
				iph->frag_off = 0;
				iph->saddr    = ((struct sockaddr_in*)(tdbp->tdb_addr_s))->sin_addr.s_addr;
				iph->daddr    = ((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr.s_addr;
				iph->protocol = IPPROTO_IPIP;
				iph->ihl      = sizeof(struct iphdr) >> 2 /* 5 */;
#ifdef IP_SELECT_IDENT
				ip_select_ident(iph, skb->dst, NULL);
#else
				iph->id       = htons(ip_id_count++);   /* Race condition here? */
#endif

				newdst = (__u32)iph->daddr;
				newsrc = (__u32)iph->saddr;
		
#ifdef NET_21
				skb->h.ipiph = skb->nh.iph;
#endif /* NET_21 */
				break;
#endif /* CONFIG_IPSEC_IPIP */
			default:
				stats->tx_errors++;
				goto cleanup;
			}
			
#ifdef NET_21
			skb->nh.raw = skb->data;
#else /* NET_21 */
			skb->ip_hdr = skb->h.iph = (struct iphdr *) skb->data;
#endif /* NET_21 */
	
			ip_send_check(iph);
#if 0
			iph->check = 0;
			iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);
#endif			
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				    "klips_debug:ipsec_tunnel_start_xmit: "
				    "after <%s%s%s>, SA:%s:\n",
				    TDB_XFORM_NAME(tdbp), sa);
			KLIPS_IP_PRINT(debug_tunnel & DB_TN_XMIT, iph);
 			
			tdbp->tdb_lifetime_bytes_c += len;
			if(!tdbp->tdb_lifetime_usetime_c) {
				tdbp->tdb_lifetime_usetime_c = jiffies / HZ;
			}
			tdbp->tdb_lifetime_usetime_l = jiffies / HZ;
			tdbp->tdb_lifetime_packets_c += 1;

			tdbp = tdbp->tdb_onext;
			
		}
		/* end encapsulation loop here XXX */


		matcher.sen_ip_src.s_addr = iph->saddr;
		matcher.sen_ip_dst.s_addr = iph->daddr;
		matcher.sen_proto = iph->protocol;
		
		// this is IP-in-IP or ESP or AH --> no ports
		matcher.sen_ip_sport = 0;
		matcher.sen_ip_dport = 0;
		
		spin_lock(&eroute_lock);
		er = ipsec_findroute(&matcher);
		if(er) {
			outgoing_said = er->er_said;
			er->er_count++;
			er->er_lasttime = jiffies/HZ;
			KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
				    "klips_debug:found another matching eroute (recurive)");
		}
		spin_unlock(&eroute_lock);
		KLIPS_PRINT(
			(orgedst != outgoing_said.dst.s_addr) &&
			 outgoing_said.dst.s_addr &&
			 er &&
			 (
			  (ike_packet && (outgoing_said.dst.s_addr != orgdst))
			  || !ike_packet
			 ) &&
			(debug_tunnel & DB_TN_XMIT),
			"klips_debug:ipsec_tunnel_start_xmit: We are recursing here.\n");
		
		if (ike_packet) ike_packet_processed = 1;
		
/* XXXX Test for multiple tunnel - 
 encapsulate again, if:
 	- we found another Eroute 
 	- destination address differs from previous 
 	  (we don't encap the packet twice to the same dest!
 	   multiple tunnemls always end on different GW)
 	- it was an IKE packet AND 
 	  Eroute's dest address differs from IKE dst address 
 	  (in CUG mode, there is ALWAYS a generic Host-to-Host 
 	  Eroute for IKE encryption)
 	  OR 
 	  it's was no IKE packet
 */

	} while((orgedst != outgoing_said.dst.s_addr) &&
		outgoing_said.dst.s_addr &&
		er &&
		 (
		  (ike_packet && (outgoing_said.dst.s_addr != orgdst))
		  || !ike_packet
		 )
		);

	KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
		    "klips_debug:ipsec_tunnel_start_xmit: "
		    "After recursive xforms -- head/tailroom: %d, %d\n",
		    skb_headroom(skb), skb_tailroom(skb));

#if 0  // was CONFIG_IPSEC_ICMP
	if((innersrc != newsrc) && (ntohs(iph->tot_len) > physmtu)) {
		ICMP_SEND(oskb ? oskb : skb,
			  ICMP_DEST_UNREACH,
			  ICMP_FRAG_NEEDED,
			  prv->mtu,
			  physdev);
		KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
			    "klips_debug:ipsec_tunnel_start_xmit: "
			    "IPSEC tunnel mode packet is larger than MTU, ICMP sent.\n");
	}
#endif /* CONFIG_IPSEC_ICMP */

	if(saved_header) {
		if(skb_headroom(skb) < hard_header_len) {
			printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: "
			       "tried to skb_push hhlen=%d, %d available.  "
			       "This should never happen, please report.\n",
			       hard_header_len, skb_headroom(skb));
			stats->tx_errors++;
			goto cleanup;
		}
		skb_push(skb, hard_header_len);
		for (i = 0; i < hard_header_len; i++) {
			skb->data[i] = saved_header[i];
		}
	}

bypass:
	KLIPS_PRINT(debug_tunnel & DB_TN_CROUT,
		    "klips_debug:ipsec_tunnel_start_xmit: "
		    "With hard_header, final head/tailroom: %d, %d\n",
		    skb_headroom(skb), skb_tailroom(skb));
#ifdef CONFIG_IPSEC_DEBUG
	if (debug_tunnel & DB_TN_CROUT)	{
		struct timeval tv;
		do_gettimeofday(&tv);
		printk(KERN_INFO "klips_debug:ipsec_tunnel_start_xmit: "
		       "ts=%02ld.%04ld calling dev_queue_xmit\n", 
		       (unsigned long int)tv.tv_sec % 60,
		       (unsigned long int)tv.tv_usec / 100);
	}
#endif /* CONFIG_IPSEC_DEBUG */

#ifdef NET_21
	/* new route/dst cache code from James Morris */
	skb->dev = physdev;
	/*skb_orphan(skb);*/

	skb->nh.iph->frag_off &= ~htons(IP_DF);
        ip_send_check(skb->nh.iph);
	
	if(ip_route_output(&rt,
			   skb->nh.iph->daddr,
			   pass ? 0 : skb->nh.iph->saddr,
			   RT_TOS(skb->nh.iph->tos),
			   physdev->iflink)) {
		stats->tx_errors++;
                KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    "klips_debug: ip_route_output failure\n");
		goto cleanup;
	}
	
	if(dev == rt->u.dst.dev) {
		ip_rt_put(rt);
		/* This is recursion, drop it. */
		stats->tx_errors++;
		KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
			    "klips_debug:ipsec_tunnel_start_xmit: "
			    "suspect recursion, dev=rt->u.dst.dev=%s, dropped\n", dev->name);
		goto cleanup;
	}
	dst_release(skb->dst);
	skb->dst = &rt->u.dst;
	stats->tx_bytes += skb->len;
	if(skb->len < skb->nh.raw - skb->data) {
		printk(KERN_WARNING "klips_error:ipsec_tunnel_start_xmit: "
		       "tried to __skb_pull nh-data=%d, %d available.  "
		       "This should never happen, please report.\n",
		       skb->nh.raw - skb->data, skb->len);
		stats->tx_errors++;
		goto cleanup;
	}
	__skb_pull(skb, skb->nh.raw - skb->data);
#ifdef SKB_RESET_NFCT
	nf_conntrack_put(skb->nfct);
	skb->nfct = NULL;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif /* CONFIG_NETFILTER_DEBUG */
#endif /* SKB_RESET_NFCT */
	KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
		    "klips_debug:ipsec_tunnel_start_xmit: "
		    "...done, calling ip_send()\n");

#ifdef CONFIG_CLOSEIPSEC
	skb->ipsec_mark = 1;
#endif

#ifdef NETDEV_23
	{
		int err;

		err = NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, rt->u.dst.dev,
			      ipsec_tunnel_xmit2);
		if(err < 0) {
			if(net_ratelimit())
				printk(KERN_ERR
				       "ipsec_tunnel_start_xmit: ip_send() failed, err=%d\n", 
				       -err);
			stats->tx_errors++;
			skb = NULL;
			goto cleanup;
		}
	}
#else
	ip_send(skb);
#endif
#else /* NET_21 */
	skb->arp = 1;
	/* ISDN/ASYNC PPP from Matjaz Godec. */
	/*	skb->protocol = htons(ETH_P_IP); */
	KLIPS_PRINT(debug_tunnel & DB_TN_XMIT,
		    "klips_debug:ipsec_tunnel_start_xmit: "
		    "...done, calling dev_queue_xmit() or ip_fragment().\n");
	IP_SEND(skb, physdev);
#endif /* NET_21 */
	stats->tx_packets++;

	skb = NULL;
 cleanup:
	//if (tdbs) put_tdb(tdbs);
	KLIPS_PRINT(debug_netlink, "ipsec_tunnel: before read_unlock\n");
	read_unlock(&sync_lock);
	wake_up(&sync_lock_wq);
	KLIPS_PRINT(debug_netlink, "ipsec_tunnel: after read_unlock\n");
#if defined(HAS_NETIF_QUEUE) || defined (HAVE_NETIF_QUEUE)
	netif_wake_queue(dev);
#else
	dev->tbusy = 0;
#endif
	if(saved_header)
		kfree(saved_header);
	if(skb) {
		dev_kfree_skb(skb, FREE_WRITE);
	}
	if(oskb)
		dev_kfree_skb(oskb, FREE_WRITE);
	return 0;
}

DEBUG_NO_STATIC struct net_device_stats *
ipsec_tunnel_get_stats(struct device *dev)
{
	return &(((struct ipsecpriv *)(dev->priv))->mystats);
}

/*
 * Revectored calls.
 * For each of these calls, a field exists in our private structure.
 */

DEBUG_NO_STATIC int
ipsec_tunnel_hard_header(struct sk_buff *skb, struct device *dev,
	unsigned short type, void *daddr, void *saddr, unsigned len)
{
	struct ipsecpriv *prv = dev->priv;
	struct device *tmp;
	int ret;

	if(!prv->hard_header) {
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "klips_debug:ipsec_tunnel_hard_header: "
			    "physical device has been detached, packet dropped "
			    "0x%p->0x%p len=%d type=%d dev=%s->NULL ",
			    saddr, daddr, len, type, dev->name);
#ifdef NET_21
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "ip=%08x->%08x\n",
			    (__u32)ntohl(skb->nh.iph->saddr),
			    (__u32)ntohl(skb->nh.iph->daddr) );
#else /* NET_21 */
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "ip=%08x->%08x\n",
			    (__u32)ntohl(skb->ip_hdr->saddr),
			    (__u32)ntohl(skb->ip_hdr->daddr) );
#endif /* NET_21 */
		return -ENODEV;
	}

#define da ((struct device *)(prv->dev))->dev_addr
	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel_hard_header: "
		    "Revectored 0x%p->0x%p len=%d type=%d dev=%s->%s "
		    "dev_addr=%02x:%02x:%02x:%02x:%02x:%02x ",
		    saddr, daddr, len, type, dev->name, prv->dev->name,
		    da[0], da[1], da[2], da[3], da[4], da[5]);
#ifdef NET_21
		    KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
				"ip=%08x->%08x\n",
				(__u32)ntohl(skb->nh.iph->saddr),
				(__u32)ntohl(skb->nh.iph->daddr) );
#else /* NET_21 */
		    KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
				"ip=%08x->%08x\n",
				(__u32)ntohl(skb->ip_hdr->saddr),
				(__u32)ntohl(skb->ip_hdr->daddr) );
#endif /* NET_21 */
	tmp = skb->dev;
	skb->dev = prv->dev;
	ret = prv->hard_header(skb, prv->dev, type, (void *)daddr, (void *)saddr, len);
	skb->dev = tmp;
	return ret;
}

DEBUG_NO_STATIC int
#ifdef NET_21
ipsec_tunnel_rebuild_header(struct sk_buff *skb)
#else /* NET_21 */
ipsec_tunnel_rebuild_header(void *buff, struct device *dev,
			unsigned long raddr, struct sk_buff *skb)
#endif /* NET_21 */
{
	struct ipsecpriv *prv = skb->dev->priv;
	struct device *tmp;
	int ret;
	
	if(!prv->rebuild_header) {
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "klips_debug:ipsec_tunnel_rebuild_header: "
			    "physical device has been detached, packet dropped "
			    "skb->dev=%s->NULL ",
			    skb->dev->name);
#ifdef NET_21
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "ip=%08x->%08x\n",
			    (__u32)ntohl(skb->nh.iph->saddr),
			    (__u32)ntohl(skb->nh.iph->daddr) );
#else /* NET_21 */
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "ip=%08x->%08x\n",
			    (__u32)ntohl(skb->ip_hdr->saddr),
			    (__u32)ntohl(skb->ip_hdr->daddr) );
#endif /* NET_21 */
		return -ENODEV;
	}

	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel: "
		    "Revectored rebuild_header dev=%s->%s ",
		    skb->dev->name, prv->dev->name);
#ifdef NET_21
	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "ip=%08x->%08x\n",
		    (__u32)ntohl(skb->nh.iph->saddr),
		    (__u32)ntohl(skb->nh.iph->daddr) );
#else /* NET_21 */
	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "ip=%08x->%08x\n",
		    (__u32)ntohl(skb->ip_hdr->saddr),
		    (__u32)ntohl(skb->ip_hdr->daddr) );
#endif /* NET_21 */
	tmp = skb->dev;
	skb->dev = prv->dev;
	
#ifdef NET_21
	ret = prv->rebuild_header(skb);
#else /* NET_21 */
	ret = prv->rebuild_header(buff, prv->dev, raddr, skb);
#endif /* NET_21 */
	skb->dev = tmp;
	return ret;
}

DEBUG_NO_STATIC int
ipsec_tunnel_set_mac_address(struct device *dev, void *addr)
{
	struct ipsecpriv *prv = dev->priv;
	
	if(!prv->set_mac_address) {
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "klips_debug:ipsec_tunnel_set_mac_address: "
			    "physical device has been detached, cannot set - "
			    "skb->dev=%s->NULL\n",
			    dev->name);
		return -ENODEV;
	}

	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel_set_mac_address: "
		    "Revectored dev=%s->%s addr=%p\n",
		    dev->name, prv->dev->name, addr);
	return prv->set_mac_address(prv->dev, addr);

}

#ifndef NET_21
DEBUG_NO_STATIC void
ipsec_tunnel_cache_bind(struct hh_cache **hhp, struct device *dev,
				 unsigned short htype, __u32 daddr)
{
	struct ipsecpriv *prv = dev->priv;
	
	if(!prv->header_cache_bind) {
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "klips_debug:ipsec_tunnel_cache_bind: "
			    "physical device has been detached, cannot set - "
			    "skb->dev=%s->NULL\n",
			    dev->name);
		return;
	}

	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel_cache_bind: "
		    "Revectored \n");
	prv->header_cache_bind(hhp, prv->dev, htype, daddr);
	return;
}
#endif /* !NET_21 */


DEBUG_NO_STATIC void
ipsec_tunnel_cache_update(struct hh_cache *hh, struct device *dev, unsigned char *  haddr)
{
	struct ipsecpriv *prv = dev->priv;
	
	if(!prv->header_cache_update) {
		KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
			    "klips_debug:ipsec_tunnel_cache_update: "
			    "physical device has been detached, cannot set - "
			    "skb->dev=%s->NULL\n",
			    dev->name);
		return;
	}

	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel: "
		    "Revectored cache_update\n");
	prv->header_cache_update(hh, prv->dev, haddr);
	return;
}

#ifdef NET_21
DEBUG_NO_STATIC int
ipsec_tunnel_neigh_setup(struct neighbour *n)
{
	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel: "
		    "ipsec_tunnel_neigh_setup\n");

        if (n->nud_state == NUD_NONE) {
                n->ops = &arp_broken_ops;
                n->output = n->ops->output;
        }
        return 0;
}

DEBUG_NO_STATIC int
ipsec_tunnel_neigh_setup_dev(struct device *dev, struct neigh_parms *p)
{
	KLIPS_PRINT(debug_tunnel & DB_TN_REVEC,
		    "klips_debug:ipsec_tunnel: "
		    "ipsec_tunnel_neigh_setup_dev\n");

        if (p->tbl->family == AF_INET) {
                p->neigh_setup = ipsec_tunnel_neigh_setup;
                p->ucast_probes = 0;
                p->mcast_probes = 0;
        }
        return 0;
}
#endif /* NET_21 */

/*
 * We call the attach routine to attach another device.
 */

DEBUG_NO_STATIC int
ipsec_tunnel_attach(struct device *tndev, struct ipsecpriv *prv, struct device *dev)
{
        int i;

	prv->dev = dev;
	prv->hard_start_xmit = dev->hard_start_xmit;
	prv->get_stats = dev->get_stats;

	if (dev->hard_header) {
		prv->hard_header = dev->hard_header;
		tndev->hard_header = ipsec_tunnel_hard_header;
	} else
		tndev->hard_header = NULL;
	
	if (dev->rebuild_header) {
		prv->rebuild_header = dev->rebuild_header;
		tndev->rebuild_header = ipsec_tunnel_rebuild_header;
	} else
		tndev->rebuild_header = NULL;
	
	if (dev->set_mac_address) {
		prv->set_mac_address = dev->set_mac_address;
		tndev->set_mac_address = ipsec_tunnel_set_mac_address;
	} else
		tndev->set_mac_address = NULL;
	
#ifndef NET_21
	if (dev->header_cache_bind) {
		prv->header_cache_bind = dev->header_cache_bind;
		tndev->header_cache_bind = ipsec_tunnel_cache_bind;
	} else
		tndev->header_cache_bind = NULL;
#endif /* !NET_21 */

	if (dev->header_cache_update) {
		prv->header_cache_update = dev->header_cache_update;
		tndev->header_cache_update = ipsec_tunnel_cache_update;
	} else
		tndev->header_cache_update = NULL;

	tndev->hard_header_len = dev->hard_header_len;

#ifdef NET_21
/*	prv->neigh_setup        = dev->neigh_setup; */
	dev->neigh_setup        = ipsec_tunnel_neigh_setup_dev;
#endif /* NET_21 */
	tndev->mtu = 16260; /* 0xfff0; */ /* dev->mtu; */
	prv->mtu = dev->mtu;

#ifdef PHYSDEV_TYPE
	tndev->type = dev->type /* ARPHRD_TUNNEL */;	/* initially */
#endif /*  PHYSDEV_TYPE */

	tndev->addr_len = dev->addr_len;
	for (i=0; i<tndev->addr_len; i++) {
		tndev->dev_addr[i] = dev->dev_addr[i];
	}
#ifdef CONFIG_IPSEC_DEBUG
	if(debug_tunnel & DB_TN_INIT) {
		printk(KERN_INFO "klips_debug:ipsec_tunnel_attach: "
		       "physical device %s being attached has HW address: %2x",
		       dev->name, dev->dev_addr[0]);
		for (i=1; i < dev->addr_len; i++) {
			printk(":%02x", dev->dev_addr[i]);
		}
		printk("\n");
	}
#endif /* CONFIG_IPSEC_DEBUG */

	return 0;
}

/*
 * We call the detach routine to detach the ipsec tunnel from another device.
 */

DEBUG_NO_STATIC int
ipsec_tunnel_detach(struct device *dev, struct ipsecpriv *prv)
{
	int i;

#ifdef CONFIG_IPSEC_NODETACH
	return -EPERM;
#endif

	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		KERN_INFO "klips_debug:ipsec_tunnel_detach: "
		       "physical device %s being detached from virtual device %s\n",
		       prv->dev->name, dev->name);

	prv->dev = NULL;
	prv->hard_start_xmit = NULL;
	prv->get_stats = NULL;

	prv->hard_header = NULL;
	dev->hard_header = NULL;
	
	prv->rebuild_header = NULL;
	dev->rebuild_header = NULL;
	
	prv->set_mac_address = NULL;
	dev->set_mac_address = NULL;
	
#ifndef NET_21
	prv->header_cache_bind = NULL;
	dev->header_cache_bind = NULL;
#endif /* !NET_21 */

	prv->header_cache_update = NULL;
	dev->header_cache_update = NULL;

#ifdef NET_21
/*	prv->neigh_setup        = NULL; */
	dev->neigh_setup        = NULL;
#endif /* NET_21 */
	dev->hard_header_len = 0;
	dev->mtu = 0;
	prv->mtu = 0;
	for (i=0; i<MAX_ADDR_LEN; i++) {
		dev->dev_addr[i] = 0;
	}
	dev->addr_len = 0;
#ifdef PHYSDEV_TYPE
	dev->type = 0;
#endif /*  PHYSDEV_TYPE */
	
	return 0;
}

/*
 * We call the clear routine to detach all ipsec tunnels from other devices.
 */
DEBUG_NO_STATIC int
ipsec_tunnel_clear(void)
{
	int i;
	struct device *ipsecdev = NULL, *prvdev;
	struct ipsecpriv *prv;
	char name[9];
	int ret;

	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_clear: called.\n");

	for(i = 0; i < IPSEC_NUM_IF; i++) {
		sprintf(name, "ipsec%d", i);
		if((ipsecdev = ipsec_dev_get(name)) != NULL) {
			if((prv = (struct ipsecpriv *)(ipsecdev->priv))) {
				prvdev = (struct device *)(prv->dev);
				if(prvdev) {
					KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
						    "klips_debug:ipsec_tunnel_clear: "
						    "physical device for device %s is %s\n",
						    name, prvdev->name);
					if((ret = ipsec_tunnel_detach(ipsecdev, prv))) {
						KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
							    "klips_debug:ipsec_tunnel_clear: "
							    "error %d detatching device %s from device %s.\n",
							    ret, name, prvdev->name);
						return ret;
					}
				}
			}
		}
	}
	return 0;
}

DEBUG_NO_STATIC int
ipsec_tunnel_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
	// struct ipsectunnelconf *cf = (struct ipsectunnelconf *)&ifr->ifr_data;
	struct ipsectunnelconf cf;
	struct ipsecpriv *prv = dev->priv;
	struct device *them; /* physical device */
#ifdef CONFIG_IP_ALIAS
	char *colon;
	char realphysname[IFNAMSIZ];
#endif /* CONFIG_IP_ALIAS */

	if (copy_from_user(&cf, ifr->ifr_data, sizeof(struct ipsectunnelconf)) < 0)
		return -EFAULT;
	
	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_ioctl: "
		    "tncfg service call #%d\n", cmd);
	switch (cmd) {
#ifdef CONFIG_CLOSEIPSEC
	case IPSEC_CTL_DEV:
#endif
	/* attach a virtual ipsec? device to a physical device */
	case IPSEC_SET_DEV:
#ifdef CONFIG_IP_ALIAS
		/* If this is an IP alias interface, get its real physical name */
		strncpy(realphysname, cf.cf_name, IFNAMSIZ);
		realphysname[IFNAMSIZ-1] = 0;
		colon = strchr(realphysname, ':');
		if (colon) *colon = 0;
		them = ipsec_dev_get(realphysname);
#else /* CONFIG_IP_ALIAS */
		them = ipsec_dev_get(cf.cf_name);
#endif /* CONFIG_IP_ALIAS */

		if (them == NULL) {
			KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
				    "klips_debug:ipsec_tunnel_ioctl: "
				    "physical device requested is null\n");

			return -ENXIO;
		}

#ifdef CONFIG_CLOSEIPSEC
		if (cmd == IPSEC_CTL_DEV)
			return closeipsec_update_behav(get_behav(them), cf.cf_behaviour);
#endif
		if (prv->dev)
			return -EBUSY;
		return ipsec_tunnel_attach(dev, dev->priv, them);

	case IPSEC_DEL_DEV:
		if (! prv->dev)
			return -ENODEV;
		return ipsec_tunnel_detach(dev, dev->priv);
	       
	case IPSEC_CLR_DEV:
		KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
			    "klips_debug:ipsec_tunnel_ioctl: "
			    "calling ipsec_tunnel_clear.\n");
		return ipsec_tunnel_clear();

	default:
		return -EOPNOTSUPP;
	}
}

int
ipsec_device_event(struct notifier_block *unused, unsigned long event, void *ptr)
{
	struct device *dev = ptr;
	struct device *ipsec_dev;
	struct ipsecpriv *priv;
	char name[9];
	int i;

	/* check for loopback devices */
	if (dev->flags & IFF_LOOPBACK)
		return(NOTIFY_DONE);

	switch (event) 
	{
		case NETDEV_DOWN:
			KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
				    "ipsec_device_event: NETDEV_DOWN...\n");
			/* find the attached physical device and detach it. */
			for(i = 0; i < IPSEC_NUM_IF; i++) {
				sprintf(name, "ipsec%d", i);
				ipsec_dev = ipsec_dev_get(name);
				if(ipsec_dev) {
					priv = (struct ipsecpriv *)(ipsec_dev->priv);
					if(priv) {
						;
						if(((struct device *)(priv->dev)) == dev) {
							dev_close(ipsec_dev);
							/* return */ ipsec_tunnel_detach(ipsec_dev, priv);
							return NOTIFY_DONE;
							break;
						}
					} else {
						KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
							    "klips_debug:ipsec_device_event: device '%s' has no private data space!\n",
							    ipsec_dev->name);
					}
				}
			}

			break;
		case NETDEV_UP:
			KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
				    "ipsec_device_event: NETDEV_UP...\n");

			/* Only handle ethernet ports */
			if(dev->type!=ARPHRD_ETHER && dev->type!=ARPHRD_LOOPBACK)
				return NOTIFY_DONE;

			break;
#ifdef NET_21
		case NETDEV_UNREGISTER:
			KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
				    "ipsec_device_event: NETDEV_UNREGISTER...\n");

			break;
#endif /* NET_21 */
	}
	return NOTIFY_DONE;
}

/*
 *	Called when an ipsec tunnel device is initialized.
 *	The ipsec tunnel device structure is passed to us.
 */
 
int
ipsec_tunnel_init(struct device *dev)
{
	int i;

	printk(KERN_INFO "klips_debug:ipsec_tunnel_init: "
	       "initialisation of device: %s\n",
	       dev->name ? dev->name : "NULL");

	/* Add our tunnel functions to the device */
	dev->open		= ipsec_tunnel_open;
	dev->stop		= ipsec_tunnel_close;
#ifdef CONFIG_IPSECTHREAD
        dev->hard_start_xmit    = hsips_start_xmit;
#else
	dev->hard_start_xmit	= ipsec_tunnel_start_xmit;
#endif
	dev->get_stats		= ipsec_tunnel_get_stats;

	dev->priv = kmalloc(sizeof(struct ipsecpriv), GFP_ATOMIC);
	if (dev->priv == NULL)
		return -ENOMEM;
	memset(dev->priv, 0, sizeof(struct ipsecpriv));

	for(i = 0; i < sizeof(zeroes); i++) {
		((__u8*)(zeroes))[i] = 0;
	}
	
#ifndef NET_21
	/* Initialize the tunnel device structure */
	for (i = 0; i < DEV_NUMBUFFS; i++)
		skb_queue_head_init(&dev->buffs[i]);
#endif /* !NET_21 */

	dev->set_multicast_list = NULL;
	dev->do_ioctl		= ipsec_tunnel_ioctl;
	dev->hard_header	= NULL;
	dev->rebuild_header 	= NULL;
	dev->set_mac_address 	= NULL;
#ifndef NET_21
	dev->header_cache_bind 	= NULL;
#endif /* !NET_21 */
	dev->header_cache_update= NULL;

#ifdef NET_21
/*	prv->neigh_setup        = NULL; */
	dev->neigh_setup        = ipsec_tunnel_neigh_setup_dev;
#endif /* NET_21 */
	dev->hard_header_len 	= 0;
	dev->mtu		= 0;
	dev->addr_len		= 0;
	dev->type		= ARPHRD_TUNNEL; /* 0 */ /* ARPHRD_ETHER; */ /* initially */
	dev->tx_queue_len	= 10;		/* Small queue */
	memset(dev->broadcast,0xFF, ETH_ALEN);	/* what if this is not attached to ethernet? */

	/* New-style flags. */
	dev->flags		= IFF_NOARP /* 0 */ /* Petr Novak */;
#ifdef NET_21
	dev_init_buffers(dev);
#else /* NET_21 */
	dev->family		= AF_INET;
	dev->pa_addr		= 0;
	dev->pa_brdaddr 	= 0;
	dev->pa_mask		= 0;
	dev->pa_alen		= 4;
#endif /* NET_21 */

	/* We're done.  Have I forgotten anything? */
	return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*  Module specific interface (but it links with the rest of IPSEC  */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int
ipsec_tunnel_probe(struct device *dev)
{
	ipsec_tunnel_init(dev); 
	return 0;
}

static struct device dev_ipsec3 = 
{
	"ipsec3\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec_tunnel_probe	/* setup */
};

static struct device dev_ipsec2 = 
{
	"ipsec2\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec_tunnel_probe	/* setup */
};

static struct device dev_ipsec1 = 
{
	"ipsec1\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec_tunnel_probe	/* setup */
};

static struct device dev_ipsec0 = 
{
	"ipsec0\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec_tunnel_probe	/* setup */
};

int 
ipsec_tunnel_init_devices(void)
{
#if 0
	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_init_devices: "
		    "registering device %s\n",
		    dev_ipsec0.name);
#endif
	if (register_netdev(&dev_ipsec0) != 0)
		return -EIO;
#if 0
	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_init_devices: "
		    "registering device %s\n",
		    dev_ipsec1.name);
#endif
	if (register_netdev(&dev_ipsec1) != 0)
		return -EIO;
#if 0
	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_init_devices: "
		    "registering device %s\n",
		    dev_ipsec2.name);
#endif
	if (register_netdev(&dev_ipsec2) != 0)
		return -EIO;
#if 0
	KLIPS_PRINT(debug_tunnel & DB_TN_INIT,
		    "klips_debug:ipsec_tunnel_init_devices: "
		    "registering device %s\n",
		    dev_ipsec3.name);
#endif
	if (register_netdev(&dev_ipsec3) != 0)
		return -EIO;
	return 0;
}

/* void */
int
ipsec_tunnel_cleanup_devices(void)
{
	int error = 0;

	unregister_netdev(&dev_ipsec0);
	unregister_netdev(&dev_ipsec1);
	unregister_netdev(&dev_ipsec2);
	unregister_netdev(&dev_ipsec3);
	kfree(dev_ipsec0.priv);
	kfree(dev_ipsec1.priv);
	kfree(dev_ipsec2.priv);
	kfree(dev_ipsec3.priv);
	dev_ipsec0.priv=NULL;
	dev_ipsec1.priv=NULL;
	dev_ipsec2.priv=NULL;
	dev_ipsec3.priv=NULL;

	return error;
}

