 /*
 * RFC2367 PF_KEYv2 Key management API message parser
 * Copyright (C) 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: pfkey_v2_parser.c,v 1.22 2003/03/06 16:19:55 br Exp $
 */

/*
 *		Template from klips/net/ipsec/ipsec/ipsec_netlink.c.
 */

char pfkey_v2_parser_c_version[] = "$Id: pfkey_v2_parser.c,v 1.22 2003/03/06 16:19:55 br Exp $";

#include <linux/config.h>
#include <linux/version.h>

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */
#include <linux/smp_lock.h>

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/skbuff.h>
#include <freeswan.h>
#ifdef SPINLOCK
#ifdef SPINLOCK_23
#include <linux/spinlock.h> /* *lock* */
#else /* SPINLOCK_23 */
#include <asm/spinlock.h> /* *lock* */
#endif /* SPINLOCK_23 */
#endif /* SPINLOCK */
#ifdef NET_21
#include <asm/uaccess.h>
#include <linux/in6.h>
#define ip_chk_addr inet_addr_type
#define IS_MYADDR RTN_LOCAL
#endif
#include <asm/checksum.h>
#include <net/ip.h>
#ifdef NETLINK_SOCK
#include <linux/netlink.h>
#else
#include <net/netlink.h>
#endif

#include <linux/random.h>	/* get_random_bytes() */

#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"
#include "ipsec_ah.h"
#include "ipsec_esp.h"
#include "ipsec_tunnel.h"
#include "ipsec_rcv.h"

#include <pfkeyv2.h>
#include <pfkey.h>

#include <linux/cryptdev.h>

extern int cug;

#ifdef CONFIG_IPSECTHREAD
#  include "ipsec_threads/hsips.h"
#endif

#define SENDERR(_x) do { error = -(_x); goto errlab; } while (0)

#ifndef min
#define min(a,b)	(((a)<(b))?(a):(b))
#endif

/* global value */
int plain_key_count = 0;
#ifndef CONFIG_INT_ENC
#define CONFIG_INT_ENC 0
#define CONFIG_MAX_PLAINKEYS 0
#endif

struct sklist_t {
	struct socket *sk;
	struct sklist_t* next;
} pfkey_sklist_head, *pfkey_sklist, *pfkey_sklist_prev;

__u32 pfkey_msg_seq = 0;

extern struct eroute *ike_er;

/* searches for an algorithm ID we got from the crypto module
 * val: Crypto Algorithm to match
 * CPI_type: one of CRYPTO_TYPE_xxx
 * alginfo_type: one of OAKLEY_ENC_ALG / OAKLEY_AUTH_ALG / IPSEC_ENC_ALG 
 *                      IPSEC_AH_AUTH_ALG / IPSEC_AUTH_ALG 
 */

int find_alginfo(long val, int CPI_type, int alginfo_type)
{
	struct list_head *cur, *head;
	struct crypto_algorithm *ca;

	head=get_crypto_algorithm_list();
	
	for ( cur= head->next; cur!= head; cur= cur->next ) {
		ca = list_entry(cur, struct crypto_algorithm, list);
		if (ca->info.type == CPI_type)
	    	if (ca->info.ipsec_id[alginfo_type] == val) {
					put_crypto_algorithm_list();
					return ca->info.index;
			}
	}
	put_crypto_algorithm_list();

	return -1;
}


int
pfkey_alloc_tdb(struct tdb** tdb)
{
	int error = 0;
	if(*tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_alloc_tdb: tdb struct already allocated\n");
		SENDERR(EEXIST);
	}

	if((*tdb = kmalloc(sizeof(**tdb), GFP_ATOMIC) ) == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_alloc_tdb: memory allocation error\n");
		SENDERR(ENOMEM);
	}
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_alloc_tdb: "
		    "allocated tdb struct=%p.\n", tdb);

	memset((caddr_t)*tdb, 0, sizeof(**tdb));
	/* init spin lock for tdb strucutre */
	spin_lock_init(&(*tdb)->lock);

#if defined CONFIG_SMP || defined CONFIG_IPSECTHREAD
        {
		int hno;
		struct tdb *p;
		
		p = *tdb;

#ifdef CONFIG_SMP
		hno = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
		hno = hsips_get_thread_cnt();
#endif
	
		p->tdb_authalg_g_handle = (void **) kmalloc(hno * sizeof(void *), GFP_ATOMIC);
		if (!p->tdb_authalg_g_handle) {
			KLIPS_PRINT(debug_pfkey, "pfkey_alloc_tdb: memory allocation error\n");
			error = ENOMEM;
			goto clean_up;
		}
		memset(p->tdb_authalg_g_handle, 0, hno * sizeof(void *));
		
		p->tdb_authalg_v_handle = (void **) kmalloc(hno * sizeof(void *), GFP_ATOMIC);
		if (!p->tdb_authalg_v_handle) {
			KLIPS_PRINT(debug_pfkey, "pfkey_alloc_tdb: memory allocation error\n");
			error = ENOMEM;
			goto clean_up;
		}
		memset(p->tdb_authalg_v_handle, 0, hno * sizeof(void *));

		p->tdb_encalg_e_handle = (void **) kmalloc(hno * sizeof(void *), GFP_ATOMIC);
		if (!p->tdb_encalg_e_handle) {
			KLIPS_PRINT(debug_pfkey, "pfkey_alloc_tdb: memory allocation error\n");
			error = ENOMEM;
			goto clean_up;
		}
                memset(p->tdb_encalg_e_handle, 0, hno * sizeof(void *));

		p->tdb_encalg_d_handle = (void **) kmalloc(hno * sizeof(void *), GFP_ATOMIC);
		if (!p->tdb_encalg_d_handle) {
			KLIPS_PRINT(debug_pfkey, "pfkey_alloc_tdb: memory allocation error\n");
			error = ENOMEM;
			goto clean_up;
		}
		memset(p->tdb_encalg_d_handle, 0, hno * sizeof(void *));

		p->tdb_i_combined_handle = (void **) kmalloc(hno * sizeof(void *), GFP_ATOMIC);
		if (!p->tdb_i_combined_handle) {
			KLIPS_PRINT(debug_pfkey, "pfkey_alloc_tdb: memory allocation error\n");
			error = ENOMEM;
			goto clean_up;
		}
		memset(p->tdb_i_combined_handle, 0, hno * sizeof(void *));

		p->tdb_o_combined_handle = (void **) kmalloc(hno * sizeof(void *), GFP_ATOMIC);
		if (!p->tdb_o_combined_handle) {
			KLIPS_PRINT(debug_pfkey, "pfkey_alloc_tdb: memory allocation error\n");
			error = ENOMEM;
			goto clean_up;
		}
		memset(p->tdb_o_combined_handle, 0, hno * sizeof(void *));
	}
#endif

errlab:
	return(error);

#if defined CONFIG_SMP || defined CONFIG_IPSECTHREAD
clean_up:
        if ((*tdb)->tdb_authalg_g_handle)
	                kfree((*tdb)->tdb_authalg_g_handle);
        if ((*tdb)->tdb_authalg_v_handle)
	                kfree((*tdb)->tdb_authalg_v_handle);
        if ((*tdb)->tdb_encalg_e_handle)
	                kfree((*tdb)->tdb_encalg_e_handle);
        if ((*tdb)->tdb_encalg_d_handle)
	                kfree((*tdb)->tdb_encalg_d_handle);
        if ((*tdb)->tdb_i_combined_handle)
	                kfree((*tdb)->tdb_i_combined_handle);
        if ((*tdb)->tdb_o_combined_handle)
	                kfree((*tdb)->tdb_o_combined_handle);

        kfree(*tdb);
        SENDERR(error);
#endif
}

int
pfkey_alloc_eroute(struct eroute** eroute)
{
	int error = 0;
	if(*eroute) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_alloc_eroute: eroute struct already allocated\n");
		SENDERR(EEXIST);
	}

	if((*eroute = kmalloc(sizeof(**eroute), GFP_ATOMIC) ) == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_alloc_eroute: memory allocation error\n");
		SENDERR(ENOMEM);
	}
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_alloc_eroute: "
		    "allocated eroute struct=%p.\n", eroute);
	memset((caddr_t)*eroute, 0, sizeof(**eroute));
	(*eroute)->er_eaddr.sen_len =
		(*eroute)->er_emask.sen_len = sizeof(struct sockaddr_encap);
	(*eroute)->er_eaddr.sen_family =
		(*eroute)->er_emask.sen_family = AF_ENCAP;
	(*eroute)->er_eaddr.sen_type = SENT_IP4;
	(*eroute)->er_emask.sen_type = 255;
	(*eroute)->er_count = 0;
	(*eroute)->er_lasttime = jiffies/HZ;

 errlab:
	return(error);
}

static int
pfkey_sa_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
	struct sadb_sa *pfkey_sa = (struct sadb_sa *)pfkey_ext;
	int error = 0;
	struct tdb* tdbp;
	struct crypto_algorithm *ca;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_sa_process: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_process: "
			    "extr or extr->tdb is NULL, fatal\n");
		SENDERR(EINVAL);
	}

#ifdef CONFIG_CLOSEIPSEC_IN
	if( pfkey_sa->sadb_sa_spi == htonl(SPI_PASS) ) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_process: "
			    "no bypass SA allowed by kernel configuration!\n");
		SENDERR(EINVAL);
	}
#endif /* CONFIG_CLOSEIPSEC_IN */

	switch(pfkey_ext->sadb_ext_type) {
	case SADB_EXT_SA:
		tdbp = extr->tdb;
		break;
	case SADB_X_EXT_SA2:
		if(pfkey_alloc_tdb(&(extr->tdb2)) == ENOMEM) {
			SENDERR(ENOMEM);
		}
		tdbp = extr->tdb2;
		break;
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_process: invalid exttype=%d.\n",
			    pfkey_ext->sadb_ext_type);
		SENDERR(EINVAL);
	}

	tdbp->tdb_said.spi = pfkey_sa->sadb_sa_spi;
	tdbp->tdb_replaywin = pfkey_sa->sadb_sa_replay;
	tdbp->tdb_state = pfkey_sa->sadb_sa_state;
	tdbp->tdb_flags = pfkey_sa->sadb_sa_flags;
	tdbp->tdb_replaywin_lastseq = tdbp->tdb_replaywin_bitmap = 0;

	switch(tdbp->tdb_said.proto) {
		case IPPROTO_AH:
			tdbp->tdb_authalg = pfkey_sa->sadb_sa_auth;

			if(tdbp->tdb_authalg)
			{
				KLIPS_PRINT(debug_pfkey, "looking for AH Auth alg: %d", pfkey_sa->sadb_sa_auth);
				// find associated algorithm numbers  
				tdbp->tdb_authalg_g_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_HMAC_GENERATE, IPSEC_AH_AUTH_ALG);
				tdbp->tdb_authalg_v_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_HMAC_VERIFY, IPSEC_AH_AUTH_ALG);
				// try MAC instead of HMAC...
				if( (tdbp->tdb_authalg_g_num < 0) || 
				    (tdbp->tdb_authalg_v_num < 0))
					tdbp->tdb_authalg_g_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_MAC_GENERATE, IPSEC_AH_AUTH_ALG);
					tdbp->tdb_authalg_v_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_MAC_VERIFY, IPSEC_AH_AUTH_ALG);
				if( (tdbp->tdb_authalg_g_num < 0) || 
				    (tdbp->tdb_authalg_v_num < 0))
				    {
				       	KLIPS_PRINT(debug_pfkey,"AH Algorithm not found\n");
				       	tdbp->tdb_authalg = 0;	// should we do this??
				       	SENDERR(EINVAL);
				    }	
				ca = get_algorithm(tdbp->tdb_authalg_g_num);
				tdbp->tdb_key_a_size = ca->info.keylen;	
			        tdbp->tdb_is_pepp= ca->info.properties & CRYPTO_PROP_PEPP;
				put_algorithm(ca);
			}
			
			tdbp->tdb_encalg = SADB_EALG_NONE;

			break;
			
		case IPPROTO_ESP:
			tdbp->tdb_authalg = pfkey_sa->sadb_sa_auth;
			
			if (tdbp->tdb_authalg)
			{
				KLIPS_PRINT(debug_pfkey, "looking for ESP Auth alg: %d", pfkey_sa->sadb_sa_auth);

				// find associated algorithm numbers  
				tdbp->tdb_authalg_g_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_HMAC_GENERATE, IPSEC_AH_AUTH_ALG);
				tdbp->tdb_authalg_v_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_HMAC_VERIFY, IPSEC_AH_AUTH_ALG);
				if( (tdbp->tdb_authalg_g_num < 0) || 
				    (tdbp->tdb_authalg_v_num < 0)) {
					tdbp->tdb_authalg_g_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_MAC_GENERATE, IPSEC_AH_AUTH_ALG);
					tdbp->tdb_authalg_v_num = find_alginfo(pfkey_sa->sadb_sa_auth, CRYPTO_TYPE_MAC_VERIFY, IPSEC_AH_AUTH_ALG);
				}
				if( (tdbp->tdb_authalg_g_num < 0) || 
				    (tdbp->tdb_authalg_v_num < 0))
				    {
				       	KLIPS_PRINT(debug_pfkey,"ESP Auth Algorithm not found\n");
				       	tdbp->tdb_authalg = 0;	// should we do this??
				       	SENDERR(EINVAL);
				    }
				ca = get_algorithm(tdbp->tdb_authalg_g_num);
				tdbp->tdb_key_a_size = ca->info.keylen;	
			        tdbp->tdb_is_pepp= ca->info.properties & CRYPTO_PROP_PEPP;
				put_algorithm(ca);
			}			
			
			tdbp->tdb_encalg = pfkey_sa->sadb_sa_encrypt;

			if (tdbp->tdb_encalg)
			{
				KLIPS_PRINT(debug_pfkey,"looking for ESP Enc alg: %d", pfkey_sa->sadb_sa_encrypt);
				tdbp->tdb_encalg_e_num = find_alginfo(pfkey_sa->sadb_sa_encrypt, CRYPTO_TYPE_ENCRYPT, IPSEC_ENC_ALG);  
				tdbp->tdb_encalg_d_num = find_alginfo(pfkey_sa->sadb_sa_encrypt, CRYPTO_TYPE_DECRYPT, IPSEC_ENC_ALG);  
				if( (tdbp->tdb_encalg_e_num < 0) || 
				    (tdbp->tdb_encalg_d_num < 0))
				    {	
				       	KLIPS_PRINT(debug_pfkey,"ESP Algorithm not found\n");
				       	tdbp->tdb_encalg = 0;	// should we do this??
				       	SENDERR(EINVAL);
				    }
				ca = get_algorithm(tdbp->tdb_encalg_e_num);
				tdbp->tdb_key_e_size = ca->info.keylen;	
				tdbp->tdb_e_bs = ca->info.blocklen;
			        tdbp->tdb_is_pepp= ca->info.properties & CRYPTO_PROP_PEPP;
				put_algorithm(ca);
			}
						
			break;
		case IPPROTO_IPIP:
			tdbp->tdb_authalg = AH_NONE;
			tdbp->tdb_encalg = ESP_NONE;
			break;
#ifdef CONFIG_IPSEC_COMP
		case IPPROTO_COMP:
			/* one of these will be used to hold the compression alg */
			tdbp->tdb_authalg = AH_NONE;
			tdbp->tdb_encalg = ESP_NONE;
			break;
#endif /* CONFIG_IPSEC_COMP */
		case IPPROTO_INT:
			tdbp->tdb_authalg = AH_NONE;
			tdbp->tdb_encalg = ESP_NONE;
			break;
		default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sa_process: "
			    "unknown or unset proto=%d.\n",
			    tdbp->tdb_said.proto);
#if 0
			SENDERR(EINVAL);
#endif
	}

errlab:
	return error;
}

static int
pfkey_lifetime_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct sadb_lifetime *pfkey_lifetime = (struct sadb_lifetime *)pfkey_ext;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_lifetime_process: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_lifetime_process: "
			    "extr or extr->tdb is NULL, fatal\n");
		SENDERR(EINVAL);
	}

	switch(pfkey_lifetime->sadb_lifetime_exttype) {
	case SADB_EXT_LIFETIME_CURRENT:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_lifetime_process: "
			    "lifetime_current not supported yet.\n");
  		SENDERR(EINVAL);
  		break;
	case SADB_EXT_LIFETIME_HARD:
		if(pfkey_lifetime->sadb_lifetime_allocations &&
		   (!extr->tdb->tdb_lifetime_allocations_h ||
		    (pfkey_lifetime->sadb_lifetime_allocations < extr->tdb->tdb_lifetime_allocations_h))) {
			extr->tdb->tdb_lifetime_allocations_h =
				pfkey_lifetime->sadb_lifetime_allocations;
			if(!extr->tdb->tdb_lifetime_allocations_s &&
			   (extr->tdb->tdb_lifetime_allocations_h < extr->tdb->tdb_lifetime_allocations_s)) {
				extr->tdb->tdb_lifetime_allocations_s = extr->tdb->tdb_lifetime_allocations_h;
			}
		}
		if(pfkey_lifetime->sadb_lifetime_bytes &&
		   (!extr->tdb->tdb_lifetime_bytes_h ||
		    (pfkey_lifetime->sadb_lifetime_bytes < extr->tdb->tdb_lifetime_bytes_h))) {
			extr->tdb->tdb_lifetime_bytes_h =
				pfkey_lifetime->sadb_lifetime_bytes;
			if(!extr->tdb->tdb_lifetime_bytes_s &&
			   (extr->tdb->tdb_lifetime_bytes_h < extr->tdb->tdb_lifetime_bytes_s)) {
				extr->tdb->tdb_lifetime_bytes_s = extr->tdb->tdb_lifetime_bytes_h;
			
			}
		}
		if(pfkey_lifetime->sadb_lifetime_addtime &&
		   (!extr->tdb->tdb_lifetime_addtime_h ||
		    (pfkey_lifetime->sadb_lifetime_addtime < extr->tdb->tdb_lifetime_addtime_h))) {
			extr->tdb->tdb_lifetime_addtime_h =
				pfkey_lifetime->sadb_lifetime_addtime;
			if(extr->tdb->tdb_lifetime_addtime_s &&
			   (extr->tdb->tdb_lifetime_addtime_h < extr->tdb->tdb_lifetime_addtime_s)) {
				extr->tdb->tdb_lifetime_addtime_s = extr->tdb->tdb_lifetime_addtime_h;
			}
		}
		if(pfkey_lifetime->sadb_lifetime_usetime &&
		   (!extr->tdb->tdb_lifetime_usetime_h ||
		    (pfkey_lifetime->sadb_lifetime_usetime < extr->tdb->tdb_lifetime_usetime_h))) {
			extr->tdb->tdb_lifetime_usetime_h =
				pfkey_lifetime->sadb_lifetime_usetime;
			if(extr->tdb->tdb_lifetime_usetime_s &&
			   (extr->tdb->tdb_lifetime_usetime_h < extr->tdb->tdb_lifetime_usetime_s)) {
				extr->tdb->tdb_lifetime_usetime_s = extr->tdb->tdb_lifetime_usetime_h;
			}
		}
		break;
	case SADB_EXT_LIFETIME_SOFT:
		if(pfkey_lifetime->sadb_lifetime_allocations &&
		   (!extr->tdb->tdb_lifetime_allocations_s ||
		    (pfkey_lifetime->sadb_lifetime_allocations < extr->tdb->tdb_lifetime_allocations_s))) {
			extr->tdb->tdb_lifetime_allocations_s =
				pfkey_lifetime->sadb_lifetime_allocations;
			if(extr->tdb->tdb_lifetime_allocations_h &&
			   (extr->tdb->tdb_lifetime_allocations_h < extr->tdb->tdb_lifetime_allocations_s)) {
				extr->tdb->tdb_lifetime_allocations_s = extr->tdb->tdb_lifetime_allocations_h;
			}
		}
		if(pfkey_lifetime->sadb_lifetime_bytes &&
		   (!extr->tdb->tdb_lifetime_bytes_s ||
		    (pfkey_lifetime->sadb_lifetime_bytes < extr->tdb->tdb_lifetime_bytes_s))) {
			extr->tdb->tdb_lifetime_bytes_s =
				pfkey_lifetime->sadb_lifetime_bytes;
			if(extr->tdb->tdb_lifetime_bytes_h &&
			   (extr->tdb->tdb_lifetime_bytes_h < extr->tdb->tdb_lifetime_bytes_s)) {
				extr->tdb->tdb_lifetime_bytes_s = extr->tdb->tdb_lifetime_bytes_h;
			}
		}
		if(pfkey_lifetime->sadb_lifetime_addtime &&
		   (!extr->tdb->tdb_lifetime_addtime_s ||
		    (pfkey_lifetime->sadb_lifetime_addtime < extr->tdb->tdb_lifetime_addtime_s))) {
			extr->tdb->tdb_lifetime_addtime_s =
				pfkey_lifetime->sadb_lifetime_addtime;
			if(extr->tdb->tdb_lifetime_addtime_h &&
			   (extr->tdb->tdb_lifetime_addtime_h < extr->tdb->tdb_lifetime_addtime_s)) {
				extr->tdb->tdb_lifetime_addtime_s = extr->tdb->tdb_lifetime_addtime_h;
			}
		}
		if(pfkey_lifetime->sadb_lifetime_usetime &&
		   (!extr->tdb->tdb_lifetime_usetime_s ||
		    (pfkey_lifetime->sadb_lifetime_usetime < extr->tdb->tdb_lifetime_usetime_s))) {
			extr->tdb->tdb_lifetime_usetime_s =
				pfkey_lifetime->sadb_lifetime_usetime;
			if(extr->tdb->tdb_lifetime_usetime_h &&
			   (extr->tdb->tdb_lifetime_usetime_h < extr->tdb->tdb_lifetime_usetime_s)) {
				extr->tdb->tdb_lifetime_usetime_s = extr->tdb->tdb_lifetime_usetime_h;
			}
		}
		break;
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_lifetime_process: invalid exttype=%d.\n",
			    pfkey_ext->sadb_ext_type);
		SENDERR(EINVAL);
	}

errlab:
	return error;
}

static int
pfkey_address_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
	int error = 0;
	int saddr_len = 0;
	char ipaddr_txt[ADDRTOA_BUF];
	unsigned char **sap;
	struct sadb_address *pfkey_address = (struct sadb_address *)pfkey_ext;
	struct sockaddr* s = (struct sockaddr*)((char*)pfkey_address + sizeof(*pfkey_address));
	struct tdb* tdbp;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_address_process:\n");
	
	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process:\n"
			    "extr or extr->tdb is NULL, fatal\n");
		SENDERR(EINVAL);
	}

	switch(s->sa_family) {
	case AF_INET:
		saddr_len = sizeof(struct sockaddr_in);
		addrtoa(((struct sockaddr_in*)s)->sin_addr, 0, ipaddr_txt, sizeof(ipaddr_txt));
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found address family=%d, AF_INET, %s.\n",
			    s->sa_family, ipaddr_txt);
		break;
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	case AF_INET6:
		saddr_len = sizeof(struct sockaddr_in6);
		break;
#endif /* defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE) */
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: s->sa_family=%d not supported.\n",
			    s->sa_family);
		SENDERR(EPFNOSUPPORT);
	}
	
	switch(pfkey_address->sadb_address_exttype) {
	case SADB_EXT_ADDRESS_SRC:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found src address.\n");
		sap = (unsigned char **)&(extr->tdb->tdb_addr_s);
		extr->tdb->tdb_addr_s_size = saddr_len;
		break;
	case SADB_EXT_ADDRESS_DST:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found dst address.\n");
		sap = (unsigned char **)&(extr->tdb->tdb_addr_d);
		extr->tdb->tdb_addr_d_size = saddr_len;
		break;
	case SADB_EXT_ADDRESS_PROXY:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found proxy address.\n");
		sap = (unsigned char **)&(extr->tdb->tdb_addr_p);
		extr->tdb->tdb_addr_p_size = saddr_len;
		break;
	case SADB_X_EXT_ADDRESS_DST2:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found 2nd dst address.\n");
		if(pfkey_alloc_tdb(&(extr->tdb2)) == ENOMEM) {
			SENDERR(ENOMEM);
		}
		sap = (unsigned char **)&(extr->tdb2->tdb_addr_d);
		extr->tdb2->tdb_addr_d_size = saddr_len;
		break;
	case SADB_X_EXT_ADDRESS_SRC_FLOW:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found src flow address.\n");
		if(pfkey_alloc_eroute(&(extr->eroute)) == ENOMEM) {
			SENDERR(ENOMEM);
		}
		sap = (unsigned char **)&(extr->eroute->er_eaddr.sen_ip_src);

		extr->eroute->er_eaddr.sen_ip_sport = ((struct sockaddr_in*)s)->sin_port;
		extr->eroute->er_eaddr.sen_proto = pfkey_address->sadb_address_proto;
		break;

	case SADB_X_EXT_ADDRESS_DST_FLOW:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found dst flow address.\n");
		if(pfkey_alloc_eroute(&(extr->eroute)) == ENOMEM) {
			SENDERR(ENOMEM);
		}
		sap = (unsigned char **)&(extr->eroute->er_eaddr.sen_ip_dst);

		extr->eroute->er_eaddr.sen_ip_dport = ((struct sockaddr_in*)s)->sin_port;
		extr->eroute->er_eaddr.sen_proto = pfkey_address->sadb_address_proto;
		break;

	case SADB_X_EXT_ADDRESS_SRC_MASK:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found src mask address.\n");
		if(pfkey_alloc_eroute(&(extr->eroute)) == ENOMEM) {
			SENDERR(ENOMEM);
		}
		sap = (unsigned char **)&(extr->eroute->er_emask.sen_ip_src);
		break;
	case SADB_X_EXT_ADDRESS_DST_MASK:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: found dst mask address.\n");
		if(pfkey_alloc_eroute(&(extr->eroute)) == ENOMEM) {
			SENDERR(ENOMEM);
		}
		sap = (unsigned char **)&(extr->eroute->er_emask.sen_ip_dst);
		break;
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_address_process: unrecognised ext_type=%d.\n",
			    pfkey_address->sadb_address_exttype);
		SENDERR(EINVAL);
	}
	
	switch(pfkey_address->sadb_address_exttype) {
	case SADB_EXT_ADDRESS_SRC:
	case SADB_EXT_ADDRESS_DST:
	case SADB_EXT_ADDRESS_PROXY:
	case SADB_X_EXT_ADDRESS_DST2:
		if(!(*sap = kmalloc(saddr_len, GFP_ATOMIC))) {
			SENDERR(ENOMEM);
		}
		memcpy(*sap, s, saddr_len);
		break;
	default:
		if(s->sa_family	!= AF_INET) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_process: s->sa_family=%d not supported.\n",
				    s->sa_family);
			SENDERR(EPFNOSUPPORT);
		}
		(unsigned int)(*sap) = ((struct sockaddr_in*)s)->sin_addr.s_addr;
#ifdef CONFIG_IPSEC_DEBUG
		if(extr->eroute) {
			char buf1[64], buf2[64];
			if (debug_pfkey) {
				subnettoa(extr->eroute->er_eaddr.sen_ip_src,
					  extr->eroute->er_emask.sen_ip_src, 0, buf1, sizeof(buf1));
				subnettoa(extr->eroute->er_eaddr.sen_ip_dst,
					  extr->eroute->er_emask.sen_ip_dst, 0, buf2, sizeof(buf2));
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_address_parse: "
					    "extr->eroute set to %s->%s\n",
					    buf1, buf2);
			}
		}
#endif /* CONFIG_IPSEC_DEBUG */
	}

	tdbp = extr->tdb;
	switch(pfkey_address->sadb_address_exttype) {
	case SADB_X_EXT_ADDRESS_DST2:
		tdbp = extr->tdb2;
	case SADB_EXT_ADDRESS_DST:
		if(s->sa_family == AF_INET) {
			tdbp->tdb_said.dst.s_addr = ((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr.s_addr;
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_process: tdbp->tdb_said.dst.s_addr=%08x,\n"
				    "klips_debug:                     ((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr.s_addr=%08x,\n",
				    tdbp->tdb_said.dst.s_addr,
				    ((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr.s_addr
				);
			addrtoa(((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr,
				0,
				ipaddr_txt,
				sizeof(ipaddr_txt));
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_process: tdb_said.dst set to %s.\n",
				    ipaddr_txt);
		} else {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_address_process: "
				    "uh, tdb_said.dst doesn't do address family=%d yet, "
				    "said will be invalid.\n",
				    s->sa_family);
		}
	default:
	}
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_address_process: successful.\n");
 errlab:
	return error;
}

static int
pfkey_key_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
        int error = 0;
        struct sadb_key *pfkey_key = (struct sadb_key *)pfkey_ext;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_key_process: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_key_process: "
			    "extr or extr->tdb is NULL, fatal\n");
		SENDERR(EINVAL);
	}

        switch(pfkey_key->sadb_key_exttype) {
        case SADB_EXT_KEY_AUTH:
                extr->tdb->tdb_key_bits_a = pfkey_key->sadb_key_bits;
		if(!(extr->tdb->tdb_key_a = kmalloc(DIVUP(pfkey_key->sadb_key_bits, 8), GFP_ATOMIC))) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_key_process: memory allocation error.\n");
			SENDERR(ENOMEM);
		}
		memcpy(extr->tdb->tdb_key_a,
		       (char*)pfkey_key + sizeof(struct sadb_key),
		       DIVUP(pfkey_key->sadb_key_bits, 8));
		break;
	case SADB_EXT_KEY_ENCRYPT: /* Key(s) */
		extr->tdb->tdb_key_bits_e = pfkey_key->sadb_key_bits;
		if(!(extr->tdb->tdb_key_e = kmalloc(DIVUP(pfkey_key->sadb_key_bits, 8), GFP_ATOMIC))) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_key_process: memory allocation error.\n");
			SENDERR(ENOMEM);
		}
		memcpy(extr->tdb->tdb_key_e,
		       (char*)pfkey_key + sizeof(struct sadb_key),
		       DIVUP(pfkey_key->sadb_key_bits, 8));
		break;
	default:
		SENDERR(EINVAL);
 	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_key_process: success.\n");
errlab:
	return error;
}

static int
pfkey_ident_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
        int error = 0;
        struct sadb_ident *pfkey_ident = (struct sadb_ident *)pfkey_ext;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_ident_process: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_ident_process: "
			    "extr or extr->tdb is NULL, fatal\n");
		SENDERR(EINVAL);
	}

	switch(pfkey_ident->sadb_ident_exttype) {
	case SADB_EXT_IDENTITY_SRC:
		extr->tdb->tdb_ident_type_s = pfkey_ident->sadb_ident_type;
		extr->tdb->tdb_ident_id_s = pfkey_ident->sadb_ident_id;
		extr->tdb->tdb_ident_len_s = pfkey_ident->sadb_ident_len -
			sizeof(struct sadb_ident) / IPSEC_PFKEYv2_ALIGN;
		if(!(extr->tdb->tdb_ident_data_s = kmalloc(pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN, GFP_ATOMIC))) {
			SENDERR(ENOMEM);
		}
		memcpy(extr->tdb->tdb_ident_data_s,
		       (char*)pfkey_ident + sizeof(struct sadb_ident),
		       pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN);
		break;
	case SADB_EXT_IDENTITY_DST: /* Identity(ies) */
		extr->tdb->tdb_ident_type_d = pfkey_ident->sadb_ident_type;
		extr->tdb->tdb_ident_id_d = pfkey_ident->sadb_ident_id;
		extr->tdb->tdb_ident_len_d = pfkey_ident->sadb_ident_len -
			sizeof(struct sadb_ident) / IPSEC_PFKEYv2_ALIGN;
		if(!(extr->tdb->tdb_ident_data_d = kmalloc(pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN, GFP_ATOMIC))) {
			SENDERR(ENOMEM);
		}
		memcpy(extr->tdb->tdb_ident_data_d,
		       (char*)pfkey_ident + sizeof(struct sadb_ident),
		       pfkey_ident->sadb_ident_len * IPSEC_PFKEYv2_ALIGN);
		break;
	default:
		SENDERR(EINVAL);
 	}
errlab:
	return error;
}

static int
pfkey_sens_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
        int error = 0;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_sens_process: Sorry, I can't process exttype=%d yet.\n",
		    pfkey_ext->sadb_ext_type);
        SENDERR(EINVAL); /* don't process these yet */
 errlab:
        return error;
}

static int
pfkey_prop_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
        int error = 0;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_prop_process: Sorry, I can't process exttype=%d yet.\n",
		    pfkey_ext->sadb_ext_type);
	SENDERR(EINVAL); /* don't process these yet */
	
 errlab:
	return error;
}

static int
pfkey_supported_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
        int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_supported_process: Sorry, I can't process exttype=%d yet.\n",
		    pfkey_ext->sadb_ext_type);
	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_spirange_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
        int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_spirange_process: .\n");
/* errlab: */
	return error;
}

static int
pfkey_x_kmprivate_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_x_kmprivate_process: Sorry, I can't process exttype=%d yet.\n",
		    pfkey_ext->sadb_ext_type);
	SENDERR(EINVAL); /* don't process these yet */

errlab:
	return error;
}

static int
pfkey_x_satype_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct sadb_x_satype *pfkey_x_satype = (struct sadb_x_satype *)pfkey_ext;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_x_satype_process: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_satype_process: "
			    "extr or extr->tdb is NULL, fatal\n");
		SENDERR(EINVAL);
	}

	if(pfkey_alloc_tdb(&(extr->tdb2)) == ENOMEM) {
		SENDERR(ENOMEM);
	}
	if(!(extr->tdb2->tdb_said.proto = satype2proto(pfkey_x_satype->sadb_x_satype_satype))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_satype_process: proto lookup from satype=%d failed.\n",
			    pfkey_x_satype->sadb_x_satype_satype);
		SENDERR(EINVAL);
	}

errlab:
	return error;
}

static int
pfkey_x_debug_process(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct sadb_x_debug *pfkey_x_debug = (struct sadb_x_debug *)pfkey_ext;

	if(!pfkey_x_debug) {
		printk("klips_debug:pfkey_x_debug_process: null pointer passed in\n");
		SENDERR(EINVAL);
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_debug_process: .\n");

#ifdef CONFIG_IPSEC_DEBUG
		if(pfkey_x_debug->sadb_x_debug_netlink >>
		   (sizeof(pfkey_x_debug->sadb_x_debug_netlink) * 8 - 1)) {
			pfkey_x_debug->sadb_x_debug_netlink &=
				~(1 << (sizeof(pfkey_x_debug->sadb_x_debug_netlink) * 8 -1));
			debug_tunnel  |= pfkey_x_debug->sadb_x_debug_tunnel;
			debug_netlink |= pfkey_x_debug->sadb_x_debug_netlink;
			debug_xform   |= pfkey_x_debug->sadb_x_debug_xform;
			debug_eroute  |= pfkey_x_debug->sadb_x_debug_eroute;
			debug_spi     |= pfkey_x_debug->sadb_x_debug_spi;
			debug_radij   |= pfkey_x_debug->sadb_x_debug_radij;
			debug_esp     |= pfkey_x_debug->sadb_x_debug_esp;
			debug_ah      |= pfkey_x_debug->sadb_x_debug_ah;
			debug_rcv     |= pfkey_x_debug->sadb_x_debug_rcv;
			debug_pfkey   |= pfkey_x_debug->sadb_x_debug_pfkey;
			if(debug_netlink)
				printk("klips_debug:pfkey_x_debug_process: set\n");
		} else {
			if(debug_netlink)
				printk("klips_debug:pfkey_x_debug_process: unset\n");
			debug_tunnel  &= pfkey_x_debug->sadb_x_debug_tunnel;
			debug_netlink &= pfkey_x_debug->sadb_x_debug_netlink;
			debug_xform   &= pfkey_x_debug->sadb_x_debug_xform;
			debug_eroute  &= pfkey_x_debug->sadb_x_debug_eroute;
			debug_spi     &= pfkey_x_debug->sadb_x_debug_spi;
			debug_radij   &= pfkey_x_debug->sadb_x_debug_radij;
			debug_esp     &= pfkey_x_debug->sadb_x_debug_esp;
			debug_ah      &= pfkey_x_debug->sadb_x_debug_ah;
			debug_rcv     &= pfkey_x_debug->sadb_x_debug_rcv;
			debug_pfkey   &= pfkey_x_debug->sadb_x_debug_pfkey;
		}
#else /* CONFIG_IPSEC_DEBUG */
		printk("klips_debug:pfkey_x_debug_process: debugging not enabled\n");
		SENDERR(EINVAL);
#endif /* CONFIG_IPSEC_DEBUG */
	
errlab:
	return error;
}


int
pfkey_tdb_init(struct tdb *tdbp, struct sadb_ext **extensions)
{
        int error = 0;
        char sa[SATOA_BUF];
	char ipaddr_txt[ADDRTOA_BUF];
	char ipaddr2_txt[ADDRTOA_BUF];

	if(!tdbp) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_tdb_init: "
			    "tdbp is NULL, fatal\n");
		SENDERR(EINVAL);
	}

	satoa(tdbp->tdb_said, 0, sa, SATOA_BUF);

        KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_tdb_init: (pfkey defined) called for SA:%s\n", sa);

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_tdb_init: calling init routine of %s%s%s\n",
		    TDB_XFORM_NAME(tdbp));
	
	switch(tdbp->tdb_said.proto) {
		
#ifdef CONFIG_IPSEC_IPIP
	case IPPROTO_IPIP: 
		addrtoa(((struct sockaddr_in*)(tdbp->tdb_addr_s))->sin_addr,
			0,
			ipaddr_txt, sizeof(ipaddr_txt));
		addrtoa(((struct sockaddr_in*)(tdbp->tdb_addr_d))->sin_addr,
			0,
			ipaddr2_txt, sizeof(ipaddr_txt));
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_tdb_init: "
			    "(pfkey defined) IPIP tdb set for %s->%s.\n",
			    ipaddr_txt,
			    ipaddr2_txt);
	break;
#endif /* !CONFIG_IPSEC_IPIP */

#ifdef CONFIG_IPSEC_AH
	case IPPROTO_AH:
		if (tdbp->tdb_authalg) {
			SData sKey;
  			CRYPTO_STATUS cs = CRYPTO_OK;
	  	    	tdbp->tdb_auth_bits = AH_AUTHBITS;
			
			/* internal encryption required, but the key is in plain */
			if ( CONFIG_INT_ENC && 
			     ((tdbp->tdb_key_bits_a / 8) == tdbp->tdb_key_a_size)
			   ) 
			{
			   if (plain_key_count >= CONFIG_MAX_PLAINKEYS) {
				   printk("Internal encryption required, max number of plain keys reached\n");
			   	   SENDERR(EINVAL);
			   }
			   else {
			   	sKey.nKEK = 0;
			   	plain_key_count++;
			   }
			}
			else {
			   if (tdbp->tdb_is_pepp) sKey.nKEK = CRYPTO_KEK_PEPP_SYM;
			   else sKey.nKEK = CONFIG_INT_ENC;	
			}
							
			sKey.pData = tdbp->tdb_key_a;
			sKey.nDataLength = tdbp->tdb_key_bits_a / 8;

			if (tdbp->tdb_flags & EMT_INBOUND) {	
                                int cnt;
				
				cnt = 1;
#ifdef CONFIG_SMP
				cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
				cnt = hsips_get_thread_cnt();
#endif
				if (cnt > 0) {
					int i;
					
					for (i = 0; i < cnt; i++) {
						if (tdbp->tdb_authalg_v_handle[i])
							crypto_close(tdbp->tdb_authalg_v_handle[i]);
						
						cs = crypto_open((crypto_handle *) &tdbp->tdb_authalg_v_handle[i],
							tdbp->tdb_authalg_v_num, &sKey);
						
						if (cs != CRYPTO_OK) {
							KLIPS_PRINT(debug_pfkey,
								"err opening auth gen alg #%d, err: %lu\n",
								tdbp->tdb_authalg_g_num, cs);
							SENDERR(EINVAL);
						}			    
					}
				} else
					printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");

			} else {
                                int cnt;
				
				cnt = 1;
#ifdef CONFIG_SMP
				cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
				cnt = hsips_get_thread_cnt();
#endif
				if (cnt > 0) {
					int i;

					for (i = 0; i < cnt; i++) {
						/* if we are in update, close old handle first */
						if (tdbp->tdb_authalg_g_handle[i]) 
							crypto_close(tdbp->tdb_authalg_g_handle[i]);
						
						cs = crypto_open((crypto_handle *) &tdbp->tdb_authalg_g_handle[i], 
								tdbp->tdb_authalg_g_num, &sKey);
						
						if (cs != CRYPTO_OK) {
							KLIPS_PRINT(debug_pfkey, 
								"error opening auth vrf alg #%d, err %lu\n",  
								tdbp->tdb_authalg_v_num, cs);
							SENDERR(EINVAL);
						} 
					}
				} else
					printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");
				
		    }
		} // endif (authalg)
		break;
	  
#endif /* CONFIG_IPSEC_AH */

#ifdef CONFIG_IPSEC_ESP
	case IPPROTO_ESP: {
		struct crypto_algorithm *ca;

	    	if ((tdbp->tdb_encalg) && (tdbp->tdb_encalg != ESP_NULL)) {
			ca = get_algorithm(tdbp->tdb_encalg_d_num);
			tdbp->tdb_iv_size = ca->info.milen;
			tdbp->tdb_iv = (caddr_t) kmalloc(tdbp->tdb_iv_size, GFP_ATOMIC);

			if(!tdbp->tdb_iv) {
				put_algorithm(ca);
				SENDERR(ENOMEM);
			}		
			
			get_random_bytes((void *)tdbp->tdb_iv, tdbp->tdb_iv_size);
			tdbp->tdb_iv_bits = tdbp->tdb_iv_size * 8;

			/* set encalg blocksize */
			tdbp->tdb_e_bs = ca->info.blocklen;
			put_algorithm(ca);
		}  // endif (encalg)
		
 	  	if (tdbp->tdb_encalg) {
			SData sKey;
			CRYPTO_STATUS cs = CRYPTO_OK;

			/* internal encryption required, but the key is in plain */
			if ( (CONFIG_INT_ENC) && 
			     (tdbp->tdb_key_bits_e / 8 == tdbp->tdb_key_e_size)
			   ) {
			   if (plain_key_count >= CONFIG_MAX_PLAINKEYS) {
				   printk("Internal encryption required, max number of plain keys reached\n");
			   	   SENDERR(EINVAL);
			   }
			   else {
			   	sKey.nKEK = 0;
			   	plain_key_count++;
			   }
			}
			else {
			   if (tdbp->tdb_is_pepp) sKey.nKEK = CRYPTO_KEK_PEPP_SYM;
			   else sKey.nKEK = CONFIG_INT_ENC;	
			}
		  
			sKey.pData = tdbp->tdb_key_e;
			sKey.nDataLength = tdbp->tdb_key_bits_e / 8;
		 
		  	if (tdbp->tdb_flags & EMT_INBOUND) {
                                int cnt;
				
				cnt = 1;
#ifdef CONFIG_SMP
				cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
				cnt = hsips_get_thread_cnt();
#endif
				if (cnt > 0) {
					int i;
					
					for (i = 0; i < cnt; i++) {
						if (tdbp->tdb_encalg_d_handle[i]) 
							crypto_close(tdbp->tdb_encalg_d_handle[i]);

						cs = crypto_open((crypto_handle *) &tdbp->tdb_encalg_d_handle[i], 
							tdbp->tdb_encalg_d_num, &sKey);
						if (cs != CRYPTO_OK) {
							KLIPS_PRINT(debug_pfkey,
								"error opening enc alg #%d\n",
								tdbp->tdb_encalg_d_num);
							SENDERR(EINVAL);
						} else {
							KLIPS_PRINT(debug_pfkey,
								"encalg D (%d) opened\n", 
								tdbp->tdb_encalg_d_num);
						}	
					}
				} else
					printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");
			} else {
                                int cnt;
				
				cnt = 1;
#ifdef CONFIG_SMP
				cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
				cnt = hsips_get_thread_cnt();
#endif
				if (cnt > 0) {
					int i;

					for (i = 0; i < cnt; i++) {
						if (tdbp->tdb_encalg_e_handle[i]) 
					    		crypto_close(tdbp->tdb_encalg_e_handle[i]);

			  			cs = crypto_open((crypto_handle *) &tdbp->tdb_encalg_e_handle[i], 
							tdbp->tdb_encalg_e_num, &sKey);
						if (cs != CRYPTO_OK) {
							KLIPS_PRINT(debug_pfkey, 
								"error opening enc alg #%d\n",  
								tdbp->tdb_encalg_e_num);
							SENDERR(EINVAL);
						} else {
							KLIPS_PRINT(debug_pfkey,
								"encalg E (%d) opened\n", 
								tdbp->tdb_encalg_e_num);
						} 
					}
				} else
					printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");
			}	
		} // endif (encalg)

	    	if (tdbp->tdb_authalg) {
			SData sKey;
			CRYPTO_STATUS cs = CRYPTO_OK;
		  
			/* internal encryption required, but the key is in plain */
			if ( (CONFIG_INT_ENC) && 
			     (tdbp->tdb_key_bits_a / 8 == tdbp->tdb_key_a_size)
			   ) {
			   if (plain_key_count >= CONFIG_MAX_PLAINKEYS) {
				   printk("Internal encryption required, max number of plain keys reached\n");
			   	   SENDERR(EINVAL);
			   }
			   else {
			   	sKey.nKEK = 0;
			   	plain_key_count++;
			   }
			}
			else {
			   if (tdbp->tdb_is_pepp) sKey.nKEK = CRYPTO_KEK_PEPP_SYM;
			   else sKey.nKEK = CONFIG_INT_ENC;	
			}
			tdbp->tdb_auth_bits = AH_AUTHBITS;

			sKey.pData = tdbp->tdb_key_a;
			sKey.nDataLength = tdbp->tdb_key_bits_a / 8;
		
			if (tdbp->tdb_flags & EMT_INBOUND) {	
                                int cnt;
				
				cnt = 1;
#ifdef CONFIG_SMP
				cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
				cnt = hsips_get_thread_cnt();
#endif
				if (cnt > 0) {
					int i;
					
					for (i = 0; i < cnt; i++) {
			    			if (tdbp->tdb_authalg_v_handle[i]) 
			    				crypto_close(tdbp->tdb_authalg_v_handle[i]); 

			    			cs = crypto_open((crypto_handle *) &tdbp->tdb_authalg_v_handle[i], 
							tdbp->tdb_authalg_v_num, &sKey);
						if (cs != CRYPTO_OK) {
			      				KLIPS_PRINT(debug_pfkey, 
								"error opening auth vrfy alg #%d, err: %lu\n",  
			 					tdbp->tdb_authalg_v_num, cs);
							SENDERR(EINVAL);
						} 
					}
				} else
					printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");
			} else {
                                int cnt;
				
				cnt = 1;
#ifdef CONFIG_SMP
				cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
				cnt = hsips_get_thread_cnt();
#endif
				if (cnt > 0) {
					int i;

					for (i = 0; i < cnt; i++) {
						if (tdbp->tdb_authalg_g_handle[i]) 
							crypto_close(tdbp->tdb_authalg_g_handle[i]);

					      	cs = crypto_open((crypto_handle *) &tdbp->tdb_authalg_g_handle[i], 
							tdbp->tdb_authalg_g_num, &sKey);
						if (cs != CRYPTO_OK) {
							KLIPS_PRINT(debug_pfkey, 
								"error opening auth gen alg #%d, err %lu\n",  
								tdbp->tdb_authalg_g_num, cs);
							SENDERR(EINVAL);
						}
					}
				} else 
					printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");
			}
		} // endif (authalg)
	   
		// combine, if we can
	    	if (tdbp->tdb_encalg && tdbp->tdb_authalg) {
			CRYPTO_HANDLE chs[2];
			CRYPTO_STATUS cs = CRYPTO_OK;

                        int cnt;

			cnt = 1;
#ifdef CONFIG_SMP
			cnt = smp_num_cpus;
#endif
#ifdef CONFIG_IPSECTHREAD
			cnt = hsips_get_thread_cnt();
#endif
									
			if (cnt > 0) {
				int i;
				for (i = 0; i < cnt; i++) {
					if (tdbp->tdb_i_combined_handle[i]) 
						crypto_close(tdbp->tdb_i_combined_handle[i]);
					if (tdbp->tdb_o_combined_handle[i]) 
						crypto_close(tdbp->tdb_o_combined_handle[i]);

					if (tdbp->tdb_flags & EMT_INBOUND) {
						chs[0] = (CRYPTO_HANDLE)tdbp->tdb_authalg_v_handle[i];
						chs[1] = (CRYPTO_HANDLE)tdbp->tdb_encalg_d_handle[i];
						cs = crypto_combine((crypto_handle *) &tdbp->tdb_i_combined_handle[i], 
							2, (crypto_handle *) &chs);
	 				        if (cs != CRYPTO_OK) {
			        			KLIPS_PRINT(debug_pfkey, 
								"error opening inbound combined alg, err %lu\n", cs);
							SENDERR(EINVAL);
						}
					} else {
						chs[0] = (CRYPTO_HANDLE)tdbp->tdb_encalg_e_handle[i];
						chs[1] = (CRYPTO_HANDLE)tdbp->tdb_authalg_g_handle[i];
						cs = crypto_combine((crypto_handle *) &tdbp->tdb_o_combined_handle[i], 
							2, (crypto_handle *) &chs);
	 				        if (cs != CRYPTO_OK) {
			        			KLIPS_PRINT(debug_pfkey, 
								"error opening outbound combined alg, err %lu\n", cs);
							SENDERR(EINVAL);
						}
					}
				}
			} else 
				printk(KERN_NOTICE "pfkey_tdb_init: no threads running\n");
		} 
	      
		break;
	} /* end case esp */
#endif /* CONFIG_IPSEC_ESP */
	  
	default:
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_tdb_init: "
			    "proto=%d unknown.\n",
			    tdbp->tdb_said.proto);
		SENDERR(EINVAL);
	} /* end switch proto */
	
errlab:
	return(error);
}


int
pfkey_safe_build(int error, struct sadb_ext *extensions[SADB_MAX+1])
{
	KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_safe_build: error=%d\n", error);
	if (!error) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_safe_build:"
			    "success.\n");
		return 1;
	} else {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_safe_build:"
			    "caught error %d\n", error);
		pfkey_extensions_free(extensions);
		return 0;
	}
}


int
pfkey_getspi_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0, i;
	ipsec_spi_t minspi = 256, maxspi = -1L;
	int found_avail = 0;
	struct tdb *tdbq;
	char sa[SATOA_BUF];
	struct sadb_ext *extensions_reply[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_reply = NULL;
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;

	for (i=0; i< SADB_EXT_MAX+1; i++) extensions_reply[i] = NULL;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_getspi_parse: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_getspi_parse: error, extr or extr->tdb pointer NULL\n");
		SENDERR(EINVAL);
	}

	if(extensions[SADB_EXT_SPIRANGE]) {
		minspi = ((struct sadb_spirange *)extensions[SADB_EXT_SPIRANGE])->sadb_spirange_min;
		maxspi = ((struct sadb_spirange *)extensions[SADB_EXT_SPIRANGE])->sadb_spirange_max;
	}

	if(maxspi == minspi) {
		extr->tdb->tdb_said.spi = maxspi;
		tdbq = __find_tdb(&(extr->tdb->tdb_said));

		if(tdbq) {
			satoa(extr->tdb->tdb_said, 0, sa, SATOA_BUF);
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug: pfkey_getspi_parse: EMT_GETSPI found an old Tunnel Descriptor Block\n"
				    "klips_debug:                for SA: %s, delete it first.\n", sa);
			SENDERR(EEXIST);
		} else {
			found_avail = 1;
		}
	} else {
		int i = 0;
		__u32 rand_val;
		while( ( i < (maxspi - minspi)) && !found_avail ) {
			get_random_bytes((void*) &(rand_val),
					 /* sizeof(extr->tdb->tdb_said.spi) */
					 ( ((maxspi - minspi) < (2^8))  ? 1 :
					   ( ((maxspi - minspi) < (2^16)) ? 2 :
					     ( ((maxspi - minspi) < (2^24)) ? 3 :
					   4 ) ) ) );
			extr->tdb->tdb_said.spi = htonl(minspi +
					      (rand_val %
					      (maxspi - minspi + 1)));
			i++;
			tdbq = __find_tdb(&(extr->tdb->tdb_said));
			if(!tdbq) {
				found_avail = 1;
			}
		}
	}

	satoa(extr->tdb->tdb_said, 0, sa, SATOA_BUF);

	if (!found_avail) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_getspi_parse: found an old Tunnel Descriptor Block\n"
			    "klips_debug:                for SA: %s, delete it first.\n", sa);
		SENDERR(EEXIST);
	}

	if(ip_chk_addr((unsigned long)extr->tdb->tdb_said.dst.s_addr) == IS_MYADDR) {
		extr->tdb->tdb_flags |= EMT_INBOUND;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_getspi_parse: existing Tunnel Descriptor Block not found (this\n"
		    "klips_debug:                is good) for SA: %s, %s-bound, allocating.\n",
		    sa, extr->tdb->tdb_flags & EMT_INBOUND ? "in" : "out");
	
	extr->tdb->tdb_rcvif = NULL;
	extr->tdb->tdb_lifetime_addtime_c = jiffies/HZ;

	extr->tdb->tdb_state = SADB_SASTATE_LARVAL;

	if(!extr->tdb->tdb_lifetime_allocations_c) {
		extr->tdb->tdb_lifetime_allocations_c += 1;
	}

	pfkey_extensions_init(extensions_reply);
	if(!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions_reply[0],
							  SADB_GETSPI,
							  satype,
							  0,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_seq,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_pid),
			      extensions_reply) &&
	     pfkey_safe_build(pfkey_sa_build(&extensions_reply[SADB_EXT_SA],
					     SADB_EXT_SA,
					     extr->tdb->tdb_said.spi,
					     0,
					     SADB_SASTATE_LARVAL,
					     0,
					     0,
					     0),
			      extensions_reply) &&
	     pfkey_safe_build(pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_SRC],
						  SADB_EXT_ADDRESS_SRC,
						  extr->tdb->tdb_said.proto,
						  0,
						  extr->tdb->tdb_addr_s),
			      extensions_reply) &&
	     pfkey_safe_build(pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_DST],
						  SADB_EXT_ADDRESS_DST,
						  extr->tdb->tdb_said.proto,
						  0,
						  extr->tdb->tdb_addr_d),
			      extensions_reply) )) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_getspi_parse: "
			    "failed to build the getspi reply message extensions\n");
		goto errlab;
	}
	
	if((error = insert_tdb(extr->tdb))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_getspi_parse: "
			    "failed to add the larval SA with error=%d.\n",
			    error);
		goto errlab;
	}
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_getspi_parse: successful for SA: %s\n", sa);
	
	if((error = pfkey_msg_build(&pfkey_reply, extensions_reply, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_getspi_parse: "
			    "failed to build the getspi reply message\n");
		goto errlab;
	}
	for(pfkey_socketsp = pfkey_open_sockets;
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_reply))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_getspi_parse: "
				    "sending up getspi reply message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_getspi_parse: "
			    "sending up getspi reply message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}

 errlab:
	if (pfkey_reply) {
		pfkey_msg_free(&pfkey_reply);
	}
	pfkey_extensions_free(extensions_reply);
	return error;
}

int
pfkey_update_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0, i;
	struct tdb* tdbq;
	char sa[SATOA_BUF];
	struct sadb_ext *extensions_reply[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_reply = NULL;
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;

	for (i=0; i< SADB_EXT_MAX+1; i++) extensions_reply[i] = NULL;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_update_parse: .\n");

	if(((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_state != SADB_SASTATE_MATURE) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_update_parse: "
			    "error, sa_state=%d must be MATURE=%d\n",
			    ((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_state,
			    SADB_SASTATE_MATURE);
		SENDERR(EINVAL);
	}

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_update_parse: error, extr or extr->tdb pointer NULL\n");
		SENDERR(EINVAL);
	}

	satoa(extr->tdb->tdb_said, 0, sa, SATOA_BUF);


	tdbq = __find_tdb(&(extr->tdb->tdb_said));
	if (!tdbq) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_update_parse: reserved Tunnel Descriptor Block\n"
			    "klips_debug:            for SA: %s not found.  Call SADB_GETSPI first or call SADB_ADD instead.\n", sa);
		SENDERR(EEXIST);
	}

	if(ip_chk_addr((unsigned long)extr->tdb->tdb_said.dst.s_addr) == IS_MYADDR) {
		extr->tdb->tdb_flags |= EMT_INBOUND;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug: pfkey_update_parse: existing Tunnel Descriptor Block found (this\n"
		    "klips_debug:                is good) for SA: %s, %s-bound, updating.\n",
		    sa, extr->tdb->tdb_flags & EMT_INBOUND ? "in" : "out");
	
	/* XXX extr->tdb->tdb_rcvif = &(enc_softc[em->em_if].enc_if);*/
	extr->tdb->tdb_rcvif = NULL;
	if ((error = pfkey_tdb_init(extr->tdb, extensions))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_update_parse: "
			    "not successful for SA: %s, deleting.\n", sa);
		ipsec_tdbwipe(extr->tdb);
		SENDERR(-error);
	}

	extr->tdb->tdb_lifetime_addtime_c = tdbq->tdb_lifetime_addtime_c;
	if((error = markdel_tdbchain(tdbq))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_update_parse: "
			    "error=%d, trouble deleting intermediate tdb for SA=%s.\n",
			    error, sa);
		SENDERR(-error);
	}


	pfkey_extensions_init(extensions_reply);
	if(!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions_reply[0],
							  SADB_UPDATE,
							  satype,
							  0,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_seq,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_pid),
			      extensions_reply)
	     && pfkey_safe_build(error = pfkey_sa_build(&extensions_reply[SADB_EXT_SA],
							SADB_EXT_SA,
							extr->tdb->tdb_said.spi,
							extr->tdb->tdb_replaywin,
							extr->tdb->tdb_state,
							extr->tdb->tdb_authalg,
							extr->tdb->tdb_encalg,
							extr->tdb->tdb_flags),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_CURRENT],
							      SADB_EXT_LIFETIME_CURRENT,
							      extr->tdb->tdb_lifetime_allocations_c,
							      extr->tdb->tdb_lifetime_bytes_c,
							      extr->tdb->tdb_lifetime_addtime_c,
							      extr->tdb->tdb_lifetime_usetime_c),
				 extensions)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_HARD],
							      SADB_EXT_LIFETIME_HARD,
							      extr->tdb->tdb_lifetime_allocations_h,
							      extr->tdb->tdb_lifetime_bytes_h,
							      extr->tdb->tdb_lifetime_addtime_h,
							      extr->tdb->tdb_lifetime_usetime_h),
				 extensions)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_SRC],
							     SADB_EXT_ADDRESS_SRC,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_s),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_DST],
							     SADB_EXT_ADDRESS_DST,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_d),
				 extensions_reply)
		)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_update_parse: "
			    "failed to build the update reply message extensions\n");
		goto errlab;
	}
		
	if((error = insert_tdb(extr->tdb))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_update_parse: "
			    "failed to add the mature SA with error=%d.\n",
			    error);
		goto errlab;
	}
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_update_parse: successful for SA: %s\n", sa);
	
	if((error = pfkey_msg_build(&pfkey_reply, extensions_reply, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_update_parse: "
			    "failed to build the update reply message\n");
		goto errlab;
	}
	for(pfkey_socketsp = pfkey_open_sockets;
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_reply))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_update_parse: "
				    "sending up update reply message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_update_parse: "
			    "sending up update reply message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}
	
 errlab:
	if (pfkey_reply) {
		pfkey_msg_free(&pfkey_reply);
	}
	pfkey_extensions_free(extensions_reply);
	return error;
}

int
pfkey_add_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct tdb* tdbq;
	char sa[SATOA_BUF];
	struct sadb_ext *extensions_reply[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_reply = NULL;
#ifndef CONFIG_IPSEC_PFKEY_LB
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;
#endif

	pfkey_extensions_init(extensions_reply);
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_add_parse: parsing add message\n");

	if(((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_state != SADB_SASTATE_MATURE) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: "
			    "error, sa_state=%d must be MATURE=%d\n",
			    ((struct sadb_sa*)extensions[SADB_EXT_SA])->sadb_sa_state,
			    SADB_SASTATE_MATURE);
		SENDERR(EINVAL);
	}

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: extr or extr->tdb pointer NULL\n");
		SENDERR(EINVAL);
	}

	satoa(extr->tdb->tdb_said, 0, sa, SATOA_BUF);

	tdbq = __find_tdb(&(extr->tdb->tdb_said));

	if (tdbq) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: found an old Tunnel Descriptor Block\n"
			    "klips_debug:                for SA: %s, delete it first.\n", sa);

                atomic_set(&tdbq->del_flag,1);
		__deloldtdbs();

	}

	if(ip_chk_addr((unsigned long)extr->tdb->tdb_said.dst.s_addr) == IS_MYADDR) {
		extr->tdb->tdb_flags |= EMT_INBOUND;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_add_parse: existing Tunnel Descriptor Block not found (this\n"
		    "klips_debug:                is good) for SA: %s, %s-bound, allocating.\n",
		    sa, extr->tdb->tdb_flags & EMT_INBOUND ? "in" : "out");
	
	extr->tdb->tdb_rcvif = NULL;
	
	if ((error = pfkey_tdb_init(extr->tdb, extensions))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_add_parse: not successful for SA: %s, deleting.\n", sa);
		ipsec_tdbwipe(extr->tdb);
		SENDERR(-error);
	}
	
	if (cug) {
		if ( (unsigned long)extr->tdb->tdb_said.dst.s_addr == 0) {
			int cnt;
		 	struct tdb *tdbp = extr->tdb;
			SData sKey;
			CRYPTO_STATUS cs = CRYPTO_OK;
			CRYPTO_HANDLE chs[2];

			printk("IKE CUG -> setting inbound SA as like as outbound\n");
			if (CONFIG_INT_ENC && (plain_key_count+2 > CONFIG_MAX_PLAINKEYS)) {
				printk("Internal encryption required, max number of plain keys reached\n");
				ipsec_tdbwipe(extr->tdb);
			   	SENDERR(EINVAL);
			}
	
			plain_key_count+=2;
			sKey.nKEK = 0;
			
			cnt = 1;
#  ifdef CONFIG_SMP
			cnt = smp_num_cpus;
#  endif
#  ifdef CONFIG_IPSECTHREAD
			cnt = hsips_get_thread_cnt();
#  endif
			
			if (cnt > 0) {
				int i;
				
				for (i = 0; i < cnt; i++) {
					/* initialize encalg_d_handles .... */
					sKey.pData = tdbp->tdb_key_e;
					sKey.nDataLength = tdbp->tdb_key_bits_e / 8;
					if (tdbp->tdb_encalg_d_handle[i]) 
						crypto_close(tdbp->tdb_encalg_d_handle[i]);
				
					cs = crypto_open((crypto_handle *) &tdbp->tdb_encalg_d_handle[i], 
						tdbp->tdb_encalg_d_num, &sKey);
					if (cs != CRYPTO_OK) {
						KLIPS_PRINT(debug_pfkey,
							"error opening enc alg #%d\n",
							tdbp->tdb_encalg_d_num);
						SENDERR(EINVAL);
					}

					/* initialize authalg_v_handles .... */
					sKey.pData = tdbp->tdb_key_a;
					sKey.nDataLength = tdbp->tdb_key_bits_a / 8;
					if (tdbp->tdb_authalg_v_handle[i]) 
						crypto_close(tdbp->tdb_authalg_v_handle[i]); 
	
					cs = crypto_open((crypto_handle *) &tdbp->tdb_authalg_v_handle[i], 
						tdbp->tdb_authalg_v_num, &sKey);
					if (cs != CRYPTO_OK) {
						KLIPS_PRINT(debug_pfkey, 
							"error opening auth vrfy alg #%d, err: %lu\n",  
							tdbp->tdb_authalg_v_num, cs);
						SENDERR(EINVAL);
					}

					/* initialize inbound_combined_handles .... */
					if (tdbp->tdb_i_combined_handle[i]) 
						crypto_close(tdbp->tdb_i_combined_handle[i]);

					chs[0] = (CRYPTO_HANDLE)tdbp->tdb_authalg_v_handle[i];
					chs[1] = (CRYPTO_HANDLE)tdbp->tdb_encalg_d_handle[i];
					cs = crypto_combine((crypto_handle *) &tdbp->tdb_i_combined_handle[i], 
						2, (crypto_handle *) &chs);
					if (cs != CRYPTO_OK) {
						KLIPS_PRINT(debug_pfkey, 
						"error opening inbound combined alg, err %lu\n", cs);
						SENDERR(EINVAL);
					}
				}
			} else 
				printk(KERN_NOTICE "pfkey_add_parse: no threads running\n");

		       /* don't care about replay for IKE SA - must be like with manual keying */
		       extr->tdb->tdb_replaywin = 0;
		}
	} // endif (cug)

	extr->tdb->tdb_lifetime_addtime_c = jiffies / HZ;
	if(!extr->tdb->tdb_lifetime_allocations_c) {
		extr->tdb->tdb_lifetime_allocations_c += 1;
	}

	if((error = insert_tdb(extr->tdb))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_add_parse: "
			    "failed to add the mature SA with error=%d.\n",
			    error);
		goto errlab;
	}
	
#ifndef CONFIG_IPSEC_PFKEY_LB      // do not build own PFKEY replies when CONFIG_IPSEC_PFKEY_LB is set

	if(!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions_reply[0],
							  SADB_ADD,
							  satype,
							  0,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_seq,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_pid),
			      extensions_reply)
	     && pfkey_safe_build(error = pfkey_sa_build(&extensions_reply[SADB_EXT_SA],
							SADB_EXT_SA,
							extr->tdb->tdb_said.spi,
							extr->tdb->tdb_replaywin,
							extr->tdb->tdb_state,
							extr->tdb->tdb_authalg,
							extr->tdb->tdb_encalg,
							extr->tdb->tdb_flags),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions_reply[SADB_EXT_LIFETIME_HARD],
							      SADB_EXT_LIFETIME_HARD,
							      extr->tdb->tdb_lifetime_allocations_h,
							      extr->tdb->tdb_lifetime_bytes_h,
							      extr->tdb->tdb_lifetime_addtime_h,
							      extr->tdb->tdb_lifetime_usetime_h),
				 extensions)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions_reply[SADB_EXT_LIFETIME_SOFT],
							      SADB_EXT_LIFETIME_SOFT,
							      extr->tdb->tdb_lifetime_allocations_s,
							      extr->tdb->tdb_lifetime_bytes_s,
							      extr->tdb->tdb_lifetime_addtime_s,
							      extr->tdb->tdb_lifetime_usetime_s),
				 extensions)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_SRC],
							     SADB_EXT_ADDRESS_SRC,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_s),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_DST],
							     SADB_EXT_ADDRESS_DST,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_d),
				 extensions_reply)
		)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_add_parse: "
			    "failed to build the add reply message extensions\n");
		goto errlab;
	}
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_add_parse: successful for SA: %s\n", sa);
	
	
	if((error = pfkey_msg_build(&pfkey_reply, extensions_reply, EXT_BITS_OUT))) {
	  KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_add_parse: "
			    "failed to build the add reply message\n");
		goto errlab;
	}

	for(pfkey_socketsp = pfkey_open_sockets;
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_reply))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_add_parse: "
				    "sending up add reply message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_add_parse: "
			    "sending up add reply message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}

#endif // PFKEY non conform behavior

 errlab:
	if (pfkey_reply) {
		pfkey_msg_free(&pfkey_reply);
	}
	pfkey_extensions_free(extensions_reply);

	return error;
}

int
pfkey_delete_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	struct tdb *tdbp;
	char sa[SATOA_BUF];
	int error = 0, i;
	struct sadb_ext *extensions_reply[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_reply = NULL;
#ifndef CONFIG_IPSEC_PFKEY_LB
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;
#endif

	for (i=0; i< SADB_EXT_MAX+1; i++) extensions_reply[i] = NULL;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_delete_parse: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_delete_parse: "
			    "extr or extr->tdb pointer NULL, fatal\n");
		SENDERR(EINVAL);
	}

	satoa(extr->tdb->tdb_said, 0, sa, SATOA_BUF);

	tdbp = __find_tdb(&(extr->tdb->tdb_said));

	if (tdbp == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_delete_parse: "
			    "Tunnel Descriptor Block not found for SA:%s, could not delete.\n",
			    sa);
		SENDERR(ESRCH);
	}

	if((error = markdel_tdbchain(tdbp))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_delete_parse: "
			    "error=%d returned trying to delete Tunnel Descriptor Block for SA:%s.\n",
			    error, sa);
		SENDERR(-error);
	}

#ifndef CONFIG_IPSEC_PFKEY_LB

	pfkey_extensions_init(extensions_reply);
	if(!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions_reply[0],
							  SADB_DELETE,
							  satype,
							  0,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_seq,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_pid),
			      extensions_reply)
	     && pfkey_safe_build(error = pfkey_sa_build(&extensions_reply[SADB_EXT_SA],
							SADB_EXT_SA,
							extr->tdb->tdb_said.spi,
							0,
							0,
							0,
							0,
							0),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_SRC],
							     SADB_EXT_ADDRESS_SRC,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_s),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_DST],
							     SADB_EXT_ADDRESS_DST,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_d),
				 extensions_reply)
		)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_delete_parse: "
			    "failed to build the delete reply message extensions\n");
		goto errlab;
	}
	
	if((error = pfkey_msg_build(&pfkey_reply, extensions_reply, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_delete_parse: "
			    "failed to build the delete reply message\n");
		goto errlab;
	}
	
	for(pfkey_socketsp = pfkey_open_sockets;
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_reply))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_delete_parse: "
				    "sending up delete reply message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_delete_parse: "
			    "sending up delete reply message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}

#endif  

 errlab:
	ipsec_tdbwipe(extr->tdb);

	if (pfkey_reply) {
		pfkey_msg_free(&pfkey_reply);
	}
	pfkey_extensions_free(extensions_reply);

	/* get rid of orphaned tdbs ... */
	__deloldtdbs();
	return error;
}

int
pfkey_get_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0, i;
	struct tdb *tdbp;
	char sa[SATOA_BUF];
	struct sadb_ext *extensions_reply[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_reply = NULL;

	for (i=0; i< SADB_EXT_MAX+1; i++) extensions_reply[i] = NULL;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_get_parse: .\n");

	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_get_parse: "
			    "extr or extr->tdb pointer NULL, fatal\n");
		SENDERR(EINVAL);
	}

	satoa(extr->tdb->tdb_said, 0, sa, SATOA_BUF);


	tdbp = __find_tdb(&(extr->tdb->tdb_said));
	if (tdbp == NULL) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_get_parse: "
			    "Tunnel Descriptor Block not found for SA=%s, could not get.\n",
			    sa);
		SENDERR(ESRCH);
	}
	
	pfkey_extensions_init(extensions_reply);
	if(!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions_reply[0],
							  SADB_GET,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype,
							  0,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_seq,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_pid),
			      extensions_reply)
	     && pfkey_safe_build(error = pfkey_sa_build(&extensions_reply[SADB_EXT_SA],
							SADB_EXT_SA,
							extr->tdb->tdb_said.spi,
							extr->tdb->tdb_replaywin,
							extr->tdb->tdb_state,
							extr->tdb->tdb_authalg,
							extr->tdb->tdb_encalg,
							extr->tdb->tdb_flags),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_CURRENT],
							      SADB_EXT_LIFETIME_CURRENT,
							      tdbp->tdb_lifetime_allocations_c,
							      tdbp->tdb_lifetime_bytes_c,
							      tdbp->tdb_lifetime_addtime_c,
							      tdbp->tdb_lifetime_usetime_c),
				 extensions)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_HARD],
							      SADB_EXT_LIFETIME_HARD,
							      tdbp->tdb_lifetime_allocations_h,
							      tdbp->tdb_lifetime_bytes_h,
							      tdbp->tdb_lifetime_addtime_h,
							      tdbp->tdb_lifetime_usetime_h),
				 extensions)
	     && pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_SOFT],
							      SADB_EXT_LIFETIME_SOFT,
							      tdbp->tdb_lifetime_allocations_s,
							      tdbp->tdb_lifetime_bytes_s,
							      tdbp->tdb_lifetime_addtime_s,
							      tdbp->tdb_lifetime_usetime_s),
				 extensions)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_SRC],
							     SADB_EXT_ADDRESS_SRC,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_s),
				 extensions_reply)
	     && pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_DST],
							     SADB_EXT_ADDRESS_DST,
							     extr->tdb->tdb_said.proto,
							     0,
							     extr->tdb->tdb_addr_d),
				 extensions_reply)
	     && extr->tdb->tdb_addr_d ? pfkey_safe_build(error = pfkey_address_build(&extensions_reply[SADB_EXT_ADDRESS_PROXY],
										     SADB_EXT_ADDRESS_PROXY,
										     extr->tdb->tdb_said.proto,
										     0,
										     extr->tdb->tdb_addr_p),
							 extensions_reply) : 1
		)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_get_parse: "
			    "failed to build the get reply message extensions\n");
		goto errlab;
	}
		
	
	if((error = pfkey_msg_build(&pfkey_reply, extensions_reply, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_get_parse: "
			    "failed to build the get reply message\n");
		goto errlab;
	}
	
	if((error = pfkey_upmsg(sk->socket, pfkey_reply))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_get_parse: "
			    "failed to send the get reply message\n");
		goto errlab;
	}
	
	KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_get_parse: "
		    "succeeded in sending get reply message.\n");
	
 errlab:
	if (pfkey_reply) {
		pfkey_msg_free(&pfkey_reply);
	}
	pfkey_extensions_free(extensions_reply);
	return error;
}

int
pfkey_acquire_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_acquire_parse: .\n");

	if((satype == 0) || (satype > SADB_SATYPE_MAX)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_acquire_parse: "
			    "SATYPE=%d invalid.\n",
			    satype);
		SENDERR(EINVAL);
	}

	if(!(pfkey_registered_sockets[satype])) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire_parse: "
			    "no sockets registered for SAtype=%d.\n",
			    satype);
		SENDERR(EPROTONOSUPPORT);
	}

	for(pfkey_socketsp = pfkey_registered_sockets[satype];
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp,
					((struct sadb_msg*)extensions[SADB_EXT_RESERVED])))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire_parse: "
				    "sending up acquire reply message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire_parse: "
			    "sending up acquire reply message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}
	
 errlab:
	return error;
}

int
pfkey_register_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	unsigned int alg_num_a = 0, alg_num_e = 0;
	struct sadb_alg *alg_a = NULL, *alg_e = NULL, *alg_ap = NULL, *alg_ep = NULL;
	struct sadb_ext *extensions_reply[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_reply = NULL;
	struct supported_list *pfkey_supported_listp;
	struct socket_list *pfkey_socketsp;
	int error = 0;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_register_parse: .\n");

	pfkey_extensions_init(extensions_reply);

	if((satype == 0) || (satype > SADB_SATYPE_MAX)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_register_parse: "
			    "SATYPE=%d invalid.\n",
			    satype);
		SENDERR(EINVAL);
	}

	if(!pfkey_list_insert_socket(sk->socket,
				 &(pfkey_registered_sockets[satype]))) {
		KLIPS_PRINT(1 || debug_pfkey,
			    "klips_debug:pfkey_register_parse: "
			    "SATYPE=%02d successfully registered by KMd on pid=%d.\n",
			    satype, key_pid(sk));
	};
	
	/* send up register msg with supported SATYPE algos */
	pfkey_supported_listp = pfkey_supported_list[satype];
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_register_parse: "
		    "pfkey_supported_list[%d]=%p\n",
		    satype,
		    pfkey_supported_list[satype]);
	while(pfkey_supported_listp) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_register_parse: "
			    "checking supported=%p\n",
			    pfkey_supported_listp);
		if(pfkey_supported_listp->supportedp->supported_alg_exttype == SADB_EXT_SUPPORTED_AUTH) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_register_parse: "
				    "adding auth alg.\n");
			alg_num_a++;
		}
		if(pfkey_supported_listp->supportedp->supported_alg_exttype == SADB_EXT_SUPPORTED_ENCRYPT) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_register_parse: "
				    "adding encrypt alg.\n");
			alg_num_e++;
		}
		pfkey_supported_listp = pfkey_supported_listp->next;
	}
	
	if(alg_num_a) {
		if((alg_a = kmalloc(alg_num_a * sizeof(struct sadb_alg), GFP_ATOMIC) ) == NULL) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_register_parse: auth alg memory allocation error\n");
			SENDERR(ENOMEM);
		}
		alg_ap = alg_a;
	}
	
	if(alg_num_e) {
		if((alg_e = kmalloc(alg_num_e * sizeof(struct sadb_alg), GFP_ATOMIC) ) == NULL) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_register_parse: enc alg memory allocation error\n");
			SENDERR(ENOMEM);
		}
		alg_ep = alg_e;
	}
	
	pfkey_supported_listp = pfkey_supported_list[satype];
	while(pfkey_supported_listp) {
		if(alg_num_a) {
			if(pfkey_supported_listp->supportedp->supported_alg_exttype == SADB_EXT_SUPPORTED_AUTH) {
				alg_ap->sadb_alg_id = pfkey_supported_listp->supportedp->supported_alg_id;
				alg_ap->sadb_alg_ivlen = pfkey_supported_listp->supportedp->supported_alg_ivlen;
				alg_ap->sadb_alg_minbits = pfkey_supported_listp->supportedp->supported_alg_minbits;
				alg_ap->sadb_alg_maxbits = pfkey_supported_listp->supportedp->supported_alg_maxbits;
				alg_ap->sadb_alg_reserved = 0;
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_register_parse: "
					    "adding auth=%p\n",
					    alg_ap);
				alg_ap++;
			}
		}
		if(alg_num_e) {
			if(pfkey_supported_listp->supportedp->supported_alg_exttype == SADB_EXT_SUPPORTED_ENCRYPT) {
				alg_ep->sadb_alg_id = pfkey_supported_listp->supportedp->supported_alg_id;
				alg_ep->sadb_alg_ivlen = pfkey_supported_listp->supportedp->supported_alg_ivlen;
				alg_ep->sadb_alg_minbits = pfkey_supported_listp->supportedp->supported_alg_minbits;
				alg_ep->sadb_alg_maxbits = pfkey_supported_listp->supportedp->supported_alg_maxbits;
				alg_ep->sadb_alg_reserved = 0;
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_register_parse: "
					    "adding encrypt=%p\n",
					    alg_ep);
				alg_ep++;
			}
		}
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_register_parse: "
			    "found satype=%d exttype=%d id=%d ivlen=%d minbits=%d maxbits=%d.\n",
			    satype,
			    pfkey_supported_listp->supportedp->supported_alg_exttype,
			    pfkey_supported_listp->supportedp->supported_alg_id,
			    pfkey_supported_listp->supportedp->supported_alg_ivlen,
			    pfkey_supported_listp->supportedp->supported_alg_minbits,
			    pfkey_supported_listp->supportedp->supported_alg_maxbits);
		pfkey_supported_listp = pfkey_supported_listp->next;
	}
	
	if(!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions_reply[0],
							  SADB_REGISTER,
							  satype,
							  0,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_seq,
							  ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_pid),
			      extensions_reply) &&
	     (alg_num_a ? pfkey_safe_build(error = pfkey_supported_build(&extensions_reply[SADB_EXT_SUPPORTED_AUTH],
									SADB_EXT_SUPPORTED_AUTH,
									alg_num_a,
									alg_a),
					  extensions_reply) : 1) &&
	     (alg_num_e ? pfkey_safe_build(error = pfkey_supported_build(&extensions_reply[SADB_EXT_SUPPORTED_ENCRYPT],
									SADB_EXT_SUPPORTED_ENCRYPT,
									alg_num_e,
									alg_e),
					  extensions_reply) : 1))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_register_parse: "
			    "failed to build the register message extensions\n");
		goto errlab;
	}
	
	if((error = pfkey_msg_build(&pfkey_reply, extensions_reply, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_register_parse: "
			    "failed to build the register message\n");
		goto errlab;
	}
	for(pfkey_socketsp = pfkey_registered_sockets[satype];
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_reply))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_register_parse: "
				    "sending up register reply message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_register_parse: "
			    "sending up register reply message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}
	
 errlab:
	if (pfkey_reply) {
		pfkey_msg_free(&pfkey_reply);
	}
	pfkey_extensions_free(extensions_reply);
	return error;
}


int
pfkey_expire_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_expire_parse: .\n");

	if(pfkey_open_sockets) {
		for(pfkey_socketsp = pfkey_open_sockets;
		    pfkey_socketsp;
		    pfkey_socketsp = pfkey_socketsp->next) {
			if((error = pfkey_upmsg(pfkey_socketsp->socketp,
						((struct sadb_msg*)extensions[SADB_EXT_RESERVED])))) {
				KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire_parse: "
					    "sending up expire reply message for satype=%d to socket=%p failed with error=%d.\n",
					    satype, pfkey_socketsp->socketp, error);
				goto errlab;
			}
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire_parse: "
				    "sending up expire reply message for satype=%d to socket=%p succeeded.\n",
				    satype, pfkey_socketsp->socketp);
		}
	}

 errlab:
	return error;
}

int
pfkey_flush_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;
	struct socket_list *pfkey_socketsp;
	uint8_t satype = ((struct sadb_msg*)extensions[SADB_EXT_RESERVED])->sadb_msg_satype;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_flush_parse: "
		    "flushing type %d SAs\n",
		    satype);

	if ((error = ipsec_tdbcleanup(satype)))
		SENDERR(-error);

	if(pfkey_open_sockets) {
		for(pfkey_socketsp = pfkey_open_sockets;
		    pfkey_socketsp;
		    pfkey_socketsp = pfkey_socketsp->next) {
			if((error = pfkey_upmsg(pfkey_socketsp->socketp,
						((struct sadb_msg*)extensions[SADB_EXT_RESERVED])))) {
				KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_flush_parse: "
					    "sending up flush reply message for satype=%d to socket=%p failed with error=%d.\n",
					    satype,
					    pfkey_socketsp->socketp,
					    error);
				goto errlab;
			}
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_flush_parse: "
				    "sending up flush reply message for satype=%d to socket=%p succeeded.\n",
				    satype,
				    pfkey_socketsp->socketp);
		}
	}

 errlab:
	return error;
}

int
pfkey_dump_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_dump_parse: .\n");

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_x_promisc_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_promisc_parse: .\n");

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_x_pchange_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_pchange_parse: .\n");

	SENDERR(ENOSYS);
 errlab:
	return error;
}

int
pfkey_x_grpsa_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	struct tdb *tdb1p, *tdb2p, *tdbp;
	struct tdb *tdb1s=NULL, *tdb2s=NULL;
	char sa1[SATOA_BUF], sa2[SATOA_BUF];
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_grpsa_parse: .\n");


	if(!extr || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_grpsa_parse: "
			    "extr or extr->tdb is NULL, fatal.\n");
		SENDERR(EINVAL);
	}

	satoa(extr->tdb->tdb_said, 0, sa1, SATOA_BUF);
	if(extr->tdb2) {
		satoa(extr->tdb2->tdb_said, 0, sa2, SATOA_BUF);
	}

	if(!(tdb1s = tdb1p = __find_tdb(&(extr->tdb->tdb_said)))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug: pfkey_x_grpsa_parse: reserved Tunnel Descriptor Block\n"
			    "klips_debug:            for SA: %s not found.  Call SADB_ADD/UPDATE first.\n", sa1);
		SENDERR(EEXIST);
	}
	if(extr->tdb2) { /* GRPSA */
		if(!(tdb2s = tdb2p = __find_tdb(&(extr->tdb2->tdb_said)))) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug: pfkey_x_grpsa_parse: reserved Tunnel Descriptor Block\n"
				    "klips_debug:            for SA: %s not found.  Call SADB_ADD/UPDATE first.\n", sa2);
			SENDERR(EEXIST);
		}

		/* Is either one already linked? */
		if(tdb1p->tdb_onext) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug: pfkey_x_grpsa_parse: Tunnel Descriptor Block\n"
				    "klips_debug:                for SA: %s is already linked.\n", sa1);
			SENDERR(EEXIST);
		}
		if(tdb2p->tdb_inext) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug: pfkey_x_grpsa_parse: Tunnel Descriptor Block\n"
				    "klips_debug:                for SA: %s is already linked.\n", sa2);
			SENDERR(EEXIST);
		}
		
		/* Is extr->tdb already linked to extr->tdb2? */
		tdbp = tdb2p;
		while(tdbp) {
			if(tdbp == tdb1p) {
				SENDERR(EEXIST);
			}
			tdbp = tdb2p->tdb_onext;
		}
		
		/* link 'em */
		tdb1p->tdb_onext = tdb2p;
		tdb2p->tdb_inext = tdb1p;
	} else { /* UNGRPSA */
		while(tdb1p->tdb_onext) {
			tdb1p = tdb1p->tdb_onext;
		}
		while(tdb1p->tdb_inext) {
			tdbp = tdb1p;
			tdb1p = tdb1p->tdb_inext;
			tdbp->tdb_inext = NULL;
			tdb1p->tdb_onext = NULL;
		}
	}

 errlab:

	ipsec_tdbwipe(extr->tdb);
	if(extr->tdb2) {
		ipsec_tdbwipe(extr->tdb2);
	}

	return error;
}

int
pfkey_x_addflow_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;
#ifdef CONFIG_IPSEC_DEBUG
	char buf1[64], buf2[64];
#endif /* CONFIG_IPSEC_DEBUG */

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_addflow_parse: .\n");

	if(!extr || !extr->tdb || !extr->eroute) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_addflow_parse: "
			    "missing extr, tdb or eroute data.\n");
		SENDERR(EINVAL);
	}

#ifdef CONFIG_IPSEC_DEBUG
	if (debug_pfkey) {
		subnettoa(extr->eroute->er_eaddr.sen_ip_src,
			  extr->eroute->er_emask.sen_ip_src, 0, buf1, sizeof(buf1));
		subnettoa(extr->eroute->er_eaddr.sen_ip_dst,
			  extr->eroute->er_emask.sen_ip_dst, 0, buf2, sizeof(buf2));
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_addflow_parse: "
			    "calling breakeroute and/or makeroute for %s->%s\n",
			    buf1, buf2);
	}
#endif /* CONFIG_IPSEC_DEBUG */
	if(extr->tdb->tdb_flags & SADB_X_SAFLAGS_REPLACEFLOW) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_addflow_parse: "
			    "REPLACEFLOW flag set, calling breakeroute.\n");
		if ((error = ipsec_breakroute(&(extr->eroute->er_eaddr),
					      &(extr->eroute->er_emask))) == EINVAL) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_x_addflow_parse: "
				    "breakeroute returned %d.\n", error);
			SENDERR(-error);
		}
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_addflow_parse: "
		    "calling makeroute.\n");

	if ((error = ipsec_makeroute(&(extr->eroute->er_eaddr),
				     &(extr->eroute->er_emask),
				     extr->tdb->tdb_said))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_addflow_parse: "
			    "makeroute returned %d.\n", error);
		SENDERR(-error);
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_addflow_parse: "
		    "makeroute call successful.\n");

	ipsec_tdbwipe(extr->tdb);

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_addflow_parse: "
		    "extr->tdb cleaned up and freed.\n");

 errlab:
	return error;
}

int
pfkey_x_delflow_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;
#ifdef CONFIG_IPSEC_DEBUG
	char buf1[64], buf2[64];
#endif /* CONFIG_IPSEC_DEBUG */

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_delflow_parse: .\n");

	if(!extr || !extr->eroute || !extr->tdb) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_delflow_parse: "
			    "extr, extr->tdb or extr->eroute is NULL, fatal\n");
		SENDERR(EINVAL);
	}

	if(extr->tdb->tdb_flags & SADB_X_SAFLAGS_CLEARFLOW) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_x_delflow_parse: "
			    "CLEARFLOW flag set, calling cleareroutes.\n");
		if ((error = ipsec_cleareroutes()))
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_x_delflow_parse: "
				    "cleareroutes returned %d.\n", error);
			SENDERR(-error);
	} else {
#ifdef CONFIG_IPSEC_DEBUG
		if (debug_pfkey) {
			subnettoa(extr->eroute->er_eaddr.sen_ip_src,
				  extr->eroute->er_emask.sen_ip_src, 0, buf1, sizeof(buf1));
			subnettoa(extr->eroute->er_eaddr.sen_ip_dst,
				  extr->eroute->er_emask.sen_ip_dst, 0, buf2, sizeof(buf2));
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_x_delflow_parse: "
				    "calling breakeroute for %s->%s\n",
				    buf1, buf2);
		}
#endif /* CONFIG_IPSEC_DEBUG */

		if((error = ipsec_breakroute(&(extr->eroute->er_eaddr),
					     &(extr->eroute->er_emask)))) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_x_delflow_parse: "
				    "breakeroute returned %d.\n", error);
			SENDERR(-error);
		}
	}
	
	ipsec_tdbwipe(extr->tdb);

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_delflow_parse: "
		    "extr->tdb cleaned up and freed.\n");

 errlab:
	return error;
}

int
pfkey_x_msg_debug_parse(struct sock *sk, struct sadb_ext **extensions, struct pfkey_extracted_data* extr)
{
	int error = 0;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_x_msg_debug_parse: .\n");

/* errlab:*/
	return error;
}


int
pfkey_expire(struct tdb *tdbp, int hard)
{
	struct sadb_ext *extensions[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_msg = NULL;
	struct socket_list *pfkey_socketsp;
	int error = 0;
	uint8_t satype = proto2satype(tdbp->tdb_said.proto);

	pfkey_extensions_init(extensions);

	if(!pfkey_open_sockets) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire: "
			    "no sockets listening.\n");
		printk("no registered sockets\n");
		goto errlab;
	}

	
	if (!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions[0],
							   SADB_EXPIRE,
							   satype,
							   0,
							   ++pfkey_msg_seq,
							   0),
			       extensions)
	      && pfkey_safe_build(error = pfkey_sa_build(&extensions[SADB_EXT_SA],
							 SADB_EXT_SA,
							 tdbp->tdb_said.spi,
							 tdbp->tdb_replaywin,
							 tdbp->tdb_state,
							 tdbp->tdb_authalg,
							 tdbp->tdb_encalg,
							 tdbp->tdb_flags),
				  extensions)

	      && pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_CURRENT],
							       SADB_EXT_LIFETIME_CURRENT,
							       tdbp->tdb_lifetime_allocations_c,
							       tdbp->tdb_lifetime_bytes_c,
							       tdbp->tdb_lifetime_addtime_c,
							       tdbp->tdb_lifetime_usetime_c),
				  extensions)
	      && (hard ? 
		  pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_HARD],
								SADB_EXT_LIFETIME_HARD,
								tdbp->tdb_lifetime_allocations_h,
								tdbp->tdb_lifetime_bytes_h,
								tdbp->tdb_lifetime_addtime_h,
								tdbp->tdb_lifetime_usetime_h),
				   extensions)
		  : pfkey_safe_build(error = pfkey_lifetime_build(&extensions[SADB_EXT_LIFETIME_SOFT],
								  SADB_EXT_LIFETIME_SOFT,
								  tdbp->tdb_lifetime_allocations_s,
								  tdbp->tdb_lifetime_bytes_s,
								  tdbp->tdb_lifetime_addtime_s,
								  tdbp->tdb_lifetime_usetime_s),
				     extensions))
	      && pfkey_safe_build(error = pfkey_address_build(&extensions[SADB_EXT_ADDRESS_SRC],
							      SADB_EXT_ADDRESS_SRC,
							      tdbp->tdb_said.proto,
							      0,
							      tdbp->tdb_addr_s),
				  extensions)
	      && pfkey_safe_build(error = pfkey_address_build(&extensions[SADB_EXT_ADDRESS_DST],
							      SADB_EXT_ADDRESS_DST,
							      tdbp->tdb_said.proto,
							      0,
							      tdbp->tdb_addr_d),
				  extensions) 

				   )) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire: failed to "
			    "build the expire message extensions\n");
		goto errlab;
	}
	

	if ((error = pfkey_msg_build(&pfkey_msg, extensions, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire: failed to "
			    "build the expire message\n");
		goto errlab;
	}
	
	for(pfkey_socketsp = pfkey_open_sockets;
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_msg))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire: "
				    "sending up expire message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_expire: "
			    "sending up expire message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}
	
 errlab:
	if (pfkey_msg) {
		pfkey_msg_free(&pfkey_msg);
	}
	pfkey_extensions_free(extensions);
	return error;
}

int
pfkey_acquire(struct tdb *tdbp)
{
	struct sadb_ext *extensions[SADB_EXT_MAX+1];
	struct sadb_msg *pfkey_msg = NULL;
	struct socket_list *pfkey_socketsp;
	int error = 0;
	struct sadb_comb comb[] = {
		/* auth; encrypt; flags; */
		/* auth_minbits; auth_maxbits; encrypt_minbits; encrypt_maxbits; */
		/* reserved; soft_allocations; hard_allocations; soft_bytes; hard_bytes; */
		/* soft_addtime; hard_addtime; soft_usetime; hard_usetime; */
		{ SADB_AALG_MD5HMAC,  SADB_EALG_3DESCBC, SADB_SAFLAGS_PFS,
		  128, 128, 168, 168,
		  0, 0, 0, 0, 0,
		  57600, 86400, 57600, 86400 },
		{ SADB_AALG_SHA1HMAC, SADB_EALG_3DESCBC, SADB_SAFLAGS_PFS,
		  160, 160, 168, 168,
		  0, 0, 0, 0, 0,
		  57600, 86400, 57600, 86400 }
	};
       
	uint8_t satype = SADB_SATYPE_ESP;

	pfkey_extensions_init(extensions);

	if((satype == 0) || (satype > SADB_SATYPE_MAX)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire: "
			    "SAtype=%d unspecified or unknown.\n",
			    satype);
		SENDERR(EINVAL);
	}

	if(!(pfkey_registered_sockets[satype])) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire: "
			    "no sockets registered for SAtype=%d.\n",
			    satype);
		SENDERR(EPROTONOSUPPORT);
	}
	
	if (!(pfkey_safe_build(error = pfkey_msg_hdr_build(&extensions[0],
							  SADB_ACQUIRE,
							  satype,
							  0,
							  ++pfkey_msg_seq,
							  0),
			      extensions)
	      && pfkey_safe_build(error = pfkey_address_build(&extensions[SADB_EXT_ADDRESS_SRC],
							      SADB_EXT_ADDRESS_SRC,
							      tdbp->tdb_said.proto,
							      0,
							      tdbp->tdb_addr_s),
				  extensions)
	      && pfkey_safe_build(error = pfkey_address_build(&extensions[SADB_EXT_ADDRESS_DST],
							      SADB_EXT_ADDRESS_DST,
							      tdbp->tdb_said.proto,
							      0,
							      tdbp->tdb_addr_d),
				  extensions)
	      && pfkey_safe_build(error = pfkey_prop_build(&extensions[SADB_EXT_PROPOSAL],
							   64, /* replay */
							   sizeof(comb)/sizeof(struct sadb_comb),
							   &(comb[0])),
				  extensions)
		)) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire: failed to "
			    "build the acquire message extensions\n");
		goto errlab;
	}
	
	if ((error = pfkey_msg_build(&pfkey_msg, extensions, EXT_BITS_OUT))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire: failed to "
			    "build the acquire message\n");
		goto errlab;
	}
	
	for(pfkey_socketsp = pfkey_registered_sockets[satype];
	    pfkey_socketsp;
	    pfkey_socketsp = pfkey_socketsp->next) {
		if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_msg))) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire: "
				    "sending up acquire message for satype=%d to socket=%p failed with error=%d.\n",
				    satype, pfkey_socketsp->socketp, error);
			goto errlab;
		}
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_acquire: "
			    "sending up acquire message for satype=%d to socket=%p succeeded.\n",
			    satype, pfkey_socketsp->socketp);
	}
	
 errlab:
	if (pfkey_msg) {
		pfkey_msg_free(&pfkey_msg);
	}
	pfkey_extensions_free(extensions);
	return error;
}



int (*ext_processors[SADB_EXT_MAX+1])(struct sadb_ext *pfkey_ext, struct pfkey_extracted_data* extr) =
{
  NULL, /* pfkey_msg_process, */
        pfkey_sa_process,
        pfkey_lifetime_process,
        pfkey_lifetime_process,
        pfkey_lifetime_process,
        pfkey_address_process,
        pfkey_address_process,
        pfkey_address_process,
        pfkey_key_process,
        pfkey_key_process,
        pfkey_ident_process,
        pfkey_ident_process,
        pfkey_sens_process,
        pfkey_prop_process,
        pfkey_supported_process,
        pfkey_supported_process,
        pfkey_spirange_process,
        pfkey_x_kmprivate_process,
        pfkey_x_satype_process,
        pfkey_sa_process,
        pfkey_address_process,
        pfkey_address_process,
        pfkey_address_process,
        pfkey_address_process,
        pfkey_address_process,
	pfkey_x_debug_process
};


int (*msg_parsers[SADB_MAX +1])(struct sock *sk, struct sadb_ext *extensions[], struct pfkey_extracted_data* extr)
 =
{
	NULL, /* RESERVED */
	pfkey_getspi_parse,
	pfkey_update_parse,
	pfkey_add_parse,
	pfkey_delete_parse,
	pfkey_get_parse,
	pfkey_acquire_parse,
	pfkey_register_parse,
	pfkey_expire_parse,
	pfkey_flush_parse,
	pfkey_dump_parse,
	pfkey_x_promisc_parse,
	pfkey_x_pchange_parse,
	pfkey_x_grpsa_parse,
	pfkey_x_addflow_parse,
	pfkey_x_delflow_parse,
	pfkey_x_msg_debug_parse
};

int
pfkey_build_reply(struct sadb_msg *pfkey_msg, struct pfkey_extracted_data *extr,
				struct sadb_msg **pfkey_reply)
{
	struct sadb_ext *extensions[SADB_EXT_MAX+1];
	int error = 0;
	int msg_type = pfkey_msg->sadb_msg_type;
	int seq = pfkey_msg->sadb_msg_seq;

	KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_build_reply: building reply"
					" with type: %d\n", msg_type);
	pfkey_extensions_init(extensions);
	if (!extr || !extr->tdb) {
			KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_build_reply: bad TDB"
							"passed\n");
			return EINVAL;
	}
	error = pfkey_safe_build(pfkey_msg_hdr_build(&extensions[0],
						     msg_type,
						     proto2satype(extr->tdb->tdb_said.proto),
						     0,
						     seq,
						     pfkey_msg->sadb_msg_pid),
				 extensions) &&
		(!(extensions_bitmaps[EXT_BITS_OUT][EXT_BITS_REQ][msg_type] &
		   1 << SADB_EXT_SA)
		 || pfkey_safe_build(pfkey_sa_build(&extensions[SADB_EXT_SA],
						    SADB_EXT_SA,
						    extr->tdb->tdb_said.spi,
						    extr->tdb->tdb_replaywin,
						    extr->tdb->tdb_state,
						    extr->tdb->tdb_authalg,
						    extr->tdb->tdb_encalg,
						    extr->tdb->tdb_flags),
				     extensions)) &&
		(!(extensions_bitmaps[EXT_BITS_OUT][EXT_BITS_REQ][msg_type] &
		   1 << SADB_EXT_LIFETIME_CURRENT)
		 || pfkey_safe_build(pfkey_lifetime_build(&extensions
							  [SADB_EXT_LIFETIME_CURRENT],
							  SADB_EXT_LIFETIME_CURRENT,
							  extr->tdb->tdb_lifetime_allocations_c,
							  extr->tdb->tdb_lifetime_bytes_c,
							  extr->tdb->tdb_lifetime_addtime_c,
							  extr->tdb->tdb_lifetime_usetime_c),
				     extensions)) &&
		(!(extensions_bitmaps[EXT_BITS_OUT][EXT_BITS_REQ][msg_type] &
		   1 << SADB_EXT_ADDRESS_SRC)
		 || pfkey_safe_build(pfkey_address_build(&extensions[SADB_EXT_ADDRESS_SRC],
							 SADB_EXT_ADDRESS_SRC,
							 extr->tdb->tdb_said.proto,
							 0,
							 extr->tdb->tdb_addr_s),
				     extensions)) &&
		(!(extensions_bitmaps[EXT_BITS_OUT][EXT_BITS_REQ][msg_type] &
		   1 << SADB_EXT_ADDRESS_DST)
		 || pfkey_safe_build(pfkey_address_build(&extensions[SADB_EXT_ADDRESS_DST],
							 SADB_EXT_ADDRESS_DST,
							 extr->tdb->tdb_said.proto,
							 0,
							 extr->tdb->tdb_addr_d),
				     extensions));

	if (error == 0) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_build_reply: "
			    "building extensions failed\n");
		return EINVAL;
	}

	KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_build_reply: built extensions"
		    ", proceed to build the message\n");
	KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_build_reply: extensions[1]="
		    " %p\n", extensions[1]);
	error = pfkey_msg_build(pfkey_reply, extensions, EXT_BITS_OUT);
	pfkey_extensions_free(extensions);

	return error;
}

int
pfkey_msg_interp(struct sock *sk, struct sadb_msg *pfkey_msg,
				struct sadb_msg **pfkey_reply)
{
	int error = 0;
	int i;
	struct sadb_ext *extensions[SADB_EXT_MAX+1];
	struct pfkey_extracted_data extr = {NULL, NULL, NULL};
	// int msg_len = pfkey_msg->sadb_msg_len * IPSEC_PFKEYv2_ALIGN;
	
	KLIPS_PRINT(debug_pfkey, "pfkey_msg_interp: before write_lock\n");

	/* we need a passiv wait on this lock, otherwise we have a deadlock here. */
	local_bh_disable();
	while (!write_trylock(&sync_lock)) {
		local_bh_enable();
		sleep_on(&sync_lock_wq);
		local_bh_disable();
	}
	// write_lock_bh2(&sync_lock);

	KLIPS_PRINT(debug_pfkey, "pfkey_msg_interp: after write_lock\n");
		
	pfkey_extensions_init(extensions);
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_interp: parsing message "
		    "ver=%d, type=%d, errno=%d, satype=%d, len=%d, res=%d, seq=%d, pid=%d.\n", 
		    pfkey_msg->sadb_msg_version,
		    pfkey_msg->sadb_msg_type,
		    pfkey_msg->sadb_msg_errno,
		    pfkey_msg->sadb_msg_satype,
		    pfkey_msg->sadb_msg_len,
		    pfkey_msg->sadb_msg_reserved,
		    pfkey_msg->sadb_msg_seq,
		    pfkey_msg->sadb_msg_pid);
	
	if((error = pfkey_alloc_tdb(&(extr.tdb)))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_interp: "
			    "something's really wrong, extr.tdb=%p should be NULL.\n", extr.tdb);
		SENDERR(-error);
	}	

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_interp: "
		    "allocated extr->tdb=%p.\n", extr.tdb);
	
	if(pfkey_msg->sadb_msg_satype > SADB_SATYPE_MAX) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_interp: satype %d > max %d\n", 
			    pfkey_msg->sadb_msg_satype, SADB_SATYPE_MAX);
		SENDERR(EINVAL);
	}
	
	switch(pfkey_msg->sadb_msg_type) {
	case SADB_GETSPI:
	case SADB_UPDATE:
	case SADB_ADD:
	case SADB_DELETE:
	case SADB_X_GRPSA:
		if(!(extr.tdb->tdb_said.proto = satype2proto(pfkey_msg->sadb_msg_satype))) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_interp: satype %d lookup failed.\n", 
				    pfkey_msg->sadb_msg_satype);
			SENDERR(EINVAL);
		}
		break;
	case SADB_X_ADDFLOW:
		extr.tdb->tdb_said.proto = satype2proto(pfkey_msg->sadb_msg_satype);
		break;
	default:
	}
	
	/* The NULL below causes the default extension parsers to be used */
	/* Parse the extensions */
	if((error = pfkey_msg_parse(pfkey_msg, NULL, extensions, EXT_BITS_IN)))
	{
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_interp: message parsing failed with error %d.\n",
			    error); 
		SENDERR(-error);
	}
	
	/* Process the extensions */
	for(i=1; i <= SADB_EXT_MAX;i++)	{
		if(extensions[i] != NULL) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_msg_interp: "
				    "processing ext %d %p with processor %p.\n", 
				    i, extensions[i], ext_processors[i]);
			if((error = ext_processors[i](extensions[i], &extr))) {
				KLIPS_PRINT(debug_pfkey,
					    "klips_debug:pfkey_msg_interp: extension processing for type %d failed with error %d.\n",
					    i,error); 
				SENDERR(-error);
			}
			
		}
		
	}
	
	/* Parse the message types */
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_msg_interp: "
		    "parsing message type %d with msg_parser %p.\n",
		    pfkey_msg->sadb_msg_type,
		    msg_parsers[pfkey_msg->sadb_msg_type]); 
	if((error = msg_parsers[pfkey_msg->sadb_msg_type](sk, extensions, &extr))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_msg_interp: message parsing failed with error %d.\n",
			    error); 
		SENDERR(-error);
	}

#ifdef CONFIG_IPSEC_PFKEY_LB // PFKEY nonconform behavior

	/* echo messages (unchanged) */
	
	switch(pfkey_msg->sadb_msg_type) {
	case SADB_X_ADDFLOW:
	case SADB_X_DELFLOW:
	case SADB_ADD:
	case SADB_DELETE:
	  {
	    
	    struct socket_list *pfkey_socketsp;
	    struct sadb_msg *pfkey_reply;
	    
	    if((pfkey_reply = (struct sadb_msg*)kmalloc((pfkey_msg->sadb_msg_len * IPSEC_PFKEYv2_ALIGN), GFP_ATOMIC)) == NULL) 
	      {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: "
			    "memory allocation error.\n");
		SENDERR(ENOBUFS);
	      }
	    
	    memcpy((void*)pfkey_reply, (void*)pfkey_msg,  pfkey_msg->sadb_msg_len * IPSEC_PFKEYv2_ALIGN);
	    pfkey_reply->sadb_msg_len = pfkey_msg->sadb_msg_len;
	    
	    for(pfkey_socketsp = pfkey_open_sockets;
		pfkey_socketsp;
		pfkey_socketsp = pfkey_socketsp->next) {
	      KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_sendmsg: "
			  "sending up error=%d message=%p to socket=%p.\n",
			  error,
			  pfkey_reply,
			  pfkey_socketsp->socketp);
	      if((error = pfkey_upmsg(pfkey_socketsp->socketp, pfkey_reply))) {
		KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_sendmsg: "
			    "sending up error message to socket=%p failed with error=%d.\n",
			    pfkey_socketsp->socketp, error);
		pfkey_msg_free(&pfkey_reply);
		goto errlab;
	      }
	      KLIPS_PRINT(debug_pfkey, "klips_debug:pfkey_sendmsg: "
			  "sending up error message to socket=%p succeeded.\n",
			  pfkey_socketsp->socketp);	      	      
	    }
	    
	    pfkey_msg_free(&pfkey_reply);	  
	    
	  }
	}
#endif
#if 0
	error = pfkey_build_reply(pfkey_msg, &extr, pfkey_reply);
	if (error) {
		*pfkey_reply = NULL;
	}
#endif	

 errlab:
	KLIPS_PRINT(debug_netlink, "pfkey_msg_interp: before write_unlock\n");
	write_unlock_bh(&sync_lock);
	KLIPS_PRINT(debug_netlink, "pfkey_msg_interp: after write_unlock\n");
	if(extr.eroute) kfree(extr.eroute);
	return(error);
}

