/*
 * filename: lamo.c
 *
 * written by: Oliver Stecklina <stecklina@secunet.de>
 *
 * last edit: 15.08.2000
 *
 */

#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <linux/module.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>

#define MAGIC_PREFIX "lamo"

extern int errno;

static inline _syscall2(unsigned long,create_module,const char *,name_user,size_t,size)
static inline _syscall5(long,query_module,const char *,name_user,int,which,char *,buf,size_t,bufsize,size_t *,ret);
static inline _syscall1(long,delete_module,const char *,name_user);

extern void *sys_call_table[];

unsigned long (*original_create_module)(const char *, size_t);
long (*original_query_module)(const char *, int, char *, size_t, size_t *);
long (*original_delete_module)(const char *);

static int (*original_modules_read_proc)(char *, char **, off_t, int, int *, void *);

static int lamo_modules_read_proc_ch = 0;

#ifdef MODULE


static int lamo_proc_match(int len, const char *name,struct proc_dir_entry * de)
{
	if (!de || !de->low_ino)
		return 0;
	if (de->namelen != len)
		return 0;
	return !memcmp(name, de->name, len);
}


static int lamo_xlate_proc_name(const char *name, struct proc_dir_entry **ret, const char **residual)
{
	const char *cp = name, *next;
	struct proc_dir_entry *de;
	int len;

	de = &proc_root;
	while (1) {
		next = strchr(cp, '/');
		if (!next)
			break;

		len = next - cp;
		for (de = de->subdir; de ; de = de->next) {
			if (lamo_proc_match(len, cp, de))
				break;
		}
		if (!de)
			return -ENOENT;
		cp += len + 1;
	}
	*residual = cp;
	*ret = de;

	return 0;
}


asmlinkage unsigned long lamo_create_module(const char *name_user, size_t size)
{
	printk(KERN_NOTICE "lamo: loading of %s for %d aborted \n", name_user, current->uid);

	return -ENOSYS;
}

asmlinkage long lamo_delete_module(const char *name_user)
{
	printk(KERN_NOTICE "lamo: unloading of %s for %d aborted \n", name_user, current->uid);

    return -ENOSYS;
}

asmlinkage long lamo_query_module(const char *name_user, int which, char *buf, size_t bufsize, size_t *ret)
{
	long res;

	res = (*original_query_module)(name_user, which, buf, bufsize, ret);

	if (!res && which == QM_MODULES) {
		char tmp[bufsize], *ptr;
		int match = 0, cnt;

		ptr = buf;

		memset(tmp, 0, bufsize);
		for (cnt = 0; cnt < bufsize && ptr; cnt++) {
			if (!match && !strncmp(MAGIC_PREFIX, ptr, 4)) {
				match++;
				*ret = *ret - 1;
				ptr += 5;
			}
	
			get_user(tmp[cnt], ptr++);
		}
	
		copy_to_user(buf, tmp, bufsize);
	}

	return res;
}

int lamo_modules_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	char *lamo_page;
	char *lamo_start;
	int lamo_count = PAGE_SIZE, lamo_eof = 0;
	off_t lamo_off = 0;

	int len = 0;

	lamo_page = __getname();
	if (!lamo_page)
		return -ENOMEM;
	
	len = (*original_modules_read_proc)(lamo_page, &lamo_start, lamo_off, lamo_count, &lamo_eof, NULL);

	if (len) {
		char *lamo = NULL, *ptr = lamo_page;
		int i, lamo_len = 0;

		/* looking for lamo line */
		for (i = 0; i < PAGE_SIZE && ptr; i++)
			if (!strncmp(MAGIC_PREFIX, ptr, 4)) {
				char *p = ptr;

				lamo = ptr;
	
				while (p && *p != '\n') {
					p++;
					lamo_len++;
				}

				if (p && *p == '\n')
					lamo_len++;

				break;
			}

		if (!lamo_len || !lamo)
			memcpy(page, lamo_page, len);
		else {
			int tocp, cnt = 0;

			if ((count + off) < (lamo - lamo_page)) {
				memcpy(page, lamo_page, count);
				*eof = 0;
				*start = page + off;
				len = count - off;

				goto out;
			}

			cnt = lamo - lamo_page;
			memcpy(page, lamo_page, cnt);

			tocp = count + off - cnt;
	
			if (tocp > 0) {
				int behind = len - (lamo - lamo_page + lamo_len);

				if (tocp > behind)
					tocp = behind;

				memcpy(page + cnt, lamo + lamo_len, tocp);
				cnt += tocp;
	
				if (cnt < (count + off))
					memset(page + cnt, 0, (count + off) - cnt);
			}
		}

    	if (len <= off + count)
			*eof = 1;

		*start = page + off;
		len -= off;

		if (len > count)
			len = count;
	}

out:
	putname(lamo_page);

	return len;
}

int init_module(void)
{
	char name[]="modules";
	struct proc_dir_entry ent, *parent;
	const char *fn = name;

	parent=&ent;

	original_create_module = sys_call_table[__NR_create_module];
	sys_call_table[__NR_create_module] = lamo_create_module;

	original_query_module = sys_call_table[__NR_query_module];
	sys_call_table[__NR_query_module] = lamo_query_module;

	original_delete_module = sys_call_table[__NR_delete_module];
	sys_call_table[__NR_delete_module] = lamo_delete_module;

	if (!lamo_xlate_proc_name(name, &parent, &fn)) {
		struct proc_dir_entry **p;
		int len = strlen(fn);

		for (p = &parent->subdir; *p; p=&(*p)->next)
			if (lamo_proc_match(len, fn, *p)) {
				struct proc_dir_entry *de = *p;

				original_modules_read_proc = de->read_proc;

				remove_proc_entry(name, NULL);
				create_proc_read_entry(name, 0, NULL, lamo_modules_read_proc, NULL);

				lamo_modules_read_proc_ch++;

				break;
			}
	} else
		printk(KERN_NOTICE "lamo: xlate_proc_name(...) failed. \n");

	printk(KERN_NOTICE "lamo: last module loaded\n");

	return 0;
}


int cleanup_module(void)
{
	sys_call_table[__NR_create_module] = original_create_module;
	sys_call_table[__NR_query_module] = original_query_module;
	sys_call_table[__NR_delete_module] = original_delete_module;

	if (lamo_modules_read_proc_ch) {
		remove_proc_entry("modules", NULL);
		create_proc_read_entry("modules", 0, NULL, original_modules_read_proc, NULL);
	}

	lamo_modules_read_proc_ch = 0;

	return -EPERM;
}
#endif
