/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_capability.c
 *
 * last edit : 15.11.2000
 *
 * content : implementation of sys_capset audit function
 *
 */

#include <linux/fs.h>

#include <linux/audit.h>
#include <linux/audit_event.h>


/*
 * name : audit_capset
 * task : Generate a capset event
 * args : status : final execute status of sys_capset
 * desc : Is the respective audit function to sys_capset
 */
int audit_capset(int status)
{
	struct audit_record *rec;
	struct audit_host_record *data;

	/* get us an empty record, for store audit data */
	if ((rec = audit_get_free_record(AU_HOST_REC, 1, AUE_CAPSET)) == NULL)
		return 0;

	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}
	
	audit_fill_static(data);
	audit_fill_filedata(data, NULL);

	rec->event = AUE_CAPSET;

	if (!status)
		data->status = AU_SYSCALL_SUCCESSFULL;
	else
		data->status = AU_SYSCALL_FAILED;

	data->arg1 = 0;
	data->arg2 = 0;

	data->env = NULL;

	data->file_len = 0;
	data->file2_len = 0;

	rec->state = AU_FULL;

	return 0;
}
