/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_quot.c
 *
 * last edit : 19.11.2000
 *
 */

#include <asm/uaccess.h>

#include <linux/audit.h>
#include <linux/audit_event.h>

#ifdef CONFIG_ACL
#include <linux/acl.h>
#endif

int audit_quotactl(unsigned long * args)
{
	struct audit_record *rec;
	struct audit_host_record *data;

	if ((rec = audit_get_free_record(AU_HOST_REC, 1, AUE_QUOTACTL)) == NULL)
		return 0;

	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}

	audit_resolve_filename(data->file, &data->file_len, (char *) args[2],
	                       STRLEN_USER((char *) args[2]));

#ifdef CONFIG_ACL
	if (data->file_len) {
		if (acl_check_host(data->file + (__u16) (data->file_len >> 16),
		                   AUE_QUOTACTL, (int) args[0])) {
			rec->state = AU_UNUSED;
			return 0;
		}
	}
#endif

	audit_fill_static(data);
	audit_fill_cfiledata(data, (char *) args[2]);

	if (!(int) args[0])
		data->status = AU_SYSCALL_SUCCESSFULL;
	else
		data->status = AU_SYSCALL_FAILED;

	data->arg1 = args[1];
	data->arg2 = args[3];
	data->env = NULL;
	data->file2_len = 0;

	rec->state = AU_FULL;
	
	return 0;
}

/* eof */
