/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_swapfile.c
 *
 * last edit : 19.11.2000
 *
 */
#include <asm/uaccess.h>

#include <linux/audit.h>
#include <linux/audit_event.h>

#ifdef CONFIG_ACL
#include <linux/acl.h>
#endif

int audit_swapon(int status, const char * filename, int swap_flags)
{
	struct audit_record *rec;
	struct audit_host_record *data;

	if ((rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SWAPON)) == NULL)
		return 0;

	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}

	audit_resolve_filename(data->file, &data->file_len, filename,
	                       STRLEN_USER(filename));

#ifdef CONFIG_ACL
	if (data->file_len) {
		if (acl_check_host(data->file + (__u16) (data->file_len >> 16),
		                    AUE_SWAPON, status)) {
			rec->state = AU_UNUSED;
			return 0;
		}
	}
#endif

	audit_fill_static(data);
	audit_fill_cfiledata(data, filename);

	if (!status)
		data->status = AU_SYSCALL_SUCCESSFULL;
	else
		data->status = AU_SYSCALL_FAILED;

	data->arg1 = swap_flags;
	data->arg2 = 0;
	data->env = NULL;
	data->file2_len = 0;

	rec->state = AU_FULL;
	
	return 0;
}


int audit_swapoff(int status, const char * filename)
{
	struct audit_record *rec;
	struct audit_host_record *data;

	if ((rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SWAPOFF)) == NULL)
		return 0;

	data = (struct audit_host_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}

	audit_resolve_filename(data->file, &data->file_len, filename,
	                       STRLEN_USER(filename));

#ifdef CONFIG_ACL
	if (data->file_len) {
		if (acl_check_host(data->file + (__u16) (data->file_len >> 16),
		                  AUE_SWAPOFF, status)) {
			rec->state = AU_UNUSED;
			return 0;
		}
	}
#endif

	audit_fill_static(data);
	audit_fill_cfiledata(data, filename);

	if (!status)
		data->status = AU_SYSCALL_SUCCESSFULL;
	else
		data->status = AU_SYSCALL_FAILED;

	data->arg1 = 0;
	data->arg2 = 0;
	data->env = NULL;
	data->file2_len = 0;

	rec->state = AU_FULL;
	
	return 0;
}

/* eof */
