/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_sys.c
 *
 * last edit : 12 Februar 2002
 *
 */

#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <linux/string.h>
#include <linux/highuid.h>

#include <linux/audit.h>
#include <linux/audit_event.h>

/* from audit/host_buffer.c */
extern struct audit_dyn_buf *au_env_buf;


static int __audit_setuid(int status, uid_t uid, uid_t ruid, uid_t euid)
{
	if (audit_doit(AUE_SETUID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETUID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				if (!status) {
					data->status = AU_SYSCALL_SUCCESSFULL;

					data->ruid = ruid;
					data->euid = euid;
				} else
					data->status = AU_SYSCALL_FAILED;

				data->arg1 = uid;

				data->arg2 = 0;
				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setuid16(old_uid_t uid)
{
	long error;
	uid_t old_ruid, old_euid;

	old_ruid = current->uid;
	old_euid = current->euid;

	error = ((long (*) (old_uid_t)) orig_sys_call_table[__NR_setuid])(uid);

	__audit_setuid((int) error, (uid_t) uid, old_ruid, old_euid);

	return error;
}

asmlinkage long audit_setuid(uid_t uid)
{
	long error;
	uid_t old_ruid, old_euid;

	old_ruid = current->uid;
	old_euid = current->euid;

	error = ((long (*) (uid_t)) orig_sys_call_table[__NR_setuid32])(uid);

	__audit_setuid((int) error, uid, old_ruid, old_euid);

	return error;
}


static int __audit_setgid(int status, gid_t gid, gid_t old_rgid, gid_t old_egid)
{
	if (audit_doit(AUE_SETGID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETGID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				if (!status) {
					data->status = AU_SYSCALL_SUCCESSFULL;
					data->rgid = old_rgid;
					data->egid = old_egid;
				} else
					data->status = AU_SYSCALL_FAILED;

				data->arg1 = gid;
				data->arg2 = 0;

				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;
			
				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setgid16(old_gid_t gid)
{
	long error;
	gid_t old_rgid, old_egid;

	old_rgid = current->gid;
	old_egid = current->egid;

	error = ((long (*) (old_gid_t)) orig_sys_call_table[__NR_setgid])(gid);

	__audit_setgid((int) error, (gid_t) gid, old_rgid, old_egid);

	return error;
}

asmlinkage long audit_setgid(gid_t gid)
{
	long error;
	gid_t old_rgid, old_egid;

	old_rgid = current->gid;
	old_egid = current->egid;

	error = ((long (*) (gid_t)) orig_sys_call_table[__NR_setgid32])(gid);

	__audit_setgid((int) error, gid, old_rgid, old_egid);

	return error;
}


static int __audit_setreuid(int status, uid_t ruid, uid_t euid, uid_t old_ruid,
                            uid_t old_euid)
{
	if (audit_doit(AUE_SETREUID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETREUID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				if (!status) {
					data->status = AU_SYSCALL_SUCCESSFULL;

					data->ruid = old_ruid;
					data->euid = old_euid;
				} else
					data->status = AU_SYSCALL_FAILED;

				data->arg1 = ruid;
				data->arg2 = euid;

				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setreuid16(old_uid_t ruid, old_uid_t euid)
{
	long error;
	uid_t old_ruid, old_euid;

	old_ruid = current->uid;
	old_euid = current->euid;

	error = ((long (*) (old_uid_t, old_uid_t)) orig_sys_call_table[__NR_setreuid])(ruid, euid);

	__audit_setreuid((int) error, low2highuid(ruid), low2highuid(euid),
	                 old_ruid, old_euid);

	return error;
}

asmlinkage long audit_setreuid(uid_t ruid, uid_t euid)
{
	long error;
	uid_t old_ruid, old_euid;

	old_ruid = current->uid;
	old_euid = current->euid;

	error = ((long (*) (uid_t, uid_t)) orig_sys_call_table[__NR_setreuid32])(ruid, euid);

	__audit_setreuid((int) error, ruid, euid, old_ruid, old_euid);

	return error;
}


static int __audit_setregid(int status, gid_t rgid, gid_t egid, gid_t old_rgid,
                            gid_t old_egid)
{
	if (audit_doit(AUE_SETREGID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETREGID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				if (!status) {
					data->status = AU_SYSCALL_SUCCESSFULL;

					data->rgid = old_rgid;
					data->egid = old_egid;
				} else
					data->status = AU_SYSCALL_FAILED;

				data->arg1 = rgid;
				data->arg2 = egid;

				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setregid16(old_gid_t rgid, old_gid_t egid)
{
	long error;
	gid_t old_rgid, old_egid;

	old_rgid = current->gid;
	old_egid = current->egid;

	error = ((long (*) (old_gid_t, old_gid_t)) orig_sys_call_table[__NR_setregid])(rgid, egid);

	__audit_setregid((int) error, low2highgid(rgid), low2highgid(egid),
	                 old_rgid, old_egid);

	return error;
}

asmlinkage long audit_setregid(gid_t rgid, gid_t egid)
{
	long error;
	gid_t old_rgid, old_egid;

	old_rgid = current->gid;
	old_egid = current->egid;

	error = ((long (*) (gid_t, gid_t)) orig_sys_call_table[__NR_setregid32])(rgid, egid);

	__audit_setregid((int) error, rgid, egid, old_rgid, old_egid);

	return error;
}


static int __audit_setfsuid(uid_t uid)
{
	int status;

	status = (current->fsuid != uid) ? 0 : -1;

	if (audit_doit(AUE_SETFSUID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETFSUID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				data->status = status ? AU_SYSCALL_SUCCESSFULL :
				                        AU_SYSCALL_FAILED;

				data->arg1 = uid;

				data->arg2 = 0;
				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;
			
				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setfsuid16(old_uid_t uid)
{
	long retval;

	retval = ((long (*) (old_uid_t)) orig_sys_call_table[__NR_setfsuid])(uid);

	__audit_setfsuid((uid_t) uid);

	return retval;
}

asmlinkage long audit_setfsuid(uid_t uid)
{
	long retval;

	retval = ((long (*) (uid_t)) orig_sys_call_table[__NR_setfsuid32])(uid);

	__audit_setfsuid(uid);

	return retval;
}


static int __audit_setfsgid(gid_t gid)
{
	int status;

	status = (current->fsgid != gid) ? 0 : -1;

	if (audit_doit(AUE_SETFSGID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETFSGID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				data->status = status ? AU_SYSCALL_SUCCESSFULL :
				                       AU_SYSCALL_FAILED;

				data->arg1 = gid;

				data->arg2 = 0;
				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setfsgid16(old_gid_t gid)
{
	long retval;

	retval = ((long (*) (old_gid_t)) orig_sys_call_table[__NR_setfsgid])(gid);
	__audit_setfsgid((gid_t) gid);

	return retval;
}

asmlinkage long audit_setfsgid(gid_t gid)
{
	long retval;

	retval = ((long (*) (gid_t)) orig_sys_call_table[__NR_setfsgid32])(gid);
	__audit_setfsgid(gid);

	return retval;
}


static int __audit_setresuid(int status, uid_t ruid, uid_t euid,
                             uid_t old_ruid, uid_t old_euid)
{
	if (audit_doit(AUE_SETRESUID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETRESUID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				if (!status) {
					data->status = AU_SYSCALL_SUCCESSFULL;

					data->ruid = old_ruid;
					data->euid = old_euid;
				} else
					data->status = AU_SYSCALL_FAILED;

				data->arg1 = ruid;
				data->arg2 = euid;

				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setresuid16(old_uid_t ruid, old_uid_t euid,
                                  old_uid_t suid)
{
	long error;
	uid_t old_ruid, old_euid;

	old_ruid = current->uid;
	old_euid = current->euid;

	error = ((long (*) (old_uid_t, old_uid_t, old_uid_t)) orig_sys_call_table[__NR_setresuid])(ruid, euid, suid);
	__audit_setresuid((int) error, low2highuid(ruid), low2highuid(euid),
	                  old_ruid, old_euid);

	return error;
}

asmlinkage long audit_setresuid(uid_t ruid, uid_t euid, uid_t suid)
{
	long error;
	uid_t old_ruid, old_euid;

	old_ruid = current->uid;
	old_euid = current->euid;

	error = ((long (*) (uid_t, uid_t, uid_t)) orig_sys_call_table[__NR_setresuid32])(ruid, euid, suid);
	__audit_setresuid((int) error, ruid, euid, old_ruid, old_euid);

	return error;
}


static int __audit_setresgid(int status, gid_t rgid, gid_t egid,
                             gid_t old_rgid, gid_t old_egid)
{
	if (audit_doit(AUE_SETRESGID, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETRESGID);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				if (!status) {
					data->status = AU_SYSCALL_SUCCESSFULL;

					data->rgid = old_rgid;
					data->egid = old_egid;
				} else
					data->status = AU_SYSCALL_FAILED;

				data->arg1 = rgid;
				data->arg2 = egid;

				data->env = NULL;
				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setresgid16(old_gid_t rgid, old_gid_t egid,
                                  old_gid_t sgid)
{
	long error;
	gid_t old_rgid, old_egid;

	old_rgid = current->gid;
	old_egid = current->egid;

	error = ((long (*) (old_gid_t, old_gid_t, old_gid_t)) orig_sys_call_table[__NR_setresgid])(rgid, egid, sgid);
	__audit_setresgid((int) error, low2highgid(rgid), low2highgid(egid), old_rgid, old_egid);

	return error;
}

asmlinkage long audit_setresgid(gid_t rgid, gid_t egid, gid_t sgid)
{
	long error;
	gid_t old_rgid, old_egid;

	old_rgid = current->gid;
	old_egid = current->egid;

	error = ((long (*) (gid_t, gid_t, gid_t)) orig_sys_call_table[__NR_setresgid32])(rgid, egid, sgid);
	__audit_setresgid((int) error, rgid, egid, old_rgid, old_egid);

	return error;
}


static int __audit_setgroups(int status, int max_size, int loc, int gidsetsize,
                             gid_t *grouplist)
{
	if (audit_doit(AUE_SETGROUPS, status)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETGROUPS);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				data->status = (!status) ? AU_SYSCALL_SUCCESSFULL :
			                                	AU_SYSCALL_FAILED;

				data->env = NULL;
				data->arg1 = gidsetsize;
				if (gidsetsize > 0) {
					data->arg2 = grouplist[0];
					if (gidsetsize > 1) {
						void *src[2];
						unsigned long str[2];

						str[0] = (max_size - 1) * sizeof(gid_t);
						str[1] = (unsigned long) &grouplist[1];

						src[0] = (void *) str;
						src[1] = NULL;

						audit_fill_dyn_buf2(au_env_buf, &data->env, src, 1,
										    loc);
					}
				} else
					data->arg2 = 0;

				data->file_len = 0;
				data->file2_len = 0;

				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return 0;
}

asmlinkage long audit_setgroups16(int gidsetsize, old_gid_t *grouplist)
{
	gid_t groups[NGROUPS];
	old_gid_t old_groups[NGROUPS];
	int i, size;
	long error;

	size = gidsetsize;
	if (size > NGROUPS)
		size = NGROUPS;

	if (copy_from_user(old_groups, grouplist, size * sizeof(old_gid_t)))
		memset(old_groups, 0, size * sizeof(old_gid_t));

	for (i = 0 ; i < size ; i++)
		groups[i] = (gid_t) old_groups[i];

	error = ((long (*) (int, old_gid_t *)) orig_sys_call_table[__NR_setgroups])(gidsetsize, grouplist);

	__audit_setgroups((int) error, size, AU_K_SPACE, gidsetsize, groups);

	return error;
}

asmlinkage long audit_setgroups(int gidsetsize, gid_t *grouplist)
{
	int size;
	long error;

	size = gidsetsize;
	if (size > NGROUPS)
		size = NGROUPS;

	error = ((long (*) (int, gid_t *)) orig_sys_call_table[__NR_setgroups32])(gidsetsize, grouplist);

	__audit_setgroups((int) error, size, AU_U_SPACE, gidsetsize, grouplist);

	return error;
}


asmlinkage long audit_sethostname(char *name, int len)
{
	long error;
	struct audit_record *rec;

	rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETHOSTNAME);

	error = ((long (*) (char *, int)) orig_sys_call_table[__NR_sethostname])(name, len);

	if (rec) {
		struct audit_host_record *data;

		data = (struct audit_host_record *) rec->data;
		if (data && audit_doit(AUE_SETHOSTNAME, (int) error)) {
			audit_fill_static(data);
			audit_fill_filedata(data, NULL);

			data->status = (!(int) error) ? AU_SYSCALL_SUCCESSFULL :
			                                AU_SYSCALL_FAILED;

			/* the new hostname is saved in file */
			data->file_len = len;
			if (strncpy_from_user(data->file, name, len))
				data->file_len = 0;

			data->arg1 = 0;
			data->arg2 = 0;
			data->env = NULL;
			data->file2_len = 0;
	
			rec->state = AU_FULL;
		} else
			rec->state = AU_UNUSED;
	}

	return error;
}


asmlinkage long audit_setdomainname(char *name, int len)
{
	long error;

	error = ((long (*) (char *, int)) orig_sys_call_table[__NR_setdomainname])(name, len);
	if (audit_doit(AUE_SETDOMAINNAME, (int) error)) {
		struct audit_record *rec;

		rec = audit_get_free_record(AU_HOST_REC, 1, AUE_SETDOMAINNAME);
		if (rec) {
			struct audit_host_record *data;

			data = (struct audit_host_record *) rec->data;
			if (data) {
				audit_fill_static(data);
				audit_fill_filedata(data, NULL);

				data->file_len = len;
				if (strncpy_from_user(data->file, name, len))
					data->file_len = 0;

				data->status = (!(int) error) ? AU_SYSCALL_SUCCESSFULL :
			                                	AU_SYSCALL_FAILED;

				data->arg1 = 0;
				data->arg2 = 0;
				data->env = NULL;
				data->file2_len = 0;
			
				rec->state = AU_FULL;
			} else
				rec->state = AU_UNUSED;
		}
	}

	return error;
}


/* eof */
