/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_tcpip_l2.c
 *
 * last edit : 13.06.2001
 *
 * content :
 *
 */

#include <linux/audit.h>
#include <linux/audit_event.h>

extern struct audit_dyn_buf *au_net_env_buf;

int audit_no_proto(struct sk_buff *skb, unsigned int sub_evt,
                   struct net_device *dev)
{
	struct audit_record *rec;
	struct audit_net_record *data;

	if ((rec = audit_get_free_record(AU_NET_REC, 0, AUE_NO_PROTO)) == NULL)
		return 0;

	data = (struct audit_net_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}
	
	rec->pid = sub_evt;

	audit_copy_dev(data, dev, IFNAMSIZ);
	audit_copy_mac(data, skb, dev);

	data->nh_prot_size = 0;
	data->h_prot_size = 0;

	if (au_net_env_buf) {
		void *src[2];
		unsigned long str[2];

		str[0] = skb->len;
		str[1] = (unsigned long) skb->data;

		src[0] = (void *) str;
		src[1] = NULL;

		audit_fill_dyn_buf2(au_net_env_buf, &data->env, src, 1, AU_K_SPACE);
	} else
		data->env = NULL;

	rec->state = AU_FULL;
	
	return 0;
}
