/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : buffer.h
 *
 * last edit: 06 June 2002
 *
 * content :
 *
 */

#ifndef AUDIT_BUFFER_H
#define AUDIT_BUFFER_H

#include <linux/sched.h>
#include <linux/audit.h>
#include <linux/smp_lock.h>

struct audit_buf {
	atomic_t state;

	struct audit_record *addr;
	struct audit_record *first_full;
	struct audit_record *first_free;

	struct audit_record *to_save;

	unsigned int session_closed;

	struct audit_record *additional;
	struct audit_record *last_add;

	spinlock_t lock;
};

#define AUDIT_BUFFER_INIT { \
	state: ATOMIC_INIT(AU_UNDEF), \
	addr: NULL, \
	first_full: NULL, \
	first_free: NULL, \
	to_save: NULL, \
	session_closed: 0, \
	additional: NULL, \
	last_add: NULL, \
	lock: SPIN_LOCK_UNLOCKED \
}

#define AU_MAX_WAKEUPS 3

struct audit_svc_info {
	atomic_t state;

	char *user_buf;
	char *svc_buf;

	pid_t auditd_pid;
	struct task_struct *auditd_task;

	spinlock_t lock;
	unsigned int wakeup_cnt;

	unsigned int buffer_cnt;
	struct audit_buffer_info *buffers[AU_REC_TYPES];
};

#define AUDIT_SERVICE_INFO_INIT { \
	state: ATOMIC_INIT(AU_UNDEF), \
	user_buf: NULL, \
	svc_buf: NULL, \
	auditd_pid: AU_UNDEF, \
	auditd_task: NULL, \
	lock: SPIN_LOCK_UNLOCKED, \
	wakeup_cnt: 0, \
	buffer_cnt: AU_REC_TYPES, \
	buffers: {NULL, NULL} \
}

extern struct audit_buf au_buf;
extern struct audit_svc_info au_svc;

#define AU_SVC_INITIALIZED(svc) \
	((atomic_read(&(svc)->state) == AU_DOIT) && ((svc)->auditd_pid != AU_UNDEF) && (svc)->auditd_task)

#endif
