/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : net_func.c
 *
 * last edit : 18 June 2002
 *
 * content :
 *
 */

#include <linux/config.h>
#include <linux/if_arp.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/module.h>
#include <net/sock.h>

#include <linux/audit.h>


#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)

#include <linux/ipv6.h>

int audit_copy_ipv6(struct audit_net_record *data, struct ipv6hdr *iphr,
                    __u8 nexthdr)
{
	int len = -EFAULT;

	if (data) {
		data->nh_prot_size = 0;

		if (iphr) {
			len = sizeof(struct ipv6hdr);

			memcpy((char *) &data->nh.ipv6h, (char *) iphr, len);
			data->nh.ipv6h.nexthdr = nexthdr;

			data->nh_prot_size = len;
		}
	}

	return len;
}
#endif


int audit_copy_mac(struct audit_net_record *data, struct sk_buff *skb,
                   struct net_device *dev)
{
	data->protocol = skb->protocol;

	if (dev)
		data->dev_type = dev->type;
	else
		data->dev_type = ARPHRD_VOID;

	if (data->dev_type == ARPHRD_ETHER && skb->mac.raw) {
		struct ethhdr *e_hdr;

		e_hdr = (struct ethhdr *) skb->mac.raw;
		if (skb->mac.raw != skb->head) {
			int hard_header_len;

			/*
			 * It seems so that the ethernet header was modified. This is e.g.
			 * done by ipsec. We need the original header.
			 */
			hard_header_len = skb->mac.raw ? (skb->data - skb->mac.raw) : 0;
			if((hard_header_len < 0) || (hard_header_len > skb_headroom(skb))) {
				memset((char *) data->mac.raw, 0, AU_NET_MAC_RAW_SIZE);
				return 0;
			}

			if (hard_header_len &&
		    	(skb->mac.raw != (skb->data - hard_header_len)) &&
		    	(hard_header_len <= skb_headroom(skb))) {

				e_hdr = (struct ethhdr *) skb->head;
			}
		}
		
		if (e_hdr) {
			memcpy((char *) &data->mac.ethernet, (char *) e_hdr,
			       sizeof(struct ethhdr));
			data->mac_prot_size = sizeof(struct ethhdr);
			return 0;
		}

#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "audit_copy_mac: zero e_hdr address.\n");
#endif
	}
	
	data->mac_prot_size = 0;

	if ((data->protocol != htons(ETH_P_IP)) &&
		(data->protocol != htons(ETH_P_IPV6)) &&
		(data->protocol != htons(ETH_P_IPX))) {

#ifdef CONFIG_AUDIT_DEBUG
		printk(KERN_DEBUG "audit_copy_mac: unsupported protocol %d.\n",
		       data->protocol);
#endif
		return -ESRCH;
	}

	return 0;
}


int audit_copy_ipv4(struct audit_net_record *data, struct iphdr *iphr)
{
	int len = -EFAULT;

	if (data) {
		data->nh_prot_size = 0;

		if (iphr) {
			len = sizeof(struct iphdr);

			/* network layer, should be ipv4 */
			memcpy((char *) &data->nh.iph, (char *) iphr, len);
			data->nh_prot_size = len;
		}
	}

	return len;
}


int audit_copy_dev(struct audit_net_record *data, struct net_device *dev,
                   unsigned int len)
{
	if (dev) {
		data->namelen = strnlen(dev->name, len);
		memcpy(data->name, dev->name, data->namelen);

		return data->namelen;
	}

	data->namelen = 0;
	data->name[0] = '\0';

	return 0;
}

/* EOF */
