/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : timer
 *
 * last edit : 06 June 2002
 *
 * content : 
 *
 */

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/smp_lock.h>
#include <linux/audit.h>

#include "buffer.h"

#define AU_TIME_OUT (CONFIG_AUDIT_TIMEOUT * 100)

static struct timer_list audit_timer;

static void audit_timer_expired(unsigned long data)
{
	struct audit_svc_info *info = (struct audit_svc_info *) data;

	if (!info || !AU_SVC_INITIALIZED(info)) {
		del_timer(&audit_timer);
		return;
	}

	/*
	 * If info is locked, the wakeup function or the service function is
	 * running. In both cases our operation isn't neccessary.
	 */
	if (!spin_is_locked(&info->lock)) {
		int i;

		for (i = 0; i < AU_REC_TYPES; i++)
			if (info->buffers[i]->free_recs != info->buffers[i]->buf_size) {
				wake_up_auditd();
				break;
			}
	}

	mod_timer(&audit_timer, jiffies + AU_TIME_OUT);

	return;
}

int create_audit_timer(struct audit_svc_info *info)
{
	init_timer(&audit_timer);

	audit_timer.function = audit_timer_expired;
	audit_timer.data = (unsigned long) info;

	return 0;
}

void enable_audit_timer(void)
{
	audit_timer.expires = jiffies + AU_TIME_OUT;
	add_timer(&audit_timer);
}

void disable_audit_timer(void)
{
	del_timer(&audit_timer);
}
