/*-
 *
 * Linux Operating System Audit - ACL Extention - ACL Extention
 *
 *   written by Oliver Stecklina <ost@secunet.de>
 *
 * filename : acl.h
 *
 * last edit : 05.08.2001
 *
 */

#ifndef _ACL_H
#define _ACL_H

#include <linux/config.h>

#include <asm/atomic.h>

#ifdef __KERNEL__


#include <linux/socket.h>
#include <linux/in.h>
#include <linux/in6.h>

#else

#include <asm/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#endif


/*
 * We have 32 possible acl options.
 */
#define ACL_EXECVE 0
#define ACL_UNLINK (1 << 1)

#define ACL_TYPE_PID	200
#define ACL_TYPE_UID	201
#define ACL_TYPE_GID	202
#define ACL_TYPE_ALL	203

/* possible acl ctl commands */

#define ACL_SET_EXEC		101

/* size definitions for the audit event masks */

#define AU_MASK_SIZE 8

#define AU_MSK_SIZE32   AU_MASK_SIZE
#define AU_MSK_SIZE16  (AU_MASK_SIZE << 1)
#define AU_MSK_SIZE8   (AU_MASK_SIZE << 2)


#define ACL_SET_AUEVT		102
#define ACL_SET_AUMSK		103

#define ACL_SET_IPV4_AUEVT	104
#define ACL_SET_IPV4_AUMSK	105

#define ACL_SET_IPV6_AUEVT	106
#define ACL_SET_IPV6_AUMSK	107

#ifdef CONFIG_AUDIT

/* global event mask list */
struct audit_acl_msk {
	struct audit_acl_msk *next;
	struct audit_acl_msk *prev;

	atomic_t count;

	union {
		__u8		u_msk8[AU_MSK_SIZE8];
		__u16		u_msk16[AU_MSK_SIZE16];
		__u32		u_msk32[AU_MSK_SIZE32];
	} perm;
#define perm_msk8		perm.u_msk8
#define perm_msk16		perm.u_msk16
#define perm_msk32		perm.u_msk32

	union {
		__u8		u_msk8[AU_MSK_SIZE8];
		__u16		u_msk16[AU_MSK_SIZE16];
		__u32		u_msk32[AU_MSK_SIZE32];
	} val;
#define val_msk8		val.u_msk8
#define val_msk16		val.u_msk16
#define val_msk32		val.u_msk32
};

#endif /* CONFIG_AUDIT */


/* Ressource Event List */
struct acl_rel_entry {
	__u16 type;
	__u16 id;

	union {
		struct {
			__u16 used;
			__u16 msk;
		} op;
#ifdef CONFIG_AUDIT
		struct audit_acl_msk *msk;
#endif
	} v;

#define op_used v.op.used
#define op_mask v.op.msk

	struct acl_entry *ace;

	struct acl_rel_entry *prev;
	struct acl_rel_entry *next;
};

/* ACL - Entry */
struct acl_entry {
	char *name;
	unsigned long hash;

	struct acl_rel_entry *rel;

	struct acl_entry *sib_r;
	struct acl_entry *cld;

	struct acl_entry *sib_l;
	struct acl_entry *parent;
};

#define ACL_ANCHOR {"/", 0, NULL, NULL, NULL, NULL, NULL}


#ifdef CONFIG_AUDIT

/* Network ACL-Structures */

/* Port ACL */
struct audit_port_acl {
	__u16	low_sport;
	__u16	high_sport;
	__u16	low_dport;
	__u16	high_dport;

	__u8	direc;

	struct audit_acl_msk  *msk;
	struct audit_port_acl *next;
};


/* IP-Range structure for IPv4 and IPv6 */
struct audit_ip_rng {
	union {
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
		struct 	in6_addr ip6;
#endif /* CONFIG_IPV6 */
		struct 	in_addr  ip;
		char	raw[16];
	} low;

	union {
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
		struct 	in6_addr ip6;
#endif /* CONFIG_IPV6 */
		struct 	in_addr  ip;
		char	raw[16];
	} high;
};


/* IP-ACL for ipv4 and ipv6 */
struct audit_ip_acl {
	/*
	 * is this entry for an ipv6-address or for a ip-range, tip contains
	 * an address of an audit_ip_rng-entry, otherwise tip is the target
	 * ip-address in network byte order
	 */
	union {
		struct in_addr  	ipv4_addr;
		struct audit_ip_rng *ip_rng;
	} tip;
	
	__u8    cidr;
	__u8    direc;

	struct audit_acl_msk  *msk;
	struct audit_port_acl *ports;

	struct audit_ip_acl   *next;
	struct audit_ip_acl   *prev;

	struct audit_ip_acl   *subnet;
	struct audit_ip_acl   *net;
};

#endif /* CONFIG_AUDIT */


#ifdef __KERNEL__

/* evt_msk.c */

#ifdef CONFIG_AUDIT

extern struct audit_acl_msk acl_msk_list;

extern int init_msk_cachep(void);
extern int kfree_audit_msk(struct audit_acl_msk *msk);
extern int audit_evt_val(struct audit_acl_msk *msk, int evt,
                         unsigned int evt_width);
extern int audit_get_acl_msk(struct audit_acl_msk *entry, __u32 *msk);
extern int audit_update_evt_msk(struct audit_acl_msk **entry_p, int evt,
                                int val, int evt_width, int alloc_flag);
extern int audit_set_evt_msk(struct audit_acl_msk **entry_p, __u32 *evt_msk,
                             int alloc_flag);

/* acl.c */

/* prototyps for the audit acl specific functions */

extern int acl_check_host(char *name, int evti, int status);
extern int acl_check_net_ipv4(unsigned int evt, unsigned int subevt,
                              void *skb_p);
extern int acl_check_net_ipv6(unsigned int evt, unsigned int subevt,
                              void *skb_p);

#endif /* CONFIG_AUDIT */

/* EXECVE is the only operation, which is suppported now */
#define acl_check_exec(name) acl_check_op(name, ACL_EXECVE)
extern int acl_check_op(char *name, int evt);


/* common acl entry point */
extern asmlinkage int sys_acl_ctl(int cmd, unsigned long args,
                                  unsigned int size);

#endif /* __KERNEL__ */

#endif
