/*-
 *
 * Linux Operating System Audit
 *
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_event.h
 *
 * last edit : 16 Februar 2002
 *
 */

#ifndef _AUDIT_EVENT_H
#define _AUDIT_EVENT_H

#define AU_MAX_EID				127
#define AU_MAX_ECLASS			31

#define AU_MIN_HOST_EVT			1
#define AU_MAX_HOST_EVT			73
#define AU_MIN_NET_EVT			80
#define AU_MAX_NET_EVT			89

#define AU_MIN_USER_EVT			100
#define AU_MAX_USER_EVT			127

/* kernel events */

#define	AUE_NULL				0	/* =no indir system call */
#define	AUE_ACCT				1	/* =ad acct(2) */
#define AUE_CAPSET				2   /* =ad */
#define	AUE_FORK				3	/* =pc fork(2) */
#define AUE_CLONE				4   /* =pc */
#define AUE_IOPERM				5   /* =io */
#define AUE_IOPL				6   /* =io */
#define AUE_INIT_MODULE			7   /* =ad */
#define AUE_DELETE_MODULE		8   /* =ad */
#define	AUE_LINK				9	/* =fc link(2) */
#define	AUE_UNLINK				10  /* =fd unlink(2) */
#define	AUE_MKNOD				11  /* =fc mknod(2) */
#define	AUE_RENAME				12	/* =fc,fd rename(2) */
#define	AUE_MKDIR				13	/* =fc mkdir(2) */
#define	AUE_RMDIR				14	/* =fd rmdir(2) */
#define	AUE_SYMLINK				15	/* =fc symlink(2) */
#define	AUE_OPEN				16  /* =fa open(2): place holder */
#define	AUE_USELIB				17  /* =fr */
#define	AUE_OPEN_R				18  /* =fr open(2): read */
#define	AUE_OPEN_RC				19	/* =fc,fr open(2): read,creat */
#define	AUE_OPEN_RT				20	/* =fd,fr open(2): read,trunc */
#define	AUE_OPEN_RTC			21	/* =fc,fd,fr open(2): rd,cr,tr */
#define	AUE_OPEN_W				22	/* =fw open(2): write */
#define	AUE_OPEN_WC				23	/* =fc,fw open(2): write,creat */
#define	AUE_OPEN_WT				24	/* =fd,fw open(2): write,trunc */
#define	AUE_OPEN_WTC			25	/* =fc,fd,fw open(2): wr,cr,tr */
#define	AUE_OPEN_RW				26	/* =fr,fw open(2): read,write */
#define	AUE_OPEN_RWC			27	/* =fc,fw,fr open(2): rd,wr,cr */
#define	AUE_OPEN_RWT			28	/* =fd,fr,fw open(2): rd,wr,tr */
#define	AUE_OPEN_RWTC			29	/* =fc,fd,fw,fr open(2): rd,wr,cr,tr */
#define	AUE_CREAT				30  /* =fc create(2) */
#define	AUE_CHDIR				31  /* =pc chdir(2) */
#define	AUE_FCHDIR				32	/* =pc fchdir(2) */
#define	AUE_CHMOD				33	/* =fm chmod(2) */
#define	AUE_FCHMOD				34	/* =fm fchmod(2) */
#define	AUE_CHOWN				35	/* =fm chown(2) */
#define	AUE_FCHOWN				36	/* =fm fchown(2) */
#define AUE_LCHOWN				37  /* =fm lchown(2) */
#define	AUE_UTIMES				38	/* =fm utimes(2) */
#define	AUE_UTIME				39	/* =fm old utime(2) */
#define	AUE_CHROOT				40	/* =pc chroot(2) */
#define	AUE_TRUNCATE			41	/* =fd truncate(2) */
#define	AUE_FTRUNCATE			42	/* =fd ftruncate(2) */
#define AUE_SYSLOG				43  /* =ad */
#define	AUE_VFORK				44	/* =pc vfork(2) */
#define	AUE_EXECVE				45	/* =ex execve(2) */
#define	AUE_VTRACE				46	/* =pc vtrace(2) */
#define	AUE_QUOTACTL			47	/* =ad quotactl(2) */
#define	AUE_ACCEPT				48	/* =nt accept(2) */
#define	AUE_CONNECT				49	/* =nt connect(2) */
#define	AUE_MOUNT				50	/* =ad mount(2) */
#define	AUE_SWAPON				51	/* =ad swapon(2) */
#define AUE_SWAPOFF				52  /* =ad */
#define	AUE_SETUID				53	/* =pc old setuid(2) */
#define	AUE_SETGID				54	/* =pc old setgid(2) */
#define	AUE_SETREUID			55	/* =pc setreuid(2) */
#define	AUE_SETREGID			56	/* =pc setregid(2) */
#define	AUE_SETHOSTNAME			57	/* =ad sethostname(2) */
#define	AUE_SETDOMAINNAME		58	/* =ad setdomainname(2) */
#define	AUE_SETGROUPS			59	/* =pc setgroups(2) */
#define AUE_SETFSUID			60  /* =pc */
#define AUE_SETFSGID			61  /* =pc */
#define AUE_SETRESUID			62  /* =pc */
#define AUE_SETRESGID			63  /* =pc */
#define	AUE_STIME				64 	/* =ad old stime(2) */
#define	AUE_SETTIMEOFDAY		65  /* =ad settimeofday(2) */
#define	AUE_ADJTIME				66  /* =ad adjtime(2) */
#define	AUE_SYSTEMBOOT			67	/* =na */
#define	AUE_AUDIT				68	/* =au audit(2) */
#define	AUE_SETAUDIT			69	/* =au setaudit(2) */
#define	AUE_SETAUID				70  /* =au setauid(2) */
#define	AUE_MMAP_EXEC			71	/* */
#define	AUE_SOCKET				72	/* */
#define	AUE_X					73	/* */

#define AUE_NO_PROTO			80
#define AUE_ARP_RCV				81
#define AUE_IP_RCV				82
#define AUE_TCP_RCV				83
#define AUE_UDP_RCV				84
#define AUE_ICMP_RCV			85
#define AUE_IPIP_RCV			86
#define AUE_IGMP_RCV			87
#define AUE_IPV6_RCV			88
#define AUE_IPSEC_RCV			89

/*
 * HoNA sub events. The events are grouped by there major event-id. This allows
 * a better discription of that what was happen, without using a lot of
 * major events.
 *
 * For setting a filter rule, we can use the normal event mask as a
 * 16 x 16 matrix. So we can define 16 major-HoNA events and for every event
 * 16 subevents.
 */

#define AUSE_MAX_EVT		15

/* NO PROTO */

#define AUSE_PROTO_BAD_TYPE		0

/* ARP_subevents */

#define AUSE_ARP_BAD_SIZE		0
#define AUSE_ARP_DENY			1
#define AUSE_ARP_NOT_IP			2
#define AUSE_ARP_BAD_DEV		3
#define AUSE_ARP_NOT_ETHER		4
#define AUSE_ARP_BAD_TYPE		5
#define AUSE_ARP_BAD_DEST		6
#define AUSE_ARP_BAD_SOURCE		7
#define AUSE_ARP_NO_ROUTE		8

/* IPv4-Subevents */

#define AUSE_IP_BAD_SRC			0
#define AUSE_IP_DENY			1
#define AUSE_IP_BAD_DEST		2
#define AUSE_IP_BAD_TYPE		3
#define AUSE_IP_TOBIG			4
#define AUSE_IP_BAD_PROTO		5
#define AUSE_IP_BAD_SIZE		6
#define AUSE_IP_BAD_VER			7
#define AUSE_IP_BAD_CHECK		8
#define AUSE_IP_SRR_DENY		9
#define AUSE_IP_BAD_SRR			10
#define AUSE_IP_SRR_RCV			11
#define AUSE_IP_SRR_FAILED		12

/* TCP-Subevents */

#define AUSE_TCP_BAD_DOFF		0
#define AUSE_TCP_FULL_SYNQ		1
#define AUSE_TCP_FULL_ACCEPTQ	2
#define AUSE_TCP_NO_SOCK		3
#define AUSE_TCP_BAD_FLAG		4
#define AUSE_TCP_NO_ACK			5
#define AUSE_TCP_CON_EST		6
#define AUSE_TCP_BAD_URG		7
#define AUSE_TCP_BAD_SIZE		8
#define AUSE_TCP_TO_LATE		9
#define AUSE_TCP_BAD_CHECK		10
#define AUSE_TCP_AUTH_ERR		11
#define AUSE_TCP_BAD_REQ		12
#define AUSE_TCP_WRITE_ERR		13
#define AUSE_TCP_OUT_OF_RES		14

/* UDP-Subevents */

#define AUSE_UDP_BAD_SIZE		0
#define AUSE_UDP_NO_SOCK		1
#define AUSE_UDP_BAD_CHECK		2

/* ICMP-Subevents */

#define AUSE_ICMP_BAD_SIZE		0
#define AUSE_ICMP_BAD_CHECK		1
#define AUSE_ICMP_BAD_TYPE		2
#define AUSE_ICMP_REQ			3
#define AUSE_ICMP_GM6_BAD_SIZE	4
#define AUSE_ICMP_GM6_BAD_SRC	5
#define AUSE_ICMP_GM6_DENY		6
#define AUSE_ICMP_BAD_DEST		7

/* IPIP-Subevents */

#define AUSE_IPIP_NO_TUN		0

/* IGMP-Subevents */

#define AUSE_IGMP_BAD_SIZE		0
#define AUSE_IGMP_BAD_CHECK		1
#define AUSE_IGMP_BAD_TYPE		2
#define AUSE_IGMP_DENY			3

/* IPv6-Subevents */

#define AUSE_IPV6_NO_ROUTE		0
#define AUSE_IPV6_BAD_VER		1
#define AUSE_IPV6_BAD_SIZE		3
#define AUSE_IPV6_BAD_PROTO		4
#define AUSE_IPV6_NO_DEST		5
#define AUSE_IPV6_BAD_PARAM		6
#define AUSE_IPV6_BAD_TYPE		7
#define AUSE_IPV6_TOBIG			8
#define AUSE_IPV6_BAD_ADDR		9
#define AUSE_IPV6_FRAG_ERR		10

/* IPSec-Subevents */

#define AUSE_IPSEC_BAD_TYPE			0
#define AUSE_IPSEC_ESP_BAD_SIZE		1
#define AUSE_IPSEC_AH_BAD_SIZE		2
#define AUSE_IPSEC_NO_SA			3
#define AUSE_IPSEC_INBOUND_BAD_SA	4
#define AUSE_IPSEC_LARVAL_STATE		5
#define AUSE_IPSEC_DEAD_STATE		6
#define AUSE_IPSEC_HARDBYTE_TO		7
#define AUSE_IPSEC_HARDADD_TO		8
#define AUSE_IPSEC_USE_TO			9
#define AUSE_IPSEC_PKT_TO			10
#define AUSE_IPSEC_BAD_AUTH			11
#define AUSE_IPSEC_REPLAY			12
#define AUSE_IPSEC_AUTH_ERR			13
#define AUSE_IPSEC_ESP_BAD_TYPE		14
#define AUSE_IPSEC_INBOUND_NO_SA	15


/* ressourcen events_mask */

#define AU_INIT_RES_EVENTMASK { \
	0, 1, 0, 0, 0, 0, 0, 1,  1, 1, 1, 1, 1, 1, 1, 1, \
	1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1, \
	1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 0, 0, 1, 0, 1, \
	0, 0, 1, 1, 1, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0, \
	0, 0, 0, 0, 0, 0, 0, 0,  1, 1, 0, 0, 0, 0, 0, 0, \
	0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0, \
	0, 0, 0, 0, 0, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1, \
	1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1}

#define AU_EVT_CHR_SIZE 128
#define	AU_EVT_CHR {"AUE_NULL", "AUE_ACCT", "AUE_CAPSET", "AUE_FORK", \
	"AUE_CLONE", "AUE_IOPERM", "AUE_IOPL", "AUE_INIT_MODULE", \
	"AUE_DELETE_MODULE", "AUE_LINK", "AUE_UNLINK", "AUE_MKNOD", "AUE_RENAME", \
	"AUE_MKDIR", "AUE_RMDIR", "AUE_SYMLINK", "AUE_OPEN", "AUE_USELIB", \
	"AUE_OPEN_R", "AUE_OPEN_RC", "AUE_OPEN_RT", "AUE_OPEN_RTC", "AUE_OPEN_W", \
	"AUE_OPEN_WC", "AUE_OPEN_WT", "AUE_OPEN_WTC", "AUE_OPEN_RW", \
	"AUE_OPEN_RWC", "AUE_OPEN_RWT", "AUE_OPEN_RWTC", "AUE_CREAT", "AUE_CHDIR", \
	"AUE_FCHDIR", "AUE_CHMOD", "AUE_FCHMOD", "AUE_CHOWN", "AUE_FCHOWN", \
	"AUE_LCHOWN", "AUE_UTIMES", "AUE_UTIME", "AUE_CHROOT", "AUE_TRUNCATE", \
	"AUE_FTRUNCATE", "AUE_SYSLOG", "AUE_VFORK", "AUE_EXECVE", "AUE_VTRACE", \
	"AUE_QUOTACTL", "AUE_ACCEPT", "AUE_CONNECT", "AUE_MOUNT", "AUE_SWAPON", \
	"AUE_SWAPOFF", "AUE_SETUID", "AUE_SETGID", "AUE_SETREUID", "AUE_SETREGID", \
	"AUE_SETHOSTNAME", "AUE_SETDOMAINNAME", "AUE_SETGROUPS", "AUE_SETFSUID", \
	"AUE_SETFSGID", "AUE_SETRESUID", "AUE_SETRESGID", "AUE_STIME", \
	"AUE_SETTIMEOFDAY", "AUE_ADJTIME", "AUE_SYSTEMBOOT", "AUE_AUDIT", \
	"AUE_SETAUDIT", "AUE_SETAUID", "AUE_NFS_SVC", "AUE_EXPORTFS", \
	"AUE_NFS_GETFH", \
	"AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", \
	"AUE_NO_PROTO", "AUE_ARP_RCV", "AUE_TCP_RCV", "AUE_UDP_RCV", \
	"AUE_ICMP_RCV", "AUE_IPIP_RCV", "AUE_IGMP_RCV", "AUE_IPV6_RCV", \
	"AUE_IPSEC_RCV", \
	"AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", \
	"AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", \
	"AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", \
	"AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", \
	"AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X", "AUE_X"\
}



#endif
