/*
 * Linux Operating System Audit
 *   written by Oliver Stecklina <stecklina@secunet.de>
 *
 * filename : audit_ipsec.c
 *
 * last edit : 11.06.2001
 *
 * content :
 *
 */

#include <asm/atomic.h>

#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>

#include <linux/audit.h>
#include <linux/audit_event.h>

#include <freeswan.h>
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_radij.h"
#include "ipsec_netlink.h"
#include "ipsec_xform.h"

extern struct audit_dyn_buf *au_net_env_buf;

int audit_ipsec_v4(struct sk_buff *skb, unsigned int sub_evt,
                   struct tdb *tdb_p, struct tdb *policy_tdb_p)
{
	struct audit_record *rec;
	struct audit_net_record *data;

	if ((rec = audit_get_free_record(AU_NET_REC, 0)) == NULL)
		return 0;

	data = (struct audit_net_record *) rec->data;
	if (!data) {
		rec->state = AU_UNUSED;
		return 0;
	}
	
	rec->event = AUE_IPSEC_RCV;
	rec->pid = sub_evt;

	audit_copy_mac(data, skb, skb->dev);
	audit_copy_ipv4(data, skb->nh.iph);
	audit_copy_dev(data, skb->dev, IFNAMSIZ);

	data->env = NULL;
	if (tdb_p) {
		void *p[5];
		unsigned long src1[2], dst1[2], src2[2], dst2[2];
		unsigned int entries, size;

		size = sizeof(__u32);

		src1[0] = size;
		if (tdb_p->tdb_addr_s)
			src1[1] = (unsigned long) &((struct sockaddr_in *) (tdb_p->tdb_addr_s))->sin_addr.s_addr;
		else
			src1[1] = 0;

		dst1[0] = size;
		if (tdb_p->tdb_addr_d)
			dst1[1] = (unsigned long) &((struct sockaddr_in *) (tdb_p->tdb_addr_d))->sin_addr.s_addr;
		else
			dst1[1] = 0;

		p[0] = (void *) src1;
		p[1] = (void *) dst1;

		entries = 2;
		if (policy_tdb_p) {
			src2[0] = size;
			if (policy_tdb_p->tdb_addr_s)
				src2[1] = (unsigned long) &((struct sockaddr_in *) (policy_tdb_p->tdb_addr_s))->sin_addr.s_addr;
			else
				src2[1] = 0;

			dst2[0] = size;
			if (policy_tdb_p->tdb_addr_s)
				dst2[1] = (unsigned long) &((struct sockaddr_in *) (policy_tdb_p->tdb_addr_d))->sin_addr.s_addr;
			else
				dst2[1] = 0;

			p[entries++] = (void *) src2;
			p[entries++] = (void *) dst2;
		}

		p[entries] = NULL;
		audit_fill_dyn_buf2(au_net_env_buf, &data->env, p, entries, AU_K_SPACE);
	}

	rec->state = AU_FULL;
	
	return 0;
}

